/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_file.h"
#include "biohdf_file_internal.h"
#include "biohdf_group.h"
#include "biohdf_group_internal.h"
#include "biohdf_utility.h"
#include "biohdf_utility_internal.h"
#include "biohdf_vector.h"

#include "bioh5g_alignments.h"
#include "bioh5g_alignments_internal.h"
#include "bioh5g_reads.h"



/*****************************************************************************
 * Internal static functions                                                 *
 *****************************************************************************/



/* ITERATOR TRAVERSALS */


/* Check if the alignment passes iterator filters (region, MAPQ, etc.) */
static int
_does_align_pass_iter_filters(bioh5g_alignments_iterator iter,
                              bioh5g_alignment_data *data);



/* Get the basic alignment data from the dataset */
static biohdf_error
_fill_basic_align_data(bioh5g_alignments aligns,
                       int64_t index,
                       /* IN-OUT */ bioh5g_alignment_data *data);



/* Fill the other (template, CIGAR, etc) data from the datasets */
static biohdf_error
_fill_remaining_align_data(bioh5g_alignments aligns,
                           int64_t index,
                           /* IN-OUT */ bioh5g_alignment_data *data);



/* Get the next alignment from an unindexed dataset.
 * Called by the master itertor "get next" function.
 */
static biohdf_error
_get_next_alignment_unindexed(bioh5g_alignments_iterator iter,
                              /*OUT*/ int64_t *index,
                              /*OUT*/ bioh5g_alignment_data **data);



/* Get the next alignment from an indexed dataset.
 * Called by the master itertor "get next" function.
 */
//static biohdf_error
//_get_next_alignment_indexed(bioh5g_alignments_iterator iter,
//                            /*OUT*/ int64_t *index,
//                            /*OUT*/ bioh5g_alignment_data **data);



/* Search through an index to find the first alignment that is within
 * the region.
 *
 * The iterator will be updated with the new information.  The iterator's
 * current index will be -1 when no alignments were found in the region.
 */
static biohdf_error
_search_index_for_region_start(bioh5g_alignments_iterator iter);


/* Perform a linear scan over the index, searching for a valid alignment
 * in the region.
 *
 * This sets the index and data element as per BIOH5Gget_next_alignment().
 */
static biohdf_error
_scan_index_for_next_alignment(bioh5g_alignments_iterator iter,
                               /*OUT*/ int64_t *index,
                               /*OUT*/ bioh5g_alignment_data **data);



/* INDEXING */

/* Create filename for temp file */
static biohdf_error
_get_temp_file_name(char *path, char *prefix, int n, /*OUT*/ char **filename);



/* Write a temp index file */
static biohdf_error
_write_index_temp_file(ref_pos_data *data, size_t n, char *filename);



/* Read data from an index temp file */
static biohdf_error
_read_index_data(FILE *f, /*OUT*/ ref_pos_data *data);



/* qsort()-compatible comparator for creating alignment hit indexes
 * based on reference and position.
 * The elements are both ref_pos_data structs (defined above).
 */
static int
_ref_pos_comparator(const void *elem1, const void *elem2);



/* TYPES */

/* Create the HDF5 enum type that describes the indexing scheme */
static biohdf_error
_create_index_enum_HDF5_type(/*OUT*/ hid_t *tid);



/* Create the HDF5 type for storing basic hit data */
static biohdf_error
_create_alignment_HDF5_type(/*OUT*/ hid_t *tid);



/* Create the HDF5 type for storing basic paired end data */
static biohdf_error
_create_template_HDF5_type(/*OUT*/ hid_t *tid);



/* ERRORS */

/* Emit a standard error message */
static void
_print_error_info(const char *collection_name)
{
    fprintf(stderr, "             \tBioHDF object: alignments collection\n");
    if(NULL == collection_name)
    {
        fprintf(stderr, "             \talignments name: NULL");
    }
    else
    {
        fprintf(stderr, "             \talignments name: %s", collection_name);
    }
    fprintf(stderr, "\n\n");

    return;
}



/*****************************************************************************
 * Create, open, close                                                       *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcheck_alignments_presence(const biohdf_file file,
                                const char *path,
                                /*OUT*/ int *presence)
{
    htri_t check;
    char **groups = NULL;    /* Returned set of groups as strings */
    int n_groups;            /* Number of groups returned */
    int i;                   /* iterator */

    hid_t new_location = -1; /* The new group to be tested as we traverse
                              * the path.
                              */
    hid_t old_location = -1; /* The group in which the new group is tested.
                              * This starts at the file id on the first
                              * and is replaced with group IDs as we traverse
                              * the path.
                              */
    char *type_name;          /* The BioHDF type name from the collection */

    char *        bad_group        = NULL;           /* The group on which an operation failed */

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Get a list of the groups in the path */
    error = _parse_groups_from_path(path, &groups, &n_groups);
    CHECK_BIO(error);

    /* Iterate over the groups, checking to see if they exist */
    old_location = file->file_id;

    for(i = 0; i < n_groups; i++)
    {
        /* Does this group exist? */
        check = (int)H5Lexists(old_location, groups[i], H5P_DEFAULT);
        if(check > 0)
        {
            /* Present. Move on to the next group */
            *presence = 1;
            new_location = H5Gopen(old_location, groups[i], H5P_DEFAULT);
            CHECK_HID_T(new_location);
        }
        else if (0 == check)
        {
            /* Not present.  Return. */
            *presence = 0;
            break;
        }
        else
        {
            /* A negative (error) value was returned  Return.*/
            *presence = 0;
            bad_group = groups[i];
            BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
            goto FAIL;
        }

        /* For the last group, we need to test if it is a collection of the
         * correct type.
         */
        if(n_groups - 1 == i)
        {
            error = _read_biohdf_type_attribute(new_location, &type_name);
            if(error != BIOHDF_NO_ERROR || 0 != strcmp(type_name, BIOH5G_ALIGNMENTS_TYPE))
            {
                *presence = 0;
                BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
                goto FAIL;
            }

            hdf_error = H5Gclose(new_location);
            CHECK_HERR_T(hdf_error);
        }

        /* Close the old "containing" group, except on the first iteration
         * (we don't want to close the file, which is old_location on
         * the first iteration).
         */
        if(H5I_GROUP == H5Iget_type(old_location))
        {
            hdf_error = H5Gclose(old_location);
            CHECK_HERR_T(hdf_error);
        }

        /* The old old/containing group is replaced with the new group */
        old_location = new_location;
    }

    /**SUCCESS***************************************************************/

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);


    if(1 == H5Iis_valid(new_location))
    {
        H5Gclose(new_location);
    }

    /* Don't close the old location if it's the file_id that was passed in! */
    if(1 == H5Iis_valid(old_location) && H5I_GROUP == H5Iget_type(old_location))
    {
        H5Gclose(old_location);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gcreate_alignments_collection(const biohdf_file file,
                                   const bioh5g_alignments_creation_properties props,
                                   const char *path,
                                   /*OUT*/ bioh5g_alignments *aligns)
{
    vector_storage_scheme refs_scheme;
    vector_storage_scheme cigar_scheme;
    vector_storage_scheme tags_scheme;

    /* For writing attributes */
    hid_t    type_id;
    hid_t    space_id;
    hid_t    attr_id;
    hsize_t  dims[1];

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *aligns = NULL;

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Create handle */
    *aligns = (bioh5g_alignments)calloc(1, sizeof(struct _bioh5g_alignments));
    CHECK_PTR(*aligns);

    /* Create path string. */
    (*aligns)->path = (char *)calloc(strlen(path) + 1, sizeof(char));
     CHECK_PTR((*aligns)->path);
    strcpy((*aligns)->path, path);

    /* Create hit data type */
    error = _create_alignment_HDF5_type(&((*aligns)->basic_data_hdf5_type));
    CHECK_BIO(error);

    /* Create template data type */
    error = _create_template_HDF5_type(&((*aligns)->template_hdf5_type));
    CHECK_BIO(error);

    /* Create groups */
    error = _create_or_open_group(file, path, &((*aligns)->group));
    CHECK_BIO(error);


    /* Add reads path attribute to the alignments group
     * TODO: use the attribute API to store this
     */
    dims[0] = strlen(props->reads_path); /* don't store the trailing \0 */
    space_id = H5Screate_simple (1, dims, NULL);
    CHECK_HID_T(space_id);

    attr_id = H5Acreate((*aligns)->group, BIOH5G_READS_PATH_ATTR,
        H5T_NATIVE_CHAR, space_id, H5P_DEFAULT, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    hdf_error = H5Awrite(attr_id, H5T_NATIVE_CHAR, props->reads_path);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Sclose(space_id);
    CHECK_HERR_T(hdf_error);


    /* Add BioHDF type attribute to the collection */

    error = _write_biohdf_type_attribute((*aligns)->group, BIOH5G_ALIGNMENTS_TYPE);
    CHECK_BIO(error);



    /* Add sort attribute to the alignments group */

    dims[0] = 1;
    space_id = H5Screate_simple (1, dims, NULL);
    CHECK_HID_T(space_id);

    error = _create_index_enum_HDF5_type(&type_id);
    CHECK_BIO(error);

    (*aligns)->index_method = UNINDEXED;
    (*aligns)->aligns_index = NULL;

    attr_id = H5Acreate((*aligns)->group, BIOH5G_INDEX_METHOD_ATTR,
        type_id, space_id, H5P_DEFAULT, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    hdf_error = H5Awrite(attr_id, type_id, &((*aligns)->index_method));
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Sclose(space_id);
    CHECK_HERR_T(hdf_error);


    /* Create data containers */

    switch(props->refs_scheme)
    {
    case FIXED_LENGTH_STRING:
        refs_scheme = FIXED_SIZE_ARRAY;
        break;
    case VARIABLE_LENGTH_STRING:
        refs_scheme = VARIABLE_SIZE_ARRAY;
        break;
    case HYBRID_STRING:
        refs_scheme = HYBRID_ARRAY;
        break;
    default:
        return BIOHDF_BADNESS;
        break;
    }

    switch(props->tags_scheme)
    {
    case FIXED_LENGTH_STRING:
        tags_scheme = FIXED_SIZE_ARRAY;
        break;
    case VARIABLE_LENGTH_STRING:
        tags_scheme = VARIABLE_SIZE_ARRAY;
        break;
    case HYBRID_STRING:
        tags_scheme = HYBRID_ARRAY;
        break;
    default:
        return BIOHDF_BADNESS;
        break;
    }

    switch(props->cigar_scheme)
    {
    case FIXED_LENGTH_STRING:
        cigar_scheme = FIXED_SIZE_ARRAY;
        break;
    case VARIABLE_LENGTH_STRING:
        cigar_scheme = VARIABLE_SIZE_ARRAY;
        break;
    case HYBRID_STRING:
        cigar_scheme = HYBRID_ARRAY;
        break;
    default:
        return BIOHDF_BADNESS;
        break;
    }

    /* Basic alignment data */
    error = _create_vector((*aligns)->group,
        BIOH5G_HIT_DATA_NAME,
        (*aligns)->basic_data_hdf5_type,
        FIXED_SIZE_ARRAY,
        1,
        props->chunk_size,
        props->compression,
        &((*aligns)->basic_data));
    CHECK_BIO(error);

    /* Template data for paired-end reads, etc. */
    error = _create_vector((*aligns)->group,
        BIOH5G_TEMPLATE_DATA_NAME,
        (*aligns)->template_hdf5_type,
        FIXED_SIZE_ARRAY,
        1,
        props->chunk_size,
        props->compression,
        &((*aligns)->templates));
    CHECK_BIO(error);

    /* Reference names */
    error = _create_vector((*aligns)->group,
        BIOH5G_REFERENCES_NAME,
        H5T_NATIVE_CHAR,
        refs_scheme,
        props->refs_length,
        props->chunk_size,
        props->compression,
        &((*aligns)->references));
    CHECK_BIO(error);

    /* SAM header */
    error = _create_vector((*aligns)->group,
        BIOH5G_SAM_HEADER_NAME,
        H5T_NATIVE_CHAR,
        VARIABLE_SIZE_ARRAY,
        -1,
        props->chunk_size,
        props->compression,
        &((*aligns)->sam_header));
    CHECK_BIO(error);

    /* SAM CIGAR strings */
    error = _create_vector((*aligns)->group,
        BIOH5G_SAM_CIGAR_NAME,
        H5T_NATIVE_CHAR,
        cigar_scheme,
        props->cigar_length,
        props->chunk_size,
        props->compression,
        &((*aligns)->sam_cigar));
    CHECK_BIO(error);

    /* SAM tags */
    error = _create_vector((*aligns)->group,
        BIOH5G_SAM_TAGS_NAME,
        H5T_NATIVE_CHAR,
        tags_scheme,
        props->tags_length,
        props->chunk_size,
        props->compression,
        &((*aligns)->sam_tags));
    CHECK_BIO(error);

    /* Next reference string (SAM RNEXT) */
    error = _create_vector((*aligns)->group,
        BIOH5G_SAM_RNEXT_NAME,
        H5T_NATIVE_CHAR,
        refs_scheme,
        props->refs_length,
        props->chunk_size,
        props->compression,
        &((*aligns)->sam_rnext));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);

    if(*aligns != NULL)
    {
        /* Free the path string */
        free((*aligns)->path);

        /* Free allocated memory */
        free((*aligns)->reads_path);

        /* Free HDF5 objects */
        if(H5Iis_valid((*aligns)->basic_data_hdf5_type)) { H5Tclose((*aligns)->basic_data_hdf5_type); }
        if(H5Iis_valid((*aligns)->template_hdf5_type))   { H5Tclose((*aligns)->template_hdf5_type); }
        if(H5Iis_valid((*aligns)->group))                { H5Gclose((*aligns)->group); }

        /* Free vectors */
        _close_vector(&((*aligns)->basic_data));
        _close_vector(&((*aligns)->references));
        _close_vector(&((*aligns)->sam_cigar));
        _close_vector(&((*aligns)->sam_header));
        _close_vector(&((*aligns)->sam_rnext));
        _close_vector(&((*aligns)->sam_tags));
        _close_vector(&((*aligns)->templates));
    }

    free(*aligns);
    *aligns = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gopen_alignments_collection(const biohdf_file file,
                                 const char *path,
                                 biohdf_open_mode mode,
                                 /*OUT*/ bioh5g_alignments *aligns)
{
    hid_t     attr_id    = -1;
    hid_t     space_id   = -1;
    hid_t     type_id    = -1;
    hsize_t   n_chars    = 0;

    int is_read_only;

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/
    
    *aligns = NULL;

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Create handle. */
    *aligns = (bioh5g_alignments)calloc(1, sizeof(struct _bioh5g_alignments));
    CHECK_PTR(*aligns);

    /* Open groups */
    error = _create_or_open_group(file, path, &((*aligns)->group));
    CHECK_BIO(error);

    /* Create path string. */
    (*aligns)->path = (char *)calloc(strlen(path) + 1, sizeof(char));
    CHECK_PTR((*aligns)->path);
    strcpy((*aligns)->path, path);

    /* Get reads path.  TODO: use attribute API to read this. */
    attr_id = H5Aopen((*aligns)->group, BIOH5G_READS_PATH_ATTR, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    space_id = H5Aget_space(attr_id);
    CHECK_HID_T(space_id);

    n_chars = H5Sget_simple_extent_npoints(space_id) + 1;
    (*aligns)->reads_path = (char *)calloc((size_t)n_chars, sizeof(char));
    CHECK_PTR((*aligns)->reads_path);

    hdf_error = H5Aread(attr_id, H5T_NATIVE_CHAR, (*aligns)->reads_path);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Sclose(space_id);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    /* Read index attribute */
    attr_id = H5Aopen((*aligns)->group, BIOH5G_INDEX_METHOD_ATTR, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    error = _create_index_enum_HDF5_type(&type_id);
    CHECK_BIO(error);

    hdf_error = H5Aread(attr_id, type_id, &((*aligns)->index_method));
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Tclose(type_id);
    CHECK_HERR_T(hdf_error);

    /* Create basic alignment data type */
    error = _create_alignment_HDF5_type(&((*aligns)->basic_data_hdf5_type));
    CHECK_BIO(error);

    /* Create template data type */
    error = _create_template_HDF5_type(&((*aligns)->template_hdf5_type));
    CHECK_BIO(error);

    /* Open datasets */
    if (READ_ONLY == mode)
    {
        is_read_only = 1;
    }
    else
    {
        is_read_only = 0;
    }

    /* Basic alignment data */
    error = _open_vector((*aligns)->group,
        BIOH5G_HIT_DATA_NAME,
        is_read_only,
        &((*aligns)->basic_data));
    CHECK_BIO(error);

    /* Template data for paired-end reads, etc. */
    error = _open_vector((*aligns)->group,
        BIOH5G_TEMPLATE_DATA_NAME,
        is_read_only,
        &((*aligns)->templates));
    CHECK_BIO(error);

    /* Reference names */
    error = _open_vector((*aligns)->group,
        BIOH5G_REFERENCES_NAME,
        is_read_only,
        &((*aligns)->references));
    CHECK_BIO(error);

    /* SAM header */
    error = _open_vector((*aligns)->group,
        BIOH5G_SAM_HEADER_NAME,
        is_read_only,
        &((*aligns)->sam_header));
    CHECK_BIO(error);

    /* SAM CIGAR strings */
    error = _open_vector((*aligns)->group,
        BIOH5G_SAM_CIGAR_NAME,
        is_read_only,
        &((*aligns)->sam_cigar));
    CHECK_BIO(error);

    /* SAM tags */
    error = _open_vector((*aligns)->group,
        BIOH5G_SAM_TAGS_NAME,
        is_read_only,
        &((*aligns)->sam_tags));
    CHECK_BIO(error);

    /* Next reference string (SAM RNEXT) */
    error = _open_vector((*aligns)->group,
        BIOH5G_SAM_RNEXT_NAME,
        is_read_only,
        &((*aligns)->sam_rnext));
    CHECK_BIO(error);

    /* Open the alignments index, if it exists */
    if(H5Lexists((*aligns)->group, BIOH5G_REF_POS_INDEX_NAME, H5P_DEFAULT) > 0)
    {
        error = _open_vector((*aligns)->group,
            BIOH5G_REF_POS_INDEX_NAME,
            is_read_only,
            &((*aligns)->aligns_index));
        CHECK_BIO(error);
    }
    else
    {
        (*aligns)->aligns_index = NULL;
    }



    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    _print_error_info(path);

    if(*aligns != NULL)
    {
        /* Free the path string */
        free((*aligns)->path);

        /* Free allocated memory */
        free((*aligns)->reads_path);

        /* Free HDF5 objects */
        if(H5Iis_valid((*aligns)->basic_data_hdf5_type)) { H5Tclose((*aligns)->basic_data_hdf5_type); }
        if(H5Iis_valid((*aligns)->template_hdf5_type))   { H5Tclose((*aligns)->template_hdf5_type); }
        if(H5Iis_valid((*aligns)->group))                { H5Gclose((*aligns)->group); }

        /* Free vectors */
        _close_vector(&((*aligns)->basic_data));
        _close_vector(&((*aligns)->references));
        _close_vector(&((*aligns)->sam_cigar));
        _close_vector(&((*aligns)->sam_header));
        _close_vector(&((*aligns)->sam_rnext));
        _close_vector(&((*aligns)->sam_tags));
        _close_vector(&((*aligns)->templates));

        /* Close index */
        _close_vector(&((*aligns)->aligns_index));

    }

    free(*aligns);
    *aligns = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gclose_alignments_collection(/*IN-OUT*/ bioh5g_alignments *aligns)
{
    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* None - NULL is ok */

    /**CODE*****************************************************************/

    if(NULL == *aligns){ return BIOHDF_NO_ERROR; }

    /* Free the path strings */
    free((*aligns)->path);
    free((*aligns)->reads_path);

    /* Close BioHDF storage */
    error = _close_vector(&((*aligns)->basic_data));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->templates));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->references));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->sam_header));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->sam_cigar));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->sam_tags));
    CHECK_BIO(error);
    error = _close_vector(&((*aligns)->sam_rnext));
    CHECK_BIO(error);

    /* Close the index */
    error = _close_vector(&((*aligns)->aligns_index));
    CHECK_BIO(error);

    /* Close the containing group. */
    hdf_error = H5Gclose((*aligns)->group);
    CHECK_HERR_T(hdf_error);

    /* Close the HDF5 types */
    hdf_error = H5Tclose((*aligns)->basic_data_hdf5_type);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Tclose((*aligns)->template_hdf5_type);
    CHECK_HERR_T(hdf_error);

    free(*aligns);
    *aligns = NULL;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    /* If aligns were NULL, we'd have exited successfully */
    _print_error_info((*aligns)->path);

    /* Try again, but don't report errors
     * This may seem redundant, but I WANT close errors to be reported since
     * they may indicate underlying problems.
     */
    if(*aligns != NULL)
    {
        /* close HDF5 objects */
        if(H5Iis_valid((*aligns)->group)) { H5Gclose((*aligns)->group); }

        /* close vectors */
        _close_vector(&((*aligns)->basic_data));
        _close_vector(&((*aligns)->templates));
        _close_vector(&((*aligns)->references));
        _close_vector(&((*aligns)->sam_header));
        _close_vector(&((*aligns)->sam_cigar));
        _close_vector(&((*aligns)->sam_tags));
        _close_vector(&((*aligns)->sam_rnext));

        /* Close index */
        _close_vector(&((*aligns)->aligns_index));
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gget_reads_path(const bioh5g_alignments aligns,
                     /* OUT */ char **reads_path)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    (*reads_path) = (char *)calloc(strlen(aligns->reads_path) + 1, sizeof(char));
    CHECK_PTR(*reads_path);
    strcpy(*reads_path, aligns->reads_path);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    *reads_path = NULL;

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Append and read data                                                      *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gget_alignments_count(const bioh5g_alignments aligns,
                           /*OUT*/ int64_t *count)
{
    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    error = _get_vector_count(aligns->basic_data, (hsize_t *)count);
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    *count = -1;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_iterator(const bioh5g_alignments aligns,
                                 /*OUT*/ bioh5g_alignments_iterator *iter)
{
    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *iter = NULL;

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    *iter = (bioh5g_alignments_iterator)calloc(1,
        sizeof(struct _bioh5g_alignments_iterator));
    CHECK_PTR(*iter);

    /* Set the alignments reference */
    (*iter)->alignments = aligns;

    /* All filters are initially off */
    (*iter)->has_region_filters    = 0;
    (*iter)->has_mapq_filter       = 0;
    (*iter)->has_sam_flags_filter  = 0;

    /* Initialize filter data */
    (*iter)->n_regions       = 0;
    (*iter)->regions         = NULL;
    (*iter)->min_mapq        = 255;     /* max MAPQ value */
    (*iter)->sam_flags_mask  = 0;       /* all flags unset */

    /* Initialize range query flags */
    (*iter)->current_index   = 0;
    (*iter)->current_region  = 0;

    error = BIOH5Gget_alignments_count(aligns, &((*iter)->max_index));
    CHECK_BIO(error);
    (*iter)->max_index -= 1;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    free(*iter);
    *iter = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_range_filter(bioh5g_alignments_iterator iter,
                                           const char *reference,
                                           int32_t start,
                                           int32_t end)
{
    int i;

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(iter);
    CHECK_PTR(reference);

    /**CODE*****************************************************************/

    iter->has_region_filters = 1;

    /* Check for overlap with existing filters */
    for(i = 0; i < iter->n_regions; i++)
    {
        if(!strcmp(reference, iter->regions[i].reference)
            && (end >= iter->regions[i].start || start <= iter->regions[i].end))
        {
            /* Merge overlapping regions and return */
            if(start < iter->regions[i].start)
            {
                iter->regions[i].start = start;
            }
            if(end < iter->regions[i].end)
            {
                iter->regions[i].end = end;
            }
            return BIOHDF_NO_ERROR;
        }
    }

    /* Add a new region if there is no overlap */

    iter->n_regions += 1;

    iter->regions = (bioh5g_reference_region *)realloc(iter->regions,
        (size_t)iter->n_regions * sizeof(bioh5g_reference_region));
    CHECK_PTR(iter->regions);

    iter->regions[iter->n_regions - 1].end = end;
    iter->regions[iter->n_regions - 1].start = start;
    strcpy(iter->regions[iter->n_regions - 1].reference, reference);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->alignments) {
        _print_error_info(iter->alignments->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_mapq_filter(bioh5g_alignments_iterator iter,
                                          unsigned char min_mapq)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(iter);

    /**CODE*****************************************************************/

    iter->has_mapq_filter = 1;
    iter->min_mapq = min_mapq;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->alignments) {
        _print_error_info(iter->alignments->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_flags_filter(bioh5g_alignments_iterator iter,
                                           uint32_t mask)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(iter);

    /**CODE*****************************************************************/

    iter->has_sam_flags_filter = 1;
    iter->sam_flags_mask = mask;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->alignments) {
        _print_error_info(iter->alignments->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gdestroy_alignments_iterator(/*IN-OUT*/ bioh5g_alignments_iterator *iter)
{
    /* Can't really fail here */

    if(NULL == *iter)
    {
        return BIOHDF_NO_ERROR;
    }

    free((*iter)->regions);
    free(*iter);
    *iter = NULL;

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gadd_alignment(const bioh5g_alignments aligns,
                    const bioh5g_alignment_data *data)
{
    /* HDF5 data storage structures */
    basic_alignment_data   a;
    template_data          t;

    biohdf_error    error   = BIOHDF_BADNESS;

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);
    CHECK_PTR(data);

    /**CODE*****************************************************************/

    /* Copy data from the in-memory struct to the HDF5 data storage structs */
    a.read_index = data->read_index;
    a.position   = data->position;
    a.length     = data->length;
    a.sam_mapq   = data->sam_mapq;
    a.sam_flags  = data->sam_flags;

    t.sam_pnext  = data->sam_pnext;
    t.sam_tlen   = data->sam_tlen;

    /* Store alignment data */
    error = _add_vector_element(aligns->basic_data, &a, 1);
    CHECK_BIO(error);

    /* Store template data */
    error = _add_vector_element(aligns->templates, &t, 1);
    CHECK_BIO(error);

    /* Store the reference string */
    error = _add_vector_element(aligns->references,
        data->reference,
        strlen(data->reference));
    CHECK_BIO(error);

    /* Store the CIGAR string */
    error = _add_vector_element(aligns->sam_cigar,
        data->sam_cigar,
        strlen(data->sam_cigar));
    CHECK_BIO(error);

    /* Store the tags string */
    error = _add_vector_element(aligns->sam_tags,
        data->sam_tags,
        strlen(data->sam_tags));
    CHECK_BIO(error);

    /* Store the rnext string */
    error = _add_vector_element(aligns->sam_rnext,
        data->sam_rnext,
        strlen(data->sam_rnext));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_index_of_last_added_alignment(const bioh5g_alignments aligns,
                                        /*OUT*/ int64_t *index)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);
    CHECK_PTR(index);

    /**CODE*****************************************************************/

    *index = (int64_t)(aligns->basic_data->index_of_last_added_element);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    *index = -1;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_next_alignment(bioh5g_alignments_iterator iter,
                         /*OUT*/ int64_t *index,
                         /*OUT*/ bioh5g_alignment_data **data)
{
    biohdf_error error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(index);
    CHECK_PTR(iter);
    CHECK_PTR(data);

    *index = -1;
    *data  = NULL;

    /**CODE*****************************************************************/

    if(REF_POS_SECONDARY == iter->alignments->index_method
        && !(iter->has_region_filters))
    {
        /* No need for binary search to locate alignments 
         * so we use the unsorted algorithm.
         */
        error = _get_next_alignment_unindexed(iter, index, data);
        CHECK_BIO(error);
    }
    else if(REF_POS_SECONDARY == iter->alignments->index_method)
    {
        error = _get_next_alignment_unindexed(iter, index, data);
        //error = _get_next_alignment_indexed(iter, index, data);
        CHECK_BIO(error);
    }
    else if(UNINDEXED == iter->alignments->index_method)
    {
        error = _get_next_alignment_unindexed(iter, index, data);
        CHECK_BIO(error);
    }
    else
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->alignments) {
        _print_error_info(iter->alignments->path);
    }
    else {
        _print_error_info(NULL);
    }

    BIOH5Gfree_alignment_data(data);
    *data = NULL;
    *index = -1;

    return BIOHDF_BADNESS;
}



static biohdf_error
_search_index_for_region_start(bioh5g_alignments_iterator iter)
{
    int64_t start = -1;
    int64_t mid   = -1;
    int64_t end   = -1;

    int     found = 0;
    int     cmp   = 0;

    ref_pos_data rpd_test;
    ref_pos_data rpd_from_idx;

    size_t ref_len = 0;

    biohdf_error error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(iter);

    /**CODE*****************************************************************/

    /* Search alignment has the same reference as the region and 
     * a start position at the first index of the region.
     */
    rpd_test.position = 0;
    ref_len = strlen(iter->regions[iter->current_region].reference);
    rpd_test.reference = (char *)calloc(ref_len, sizeof(char));
    CHECK_PTR(rpd_test.reference);
    strcpy(iter->regions[iter->current_region].reference,
        rpd_test.reference);

    /* Set start, end */
    start = 0;
    error = BIOH5Gget_alignments_count(iter->alignments, &end);
    CHECK_BIO(error);

    /* While not found... */
    while(!found)
    {
        /* Calculate mid */
        mid = (end - start) / 2;

        /* Get basic data @ mid */

        /* Compare values at mid */
        cmp = _ref_pos_comparator(&rpd_test, &rpd_from_idx);

        /* Re-evaluate mid and found based on < = > */
        if(cmp < 0)
        {
            /* Too low. Check higher elements. */
            mid = start;
        }
        if(cmp > 0)
        {
            /* Too high. Check lower elements. */
            mid = end;
        }
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    return BIOHDF_BADNESS;
}



static biohdf_error
_scan_index_for_next_alignment(bioh5g_alignments_iterator iter,
                               /*OUT*/ int64_t *index,
                               /*OUT*/ bioh5g_alignment_data **data)
{
    int is_valid = 0;

    biohdf_error error;                 /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(index);
    CHECK_PTR(iter);
    CHECK_PTR(data);

    /**CODE*****************************************************************/

    while(!is_valid)
    {
        /* Get the next alignment ID from the index */

        /* Get alignment data at that ID */
        error = BIOH5Gget_alignment(iter->alignments,*index, data);
        CHECK_BIO(error);

        /* Make sure the references are the same.
         * If not, we are done with this region.
         */
        if(strcmp(iter->regions[iter->current_region].reference,
            (*data)->reference))
        {
            error = BIOH5Gfree_alignment_data(data);
            CHECK_BIO(error);
            break;
        }

        /* Check filters */
        is_valid = _does_align_pass_iter_filters(iter, *data);
        if(!is_valid)
        {
            error = BIOH5Gfree_alignment_data(data);
            CHECK_BIO(error);
        }
    }


    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    BIOH5Gfree_alignment_data(data);

    *index = -1;
    *data  = NULL;

    return BIOHDF_BADNESS;
}



static int
_does_align_pass_iter_filters(bioh5g_alignments_iterator iter,
                              bioh5g_alignment_data *data)
{
    int i = -1;

    assert(iter);
    assert(data);

    /* Check to see if the MAPQ is ok */
    if(iter->has_mapq_filter && data->sam_mapq < iter->min_mapq)
    {
        /* Bad MAPQ (FAIL) */
        return 0;
    }

    /* Check to see if the the SAM flags are ok */
    if(iter->has_sam_flags_filter
        && iter->sam_flags_mask != (data->sam_flags & iter->sam_flags_mask))
    {
        /* Bad SAM flags (FAIL) */
        return 0;
    }

    /* Check to see if the alignment is in the region */
    if(!(iter->has_region_filters))
    {
        /* No regions to test (PASS) */
        return 1;
    }
    else
    {
        /* Exhaustively check all regions until a match is found */
        for(i = 0; i < iter->n_regions; i++)
        {
            if(data->position <= (iter->regions[i]).end &&
                data->position + data->length - 1 >= (iter->regions[i]).start &&
                0 == strcmp(data->reference, (iter->regions[i]).reference)
                )
            {
                /* Match to a region (PASS) */
                return 1;
            }
        }
    }

    /* No regions matched (FAIL) */
    return 0;
}



static biohdf_error
_fill_basic_align_data(bioh5g_alignments alignments,
                       int64_t index,
                       /* IN-OUT */ bioh5g_alignment_data *data)
{
    basic_alignment_data  *a = NULL;    /* For reading from HDF5 */

    int is_c_string;                    /* Append ASCII nul to strings */

    biohdf_error error;                 /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    assert(alignments);
    assert(index >= 0);
    assert(data);

    /**CODE*****************************************************************/

    /* Get template data */
    a = (basic_alignment_data *)calloc(1, sizeof(basic_alignment_data));
    CHECK_PTR(a);

    is_c_string = 0;
    error = _get_vector_element(alignments->basic_data,
        index, 
        is_c_string, 
        (void **)(&a));
    CHECK_BIO(error);

    is_c_string = 1;
    error = _get_vector_element(alignments->references,
        index,
        is_c_string, 
        (void **)&(data->reference));
    CHECK_BIO(error);

    data->read_index = a->read_index;
    data->position   = a->position;
    data->length     = a->length;
    data->sam_mapq   = a->sam_mapq;
    data->sam_flags  = a->sam_flags;

    /**SUCCESS***************************************************************/

    free(a);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    free(a);
    free(data->reference);

    return BIOHDF_BADNESS;
}



static biohdf_error
_fill_remaining_align_data(bioh5g_alignments alignments,
                           int64_t index,
                           /* IN-OUT */ bioh5g_alignment_data *data)
{
    template_data *t = NULL;    /* For reading from HDF5 */

    int is_c_string;            /* Append ASCII nul to strings */

    biohdf_error error;         /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    assert(alignments);
    assert(index >= 0);
    assert(data);

    /**CODE*****************************************************************/

    /* Get template data */
    t = (template_data *)calloc(1, sizeof(template_data));
    CHECK_PTR(t);

    is_c_string = 0;
    error = _get_vector_element(alignments->templates, 
        index, 
        is_c_string, 
        (void **)(&t));
    CHECK_BIO(error);

    data->sam_pnext  = t->sam_pnext;
    data->sam_tlen   = t->sam_tlen;

    /* Get the CIGAR string */
    is_c_string = 1; /* string */
    error = _get_vector_element(alignments->sam_cigar, 
        index,
        is_c_string,
        (void **)&(data->sam_cigar));
    CHECK_BIO(error);

    /* Get the SAM tag string */
    is_c_string = 1; /* string */
    error = _get_vector_element(alignments->sam_tags, 
        index,
        is_c_string,
        (void **)&(data->sam_tags));
    CHECK_BIO(error);

    /* Get the SAM rnext string */
    is_c_string = 1; /* string */
    error = _get_vector_element(alignments->sam_rnext, 
        index,
        is_c_string,
        (void **)&(data->sam_rnext));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    free(t);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    free(t);
    free(data->sam_cigar);
    free(data->sam_tags);
    free(data->sam_rnext);

    return BIOHDF_BADNESS;

}



static biohdf_error
_get_next_alignment_unindexed(bioh5g_alignments_iterator iter,
                              /*OUT*/ int64_t *index,
                              /*OUT*/ bioh5g_alignment_data **data)
{
    int found   = 0;    /* flag for alignment filter match */

    biohdf_error error = BIOHDF_BADNESS;    /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *index = -1;
    *data  = NULL;

    assert(iter);

    /**CODE*****************************************************************/

    /* Set up output data structure */
    *data = (bioh5g_alignment_data *)calloc(1, sizeof(bioh5g_alignment_data));
    CHECK_PTR(*data);


    /* Look for an alignment in the allowed regions */

    /* Traverse the alignments until we find one that meets our criteria
     * or we run out of alignments.
     */
    while(0 == found && (iter->current_index <= iter->max_index))
    {
        /* WHEN THE DATA ARE SORTED BUT THERE ARE NO REGION FILTERS,
         * THE ALGORITHM IS IDENTICAL TO THE UNSORTED TRAVERSAL.
         * WE STILL HAVE TO USE THE INDEX, THOUGH, SO THAT THE OUTPUT
         * IS IN SORTED ORDER.
         *
         * THAT CODE WILL GO HERE.
         */

        /* Read the basic alignment data */
        error = _fill_basic_align_data(iter->alignments,
            iter->current_index,
            *data);
        CHECK_BIO(error);

        /* Check to see if the alignment passes the filters */
        found = _does_align_pass_iter_filters(iter, *data);

        /* If this alignment isn't in the region, clean up for the next try */
        if(0 == found)
        {
            iter->current_index += 1;
            free((*data)->reference);
        }
    }

    /* Add rest of data if we had a good match */
    if(found)
    {
        error = _fill_remaining_align_data(iter->alignments,
            iter->current_index,
            *data);
        CHECK_BIO(error);

        /* Increment the iterator */
        *index = iter->current_index;
        iter->current_index += 1;
    }
    else
    {
        free(*data);
        *data = NULL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->alignments) {
        _print_error_info(iter->alignments->path);
    }
    else {
        _print_error_info(NULL);
    }

    BIOH5Gfree_alignment_data(data);
    *index = -1;
    *data = NULL;

    return BIOHDF_BADNESS;
}



//static biohdf_error
//_get_next_alignment_indexed(bioh5g_alignments_iterator iter,
//                            /*OUT*/ int64_t *index,
//                            /*OUT*/ bioh5g_alignment_data **data)
//{
//    int found       = 0;    /* Set when a suitable alignment is found */
//
//    int nothing_matched    = 0;
//
//    biohdf_error error  = BIOHDF_BADNESS; /* BioHDF error code */
//
//    /**CHECK*PARAMETERS*****************************************************/
//
//    *index = -1;
//    *data  = NULL;
//
//    assert(iter);
//
//    /**CODE*****************************************************************/
//
//    /* Look for a suitable alignment until we either match or run out of
//     * regions (we break on a match).
//     */
//    while(iter->current_region < iter->n_regions)
//    {
//        /* Check to see if this is a new region.
//         * If so, we have to bsearch in the index to find the first
//         * alignment in the region.
//         */
//        if(iter->is_new_region)
//        {
//            /* Find the start of this region in the index */
//            error = _search_index_for_region_start(iter);
//            CHECK_BIO(error);
//
//            if(-1 == iter->current_index)
//            {
//                /* Nothing in this region */
//                iter->current_region += 1;
//            }
//            else
//            {
//                iter->is_new_region = 0;
//            }
//        }
//        else
//        {
//            /* Not a new region
//             * 
//             * Perform a linear scan of index to get next matching alignment.
//             */
//            error = _scan_index_for_next_alignment(iter, index, data);
//            CHECK_BIO(error);
//
//            if(NULL == data)
//            {
//                /* Nothing more in this region */
//                iter->current_region += 1;
//            }
//       }
//    } /* end while */
//
//    /**SUCCESS***************************************************************/
//
//    return BIOHDF_BADNESS;
//
//    /**FAILURE***************************************************************/
//
//    FAIL:
//
//    if(iter && iter->alignments) {
//        _print_error_info(iter->alignments->path);
//    }
//    else {
//        _print_error_info(NULL);
//    }
//
//    BIOH5Gfree_alignment_data(data);
//    *index = -1;
//    *data = NULL;
//
//    return BIOHDF_BADNESS;
//}



BIOHDF_API biohdf_error
BIOH5Gget_alignment(const bioh5g_alignments aligns,
                    int64_t index,
                    /*OUT*/ bioh5g_alignment_data **data)
{
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *data  = NULL;

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    /* Set up output data structure */
    *data = (bioh5g_alignment_data *)calloc(1, sizeof(bioh5g_alignment_data));
    CHECK_PTR(*data);

    /* Read the basic alignment data */
    error = _fill_basic_align_data(aligns, index, *data);
    CHECK_BIO(error);

    /* Fill the rest of the alignment data */
    error = _fill_remaining_align_data(aligns, index, *data);
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    BIOH5Gfree_alignment_data(data);
    *data = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gfree_alignment_data(/*IN-OUT*/ bioh5g_alignment_data **data)
{
    /* There is very little error checking that this function could
     * reasonably perform.
     */

    if(*data != NULL)
    {
        free((*data)->reference);
        free((*data)->sam_cigar);
        free((*data)->sam_tags);
        free((*data)->sam_rnext);

        free(*data);
        (*data) = NULL;
    }

    return BIOHDF_NO_ERROR;
}



/*****************************************************************************
 * Alignments Indexing                                                       *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_index(bioh5g_alignments aligns,
                              bioh5g_alignments_index_method method,
                              biohdf_index_creation_properties props)
{
    /* Single array flag
     * When there is only one data array, there's no need to
     * write temp files
     */
    int     no_temp_files = 1;

    size_t  i;      /* iterator over array elements in special case */

    /* Property data */
    int64_t         max_memory_bytes;
    char            *temp_file_path;

    /* # elements in the collection */
    int64_t         count = -1;

    /* Reference / Position data array */
    ref_pos_data  *rp_data    = NULL;   /* Array of reference and position data */
    size_t        rp_n        = 0;      /* # elements in the ref/pos array */
    char          *rp_front   = NULL;   /* Pointer to where the basic data will be written */
    char          *rp_back    = NULL;   /* Pointer to where the char data ends */
    
    /* Alignment iterator and data */
    bioh5g_alignments_iterator   iter          = NULL;   /* alignments iterator */
    bioh5g_alignment_data        *alignment    = NULL;   /* alignment data */
    int64_t                      align_index   = 0;      /* alignment index (not used) */

    /* Temp file creation */
    char     *filename = NULL;  /* name of temp file to create or open */
    int      n_files   = 0;     /* # of temp files written */
    FILE     **files   = NULL;  /* array of temp files for k-way merge */
    int      n_closed  = 0;     /* # of files closed */

    /* k-way merge of temp files */
    int k       = -1;   /* file pointer and data iterator for k-way merge */
    int min_k   = -1;   /* index where min value is found */

    /* BioHDF index vector */
    vector *index = NULL;

    /* Attribute writing */
    hid_t   type_id = -1;
    hid_t   attr_id = -1;

    int     bad_close   = EOF;   /* return value of fclose() */


    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    /* Nothing to do if an empty collection */
    /* This shortcut exists because it turned out that, as of v0.4, there
     * is a significant amount of time devoted to indexing a zero element
     * collection in the alignments test.
     */
    error = BIOH5Gget_alignments_count(aligns, &count);
    CHECK_BIO(error);
    CHECK_INT_GE(count, 0);
    if(0 == count)
    {
        return BIOHDF_NO_ERROR;
    }

    /* Set parameters from properties or defaults */
    if(NULL == props)
    {
        max_memory_bytes = DEFAULT_INDEX_MAX_MEMORY_TO_USE_MB * 1024 * 1024;
        temp_file_path   = DEFAULT_INDEX_TEMP_FILE_PATH;
    }
    else
    {
        max_memory_bytes = props->max_memory_to_use_mb * 1024 * 1024;
        temp_file_path   = props->temp_file_path;
    }

    /* Get an alignments iterator */
    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);

    /* Create the temp array for storing data to be sorted */

    /* NOTE: This will allocate a LOT of memory, even when just a few
       alignments are being sorted.  Might want to start smaller and
       realloc our way up to the max size as needed. */
    rp_data = (ref_pos_data *)malloc((size_t)max_memory_bytes);
    CHECK_PTR(rp_data);


    /************************************************************************
     * Create temp files of sorted data from the alignments collection data *
     ************************************************************************/


    /* While there are sill alignments to process, fill the rps array from the collection */

    no_temp_files = 0;
    n_files = 0;
    rp_n = 0;
    rp_front = (char *)rp_data;
    rp_back  = (char *)(rp_front + max_memory_bytes); /* points to 1 byte AFTER the end of the array */

    error = BIOH5Gget_next_alignment(iter, &align_index, &alignment);
    CHECK_BIO(error);
    while(BIOHDF_NO_ERROR == error)
    {
        /* If pointers overlap or if the collection is empty,
         * sort the data array and store it on the disk.
         */
        if(NULL == alignment && 0 == n_files)
        {
            /* Special case where there is only enough data for one file */

            /* Sort the rps array */
            qsort(rp_data, rp_n, sizeof(ref_pos_data), _ref_pos_comparator);

            /* Set the 'no temp files written' flag */
            no_temp_files = 1;
            
        }
        else if(NULL == alignment
            || rp_front + sizeof(ref_pos_data) >= rp_back - (strlen(alignment->reference) + 1))
        {
            /* Normal, 'array is full' or end of data condition */

            /* Sort the rps array */
            qsort(rp_data, rp_n, sizeof(ref_pos_data), _ref_pos_comparator);

            /* Propagate the data to the disk */
            error = _get_temp_file_name(temp_file_path, "REF_POS_INDEX_DATA_", n_files, &filename);
            CHECK_BIO(error);

            error = _write_index_temp_file(rp_data, rp_n, filename);
            CHECK_BIO(error);
            
            free(filename);
            filename = NULL;

            /* reset data */
            rp_n = 0;
            rp_front = (char *)rp_data;
            rp_back  = (char *)(rp_front + max_memory_bytes);

            /* Add 1 to the number of files written */
            n_files++;
        }

        /* If the collection is empty, break.  If not, write the data to the array */
        if(alignment == NULL)
        {
            break;
        }
        else
        {
            /* Increment back pointer and copy string data there */
            rp_back  -= (strlen(alignment->reference) + 1);
            strcpy(rp_back, alignment->reference);
            
            /* Copy data to the array and increment the front pointer */
            rp_data[rp_n].original_index  = align_index;
            rp_data[rp_n].position        = alignment->position;
            rp_data[rp_n].reference       = rp_back;

            rp_front += sizeof(ref_pos_data);

            /* For next entry */
            rp_n++;
        }

        error = BIOH5Gfree_alignment_data(&alignment);
        CHECK_BIO(error);
        error = BIOH5Gget_next_alignment(iter, &align_index, &alignment);
        CHECK_BIO(error);
    }

    /***************************
     * Create the index vector *
     ***************************/

    /* Check to see if the alignments have already been indexed.
     * If so, delete the existing index (VERY inefficient - needs to be repacked)
     *
     * TODO: consider re-using the existing index space.
     */
    if(H5Lexists(aligns->group, BIOH5G_REF_POS_INDEX_DSET_NAME, H5P_DEFAULT))
    {
        hdf_error = H5Ldelete(aligns->group, BIOH5G_REF_POS_INDEX_DSET_NAME, H5P_DEFAULT);
        CHECK_HERR_T(hdf_error);
    }

    /* Create secondary index vector in the same group as the alignments */
    error = _create_vector(aligns->group, 
        BIOH5G_REF_POS_INDEX_NAME,
        H5T_STD_I64LE,
        FIXED_SIZE_ARRAY,
        1, 
        BIOHDF_DEFAULT_CHUNK_SIZE,
        BIOHDF_DEFAULT_COMPRESSION_LEVEL,
        &index);
    CHECK_BIO(error);


    /*************************
     * Fill the index vector *
     *************************/

    if(1 == no_temp_files)
    {
        for(i = 0; i < rp_n; i++)
        {
            error = _add_vector_element(index, &(rp_data[i].original_index), 1);
            CHECK_BIO(error);
        }
    }
    else
    {
        /* Create the file pointer array */
        files = (FILE **)calloc(n_files, sizeof(FILE *));
        CHECK_PTR(files);

        for(k = 0; k < n_files; k++)
        {
            /* Get filename */
            error = _get_temp_file_name(temp_file_path, "REF_POS_INDEX_DATA_", k, &filename);
            CHECK_BIO(error);

            /* Open that particular file */
            files[k] = fopen(filename, "rb");
            CHECK_PTR(files[k]);

            free(filename);
            filename = NULL;
        }

        /* Initialize the file data array */
        rp_data = (ref_pos_data *)realloc(rp_data, n_files * sizeof(ref_pos_data));
        CHECK_PTR(rp_data);

        for(k = 0; k < n_files; k++)
        {
            error = _read_index_data(files[k], &(rp_data[k]));
            CHECK_BIO(error);
        }

        /* Fill the index vector */
        n_closed = 0;
        while(n_closed < n_files)
        {
            /* Find first non-NULL files entry and declare that index min_d */
            for(k = 0; k < n_files; k++)
            {
                if(NULL != files[k])
                {
                    min_k = k;
                    break;
                }
            }

            for(k = min_k + 1; k < n_files; k++)
            {
                if(NULL != files[k]
                    && rp_data[k].position < rp_data[min_k].position
                    && strcmp(rp_data[k].reference, rp_data[min_k].reference) < 0)
                {
                    /* Reference and position are < */
                    /* If this rp_data != NULL && is less than min_d, it's the new min_d */
                    min_k = k;
                }
            }

            /* Write min_d's index into the index vector */
            error = _add_vector_element(index, &(rp_data[min_k].original_index), 1);
            CHECK_BIO(error);

            /* Attempt to read another element from min_d's file */
            free(rp_data[min_k].reference);
            rp_data[min_k].reference = NULL;
            error = _read_index_data(files[min_k], &(rp_data[min_k]));
            CHECK_BIO(error);
            
            /* If that doesn't work, close the file pointer and make it NULL */
            if(rp_data[min_k].original_index < 0)
            {
                bad_close = fclose(files[min_k]);
                if(EOF == bad_close)
                {
                    BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
                    goto FAIL;
                }
                files[min_k] = NULL;
                n_closed++;
            }
        }

        /* Remove temp files */
        for(k = 0; k < n_files; k++)
        {
            /* Get filename */
            error = _get_temp_file_name(temp_file_path, "REF_POS_INDEX_DATA_", k, &filename);
            CHECK_BIO(error);

            /* Remove that particular file */
            bad_close = remove(filename);
            CHECK_INT_NE(bad_close, 0);

            free(filename);
            filename = NULL;
        }
    }

    /****************************
     * Write the sort attribute *
     ****************************/

    error = _create_index_enum_HDF5_type(&type_id);
    CHECK_BIO(error);

    aligns->index_method = REF_POS_SECONDARY;

    attr_id = H5Aopen(aligns->group, BIOH5G_INDEX_METHOD_ATTR, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    hdf_error = H5Awrite(attr_id, type_id, &(aligns->index_method));
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Tclose(type_id);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);

    error = _close_vector(&index);
    CHECK_BIO(error);

    free(rp_data);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    BIOH5Gdestroy_alignments_iterator(&iter);
    free(rp_data);
    free(filename);

    if(files != NULL)
    {
        for(k = 0; k < n_files; k++)
        {
            if(files[k] != NULL)
            {
                fclose(files[k]);

                /* Leave the files for debugging! */
            }
        }
    }

    return BIOHDF_BADNESS;
}



static biohdf_error
_read_index_data(FILE *f, /*OUT*/ ref_pos_data *data)
{
    size_t record_size      = 0;     /* Size of this record, in bytes */
    size_t len              = 0;     /* Length of the reference string */

    size_t n_items_read     = 0;     /* # of items read by fread() */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(f);
    CHECK_PTR(data);

    /**CODE*****************************************************************/

    /* Read record size from file */
    n_items_read = fread(&record_size, sizeof(size_t), 1, f);
    if(n_items_read != 1)
    {
        if(feof(f))
        {
            data->original_index = -1;
            data->position = -1;
            data->reference = NULL;
            return BIOHDF_NO_ERROR;
        }
        else
        {
            BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
            goto FAIL;
        }
        
    }

    /* Read original index from file */
    n_items_read = fread(&(data->original_index), sizeof(int64_t), 1, f);
    CHECK_INT_EQ(n_items_read, 1);

    /* Read position from file */
    n_items_read = fread(&(data->position), sizeof(int32_t), 1, f);
    CHECK_INT_EQ(n_items_read, 1);

    /* Calculate string size (+1 for nul) and allocate memory */
    len = record_size - sizeof(size_t) - sizeof(int32_t) - sizeof(int64_t);
    CHECK_INT_GT(len, -1);

    data->reference = (char *)calloc(len + 1 /* nul */, sizeof(char));
    CHECK_PTR(data->reference);

    /* Read characters */
    n_items_read = fread(data->reference, sizeof(char), len, f);
    CHECK_INT_NE(n_items_read, len);

    data->reference[len] = '\0';

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    data->original_index = -1;
    data->position = -1;
    data->reference = NULL;

    return BIOHDF_BADNESS;
}



static biohdf_error
_get_temp_file_name(char *path, char *prefix, int n, /*OUT*/ char **filename)
{
    char     *n_str      = NULL;

    size_t   filename_len = 0;      /* Length of the filename */
    size_t   fn_len       = 0;      /* Length of the "file #" string */

    /**CHECK*PARAMETERS*****************************************************/

    *filename = NULL;

    CHECK_PTR(path);
    CHECK_PTR(prefix);
    CHECK_INT_GT(n, -1);

    /**CODE*****************************************************************/

    fn_len = 20; /* max # of digits in a 64-bit int */

    n_str  = (char *)calloc(fn_len, sizeof(char));
    CHECK_PTR(n_str);
    sprintf(n_str, "%d", n);

    filename_len = strlen(path)
        + strlen(prefix)
        + fn_len
        + 4          /* ".tmp" */
        + 1;         /* \0 */

    *filename = (char *)calloc(filename_len, sizeof(char));
    CHECK_PTR(*filename);

    strcpy(*filename, path);
    strcat(*filename, prefix);
    strcat(*filename, n_str);
    strcat(*filename, ".tmp");

    /**SUCCESS***************************************************************/

    free(n_str);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    free(n_str);

    free(*filename);
    *filename = NULL;

    return BIOHDF_BADNESS;
}



static biohdf_error
_write_index_temp_file(ref_pos_data *data,
                       size_t n,
                       char *filename)
{
    size_t   i            = 0;      /* iterator */
    
    size_t   entry_size   = 0;      /* Size of the entry in bytes */
    size_t   ref_len      = 0;      /* Length of the reference string */
    
    FILE     *f           = NULL;   /* File pointer */

    size_t   n_items_written  = 0;     /* # of items written by fwrite() */
    int      bad_close        = EOF;   /* return value of fclose() */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(data);
    CHECK_PTR(filename);

    /**CODE*****************************************************************/

    /* Open file */
    f = fopen(filename, "w+b");
    CHECK_PTR(f);

    /* Write entries to the disk */
    for(i = 0; i < n; i++)
    {
        /* Calculate size of entry */
        ref_len = strlen(data[i].reference);  /* ignore trailing nul */

        entry_size = sizeof(entry_size)
            + sizeof(data[i].original_index)
            + sizeof(data[i].position)
            + ref_len;

        /* Write size of entry */
        n_items_written = fwrite(&entry_size, sizeof(size_t), 1, f);
        CHECK_INT_EQ(n_items_written, 1);

        /* Write index and position */
        n_items_written = fwrite(&(data[i].original_index), sizeof(int64_t), 1, f);
        CHECK_INT_EQ(n_items_written, 1);
        n_items_written = fwrite(&(data[i].position),       sizeof(int32_t), 1, f);
        CHECK_INT_EQ(n_items_written, 1);

        /* Write string */
        n_items_written = fwrite(data[i].reference, sizeof(char), ref_len, f);
        CHECK_INT_EQ(n_items_written, ref_len);
    }

    /**SUCCESS***************************************************************/

    bad_close = fclose(f);
    if(EOF == bad_close)
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(filename);

    if(NULL != f)
    {
        fclose(f);
    }

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Alignment hit file header storage                                         *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gstore_alignment_file_header(const bioh5g_alignments aligns,
                                  bioh5g_alignments_format format,
                                  const char *header)
{
    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    /* TODO: store the format in which the header was stored */
    /* TODO: make it obvious that only one header can be stored per collection */

    /* Create and store header (w/o terminal nul) */
    error = _add_vector_element(aligns->sam_header,
        header,
        strlen(header));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_file_header(const bioh5g_alignments aligns,
                                /*OUT*/ bioh5g_alignments_format *format,  /* NOT USED AT THIS TIME */
                                /*OUT*/ char **header)
{
    hsize_t     size = 0;
    int         is_c_string = 1;

    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *header = NULL;

    CHECK_PTR(aligns);

    /**CODE*****************************************************************/

    error = _get_vector_count(aligns->sam_header, &size);

    if(0 != size)
    {
        error = _get_vector_element(aligns->sam_header, 
            0, /* The header is the only thing in the vector, so index 0 */
            is_c_string, 
            (void **)header);
        CHECK_BIO(error);
    }
    else
    {
        /* Return empty string on no header */
        /* TODO: Might want to consider returning NULL here */
        *header = (char *)calloc(1, sizeof(char));
        CHECK_PTR(*header);

        strcpy(*header, "\0");
    }

    /* TODO - cheating! */
    *format = SAM_FORMAT;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(aligns) {
        _print_error_info(aligns->path);
    }
    else {
        _print_error_info(NULL);
    }

    free(*header);
    *header = NULL;

    return BIOHDF_BADNESS;
}


/*****************************************************************************
 * Data formats                                                              *
 *****************************************************************************/

static int64_t FAKE_RNAME_g = 0;

BIOHDF_API biohdf_error
BIOH5Gcreate_alignment_string(const bioh5g_alignment_data *alignment,
                              const bioh5g_read_data *read,
                              bioh5g_alignments_format format,
                              /*OUT*/ char **alignment_string)
{
    if(SAM_FORMAT == format)
    {
    }
    else
    {
        return BIOHDF_BADNESS;
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gwrite_alignment_to_stream(const bioh5g_alignment_data *alignment,
                                const bioh5g_read_data *read,
                                bioh5g_alignments_format format,
                                FILE *stream)
{
    int printf_error    = -1;   /* fprintf error value */

    /**CHECK*PARAMETERS*****************************************************/

    /* NOTE: read CAN be NULL! */
    CHECK_PTR(alignment);
    CHECK_PTR(stream);

    /**CODE*****************************************************************/

    if(SAM_FORMAT == format)
    {
        /* QNAME */
        if(NULL == read || 0 == strlen(read->identifier))
        {
            /* Arbitrary integers are used as "unique" names when no
             * read data is supplied.  This will obviously have
             * parallel issues
             */
            printf_error = fprintf(stream, "%ld", (long int)(FAKE_RNAME_g++));
        }
        else
        {
            printf_error = fprintf(stream, "%s", read->identifier);
        }
        CHECK_INT_GT(printf_error, -1);

        /* FLAG */
        printf_error = fprintf(stream, "\t%d", (int)(alignment->sam_flags));
        CHECK_INT_GT(printf_error, -1);


        /* RNAME */
        if(0 == strlen(alignment->reference))
        {
            printf_error = fprintf(stream, "\t*");
        }
        else
        {
            printf_error = fprintf(stream, "\t%s", alignment->reference);
        }
        CHECK_INT_GT(printf_error, -1);


        /* POS */
        printf_error = fprintf(stream, "\t%d", (int)(alignment->position));
        CHECK_INT_GT(printf_error, -1);


        /* MAPQ */
        printf_error = fprintf(stream, "\t%d", (int)(alignment->sam_mapq));
        CHECK_INT_GT(printf_error, -1);


        /* CIGAR */
        if(0 == strlen(alignment->sam_cigar))
        {
            printf_error = fprintf(stream, "\t*");
        }
        else
        {
            printf_error = fprintf(stream, "\t%s", alignment->sam_cigar);
        }
        CHECK_INT_GT(printf_error, -1);


        /* RNEXT */
        if(0 == strlen(alignment->sam_rnext))
        {
            printf_error = fprintf(stream, "\t*");
        }
        else
        {
            printf_error = fprintf(stream, "\t%s", alignment->sam_rnext);
        }
        CHECK_INT_GT(printf_error, -1);


        /* PNEXT */
        printf_error = fprintf(stream, "\t%d", (int)(alignment->sam_pnext));
        CHECK_INT_GT(printf_error, -1);


        /* TLEN */
        printf_error = fprintf(stream, "\t%d", (int)(alignment->sam_tlen));
        CHECK_INT_GT(printf_error, -1);


        /* SEQ */
        if(NULL == read || 0 == strlen(read->sequence))
        {
            printf_error = fprintf(stream, "\t*");
        }
        else
        {
            printf_error = fprintf(stream, "\t%s", read->sequence);
        }
        CHECK_INT_GT(printf_error, -1);

        
        /* QUAL */
        if(NULL == read || 0 == strlen(read->quality_values))
        {
            printf_error = fprintf(stream, "\t*");
        }
        else
        {
            printf_error = fprintf(stream, "\t%s", read->quality_values);
        }
        CHECK_INT_GT(printf_error, -1);


        /* Tags are special - there is no ouput on an empty tag string,
         * unlike the other fields, which print '*' for an empty string.
         */
        if(0 != strlen(alignment->sam_tags))
        {
            printf_error = fprintf(stream, "\t%s", alignment->sam_tags);
        }
        CHECK_INT_GT(printf_error, -1);


        printf_error = fprintf(stream, "\n");
        CHECK_INT_GT(printf_error, -1);
    }
    else
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Internal functions                                                        *
 *****************************************************************************/



static int
_ref_pos_comparator(const void *elem1, const void *elem2)
{
    int r = 0;   /* return value */

    int64_t pos1;
    int64_t pos2;

    char *ref1;
    char *ref2;

    /* Set values */
    /* TODO: get rid of the aliases if there is a performance hit
     * If we create a references dictionary, this will all get rewritten
     * anyway.
     */
    pos1 = ((ref_pos_data *)elem1)->position;
    pos2 = ((ref_pos_data *)elem2)->position;
    ref1 = ((ref_pos_data *)elem1)->reference;
    ref2 = ((ref_pos_data *)elem2)->reference;

    /* compare reference */
    r = strcmp(ref1, ref2);

    /* compare position */
    if(0 == r)
    {
        if(pos1 < pos2)  { r = -1; }
        if(pos1 > pos2)  { r = 1; }
    }

    return r;
}



static biohdf_error
_create_alignment_HDF5_type(/*OUT*/ hid_t *tid)
{
    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* n/a */

    /**CODE*****************************************************************/

    *tid = -1;

    *tid = H5Tcreate(H5T_COMPOUND, sizeof(basic_alignment_data));
    CHECK_HID_T(*tid);

    hdf_error = H5Tinsert(*tid, "read_index", HOFFSET(basic_alignment_data, read_index), H5T_STD_I64LE);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Tinsert(*tid, "position", HOFFSET(basic_alignment_data, position), H5T_STD_I32LE);
    CHECK_HERR_T(hdf_error);
    
    hdf_error = H5Tinsert(*tid, "length", HOFFSET(basic_alignment_data, length), H5T_STD_I32LE);
    CHECK_HERR_T(hdf_error);
    
    hdf_error = H5Tinsert(*tid, "sam_mapq", HOFFSET(basic_alignment_data, sam_mapq), H5T_STD_U8LE);
    CHECK_HERR_T(hdf_error);
    
    hdf_error = H5Tinsert(*tid, "sam_flags", HOFFSET(basic_alignment_data, sam_flags),  H5T_STD_U32LE);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*tid > 0)
    {
        H5Tclose(*tid);
    }

    *tid = -1;

    return BIOHDF_BADNESS;
}



static biohdf_error
_create_template_HDF5_type(/*OUT*/ hid_t *tid)
{
    herr_t  hdf_error   = -1;   /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* n/a */

    /**CODE*****************************************************************/

    *tid = -1;
    
    *tid = H5Tcreate(H5T_COMPOUND, sizeof(template_data));
    CHECK_HID_T(*tid);

    hdf_error = H5Tinsert(*tid, "sam_pnext",    HOFFSET(template_data, sam_pnext),    H5T_STD_I32LE);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Tinsert(*tid, "sam_tlen",     HOFFSET(template_data, sam_tlen),     H5T_STD_I32LE);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*tid > 0)
    {
        H5Tclose(*tid);
    }

    *tid = -1;

    return BIOHDF_BADNESS;
}



static biohdf_error
_create_index_enum_HDF5_type(/*OUT*/ hid_t *tid)
{
    int index_type;

    herr_t  hdf_error   = -1;   /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* n/a */

    /**CODE*****************************************************************/

    *tid = -1;

    *tid = H5Tenum_create(H5T_NATIVE_INT);
    CHECK_HID_T(*tid);

    index_type = UNINDEXED;
    hdf_error = H5Tenum_insert(*tid, "UNINDEXED", &index_type);
    CHECK_HERR_T(hdf_error);

    index_type = REF_POS_SECONDARY;
    hdf_error = H5Tenum_insert(*tid, "REF_POS_SECONDARY", &index_type);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*tid > 0)
    {
        H5Tclose(*tid);
    }

    *tid = -1;

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Accessor functions (needed for higher-language interoperation)            *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_alignment_data(/*OUT*/ bioh5g_alignment_data **data)
{
    (*data) = (bioh5g_alignment_data *)calloc(1, sizeof(bioh5g_alignment_data));
    (*data)->read_index  = 0;
    (*data)->reference   = NULL;
    (*data)->position    = 0;
    (*data)->length      = 0;
    (*data)->sam_mapq    = 0;
    (*data)->sam_flags   = 0;
    (*data)->sam_cigar   = NULL;
    (*data)->sam_tags    = NULL;
    (*data)->sam_rnext   = NULL;
    (*data)->sam_pnext   = 0;
    (*data)->sam_tlen    = 0;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_read_index(bioh5g_alignment_data *data,
                               /*OUT*/ int64_t *read_index)
{
    *read_index = data->read_index;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_read_index(bioh5g_alignment_data *data,
                               int64_t read_index)
{
    data->read_index = read_index;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_reference(bioh5g_alignment_data *data,
                              /*OUT*/ char **reference)
{
    (*reference) = (char *)calloc(strlen(data->reference) + 1, sizeof(char));
    strcpy(*reference, data->reference);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_reference(bioh5g_alignment_data *data,
                              char *reference)
{
    if(NULL != data->reference)
    { 
        free(data->reference);
    }
    data->reference = (char *)calloc(strlen(reference) + 1, sizeof(char));
    strcpy(data->reference, reference);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_position(bioh5g_alignment_data *data,
                             /*OUT*/ int32_t *position)
{
    *position = data->position;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_position(bioh5g_alignment_data *data,
                             int32_t position)
{
    data->position = position;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_length(bioh5g_alignment_data *data,
                           /*OUT*/ int32_t *length)
{
    *length = data->length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_length(bioh5g_alignment_data *data,
                           int32_t length)
{
    data->length = length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_mapq(bioh5g_alignment_data *data,
                             /*OUT*/ unsigned char *sam_mapq)
{
    *sam_mapq = data->sam_mapq;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_mapq(bioh5g_alignment_data *data,
                             unsigned char sam_mapq)
{
    data->sam_mapq = sam_mapq;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_flags(bioh5g_alignment_data *data,
                              /*OUT*/ uint32_t *sam_flags)
{
    *sam_flags = data->sam_flags;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_flags(bioh5g_alignment_data *data,
                              uint32_t sam_flags)
{
    data->sam_flags = sam_flags;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_cigar(bioh5g_alignment_data *data,
                              /*OUT*/ char **sam_cigar)
{
    (*sam_cigar) = (char *)calloc(strlen(data->sam_cigar) + 1, sizeof(char));
    strcpy(*sam_cigar, data->sam_cigar);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_cigar(bioh5g_alignment_data *data,
                              char *sam_cigar)
{
    if(NULL != data->sam_cigar)
    { 
        free(data->sam_cigar);
    }
    data->sam_cigar = (char *)calloc(strlen(sam_cigar) + 1, sizeof(char));
    strcpy(data->sam_cigar, sam_cigar);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_tags(bioh5g_alignment_data *data,
                             /*OUT*/ char **sam_tags)
{
    (*sam_tags) = (char *)calloc(strlen(data->sam_tags) + 1, sizeof(char));
    strcpy(*sam_tags, data->sam_tags);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_tags(bioh5g_alignment_data *data,
                             char *sam_tags)
{
    if(NULL != data->sam_tags)
    { 
        free(data->sam_tags);
    }
    data->sam_tags = (char *)calloc(strlen(sam_tags) + 1, sizeof(char));
    strcpy(data->sam_tags, sam_tags);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_rnext(bioh5g_alignment_data *data,
                              /*OUT*/ char **sam_rnext)
{
    (*sam_rnext) = (char *)calloc(strlen(data->sam_rnext) + 1, sizeof(char));
    strcpy(*sam_rnext, data->sam_rnext);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_rnext(bioh5g_alignment_data *data,
                              char *sam_rnext)
{
    if(NULL != data->sam_rnext)
    { 
        free(data->sam_rnext);
    }
    data->sam_rnext = (char *)calloc(strlen(sam_rnext) + 1, sizeof(char));
    strcpy(data->sam_rnext, sam_rnext);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_pnext(bioh5g_alignment_data *data,
                              /*OUT*/ int32_t *sam_pnext)
{
    *sam_pnext = data->sam_pnext;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_pnext(bioh5g_alignment_data *data,
                              int32_t sam_pnext)
{
    data->sam_pnext = sam_pnext;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_tlen(bioh5g_alignment_data *data,
                             /*OUT*/ int32_t *sam_tlen)
{
    *sam_tlen = data->sam_tlen;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_tlen(bioh5g_alignment_data *data,
                             int32_t sam_tlen)
{
    data->sam_tlen = sam_tlen;
    return BIOHDF_NO_ERROR;
}



/*****************************************************************************
 * Alignments properties - create, destroy, access                           *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_properties(/*OUT*/ bioh5g_alignments_creation_properties *props)
{
    *props = (bioh5g_alignments_creation_properties)calloc(1, sizeof(struct _bioh5g_alignments_creation_properties));

    (*props)->reads_path     = (char *)calloc(1, sizeof(char));  /* empty string */

    (*props)->chunk_size     = BIOHDF_DEFAULT_CHUNK_SIZE;
    (*props)->compression    = BIOHDF_DEFAULT_COMPRESSION_LEVEL;
    (*props)->refs_scheme    = VARIABLE_LENGTH_STRING;
    (*props)->tags_scheme    = VARIABLE_LENGTH_STRING;
    (*props)->cigar_scheme   = VARIABLE_LENGTH_STRING;
    (*props)->refs_length    = 0;
    (*props)->tags_length    = 0;
    (*props)->cigar_length   = 0;

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gdestroy_alignments_properties(/*OUT*/ bioh5g_alignments_creation_properties *props)
{
    if(*props)
    {
        if((*props)->reads_path)
        {
            free((*props)->reads_path);
        }

        free(*props);
        (*props) = NULL;
    }

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_reads_path(bioh5g_alignments_creation_properties props,
                                           char *reads_path)
{
    if(NULL != props->reads_path)
    {
        free(props->reads_path);
    }

    if(NULL == reads_path)
    {
        props->reads_path = NULL;
    }
    else
    {
        props->reads_path = (char *)calloc(strlen(reads_path) + 1, sizeof(char));
        strcpy(props->reads_path, reads_path);
    }

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_refs_scheme(bioh5g_alignments_creation_properties props,
                                            biohdf_string_storage_scheme scheme)
{
    props->refs_scheme = scheme;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_tags_scheme(bioh5g_alignments_creation_properties props,
                                                 biohdf_string_storage_scheme scheme)
{
    props->tags_scheme = scheme;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_cigar_scheme(bioh5g_alignments_creation_properties props,
                                             biohdf_string_storage_scheme scheme)
{
    props->cigar_scheme = scheme;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_refs_length(bioh5g_alignments_creation_properties props,
                                            size_t length)
{
    props->refs_length = length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_tags_length(bioh5g_alignments_creation_properties props,
                                            size_t length)
{
    props->tags_length = length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_cigar_length(bioh5g_alignments_creation_properties props,
                                             size_t length)
{
    props->cigar_length = length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_chunk_size(bioh5g_alignments_creation_properties props,
                                           int64_t chunk_size)
{
    props->chunk_size = chunk_size;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_compression_level(bioh5g_alignments_creation_properties props,
                                                  compression_level level)
{
    props->compression = level;
    return BIOHDF_NO_ERROR;
}

