/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/
 
/** @defgroup bioh5g_alignments bioh5g_alignments
 *
 *  Represents NGS alignments (SAM, etc. entries).
 *
 *  @{
 */


#ifndef _BIOH5G_ALIGNMENTS_H
#define _BIOH5G_ALIGNMENTS_H

#include <stdio.h>

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_file.h"
#include "biohdf_utility.h"

#include "bioh5g_api.h"
#include "bioh5g_reads.h"



/*****************************************************************************
 * Attribute Names                                                           *
 *****************************************************************************/



/** Attribute name: BioHDF path to the associated reads */
#define BIOH5G_READS_PATH_ATTR "READS_PATH"



/** Attribute name: Describes how the alignments are indexed */
#define BIOH5G_INDEX_METHOD_ATTR "INDEX_METHOD"



/*****************************************************************************
 * Type definitions                                                          *
 *****************************************************************************/



/** BioHDF alignments collection handle */
typedef struct _bioh5g_alignments *bioh5g_alignments;



/** BioHDF alignments creation properties */
typedef struct _bioh5g_alignments_creation_properties *bioh5g_alignments_creation_properties;



/** BioHDF alignments iterator handle */
typedef struct _bioh5g_alignments_iterator *bioh5g_alignments_iterator;



/*****************************************************************************
 * Structs and enums                                                         *
 *****************************************************************************/



/** The text file format for alignment I/O */
typedef enum
{
    SAM_FORMAT      /**< SAM format */
} bioh5g_alignments_format;



/** The indexing scheme used on the alignments in this collection */
typedef enum
{
    UNINDEXED = 0,       /**< No indexing, all queries are linear scans */
    REF_POS_SECONDARY = 1  /**< Sorted on the reference (strcmp order) and start position.  Secondary index (no physical data ordering) */

    /* REF_POS_PRIMARY = 2, */
    /* REF_POS_NCLIST_PRIMARY = 3, */
    /* REF_POS_NCLIST_PRIMARY = 4, */

} bioh5g_alignments_index_method;



/*****************************************************************************
 * Data container                                                            *
 *****************************************************************************/



/** Alignment data container
 *
 * Currently very SAM-centric but will be made more generic in the future.
 */
typedef struct
{
    /* basic alignment data */
    int64_t   read_index;   /**< The HDF5 index/offset for the associated read */
    char      *reference;   /**< The reference name */
    int32_t   position;     /**< The position on the reference (left endpoint, 1-based inclusive) */
    int32_t   length;       /**< The length of the alignment */

    /* SAM data */
    unsigned char  sam_mapq;     /**< SAM quality (MAPQ) value */
    int32_t        sam_flags;    /**< The SAM FLAGS */
    char           *sam_cigar;   /**< The SAM CIGAR string */
    char           *sam_tags;    /**< The SAM tag string */

    /* SAM template data */
    char      *sam_rnext;   /**< The reference of the next alignment in the template */
    int32_t   sam_pnext;    /**< The position of the next alignment in the template */
    int32_t   sam_tlen;     /**< The template length */

} bioh5g_alignment_data;



/*****************************************************************************
 * Create, open, close                                                       *
 *****************************************************************************/



/** Test if an alignments collection exists
 *
 * This function will return TRUE if a collection of the same type exists
 * at the named location.  If any other HDF5 or BioHDF object with that
 * same name exists, TRUE will be returned as well as an error code, the
 * assumption being that Bad Code(tm) that does not check return values
 * will be more likely to attempt to open code (and fail), rather than
 * create things (which my partially succeed, making a mess).
 *
 * @param file A BioHDF file handle
 * @param path The BioHDF path to the collection
 * @param[out] presence TRUE if the collection exists, FALSE if it does not.
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcheck_alignments_presence(const biohdf_file file,
                                const char *path,
                                /*OUT*/ int *presence);



/** Create (and open) a new alignments collection
 *
 *  The collection handle returned by this function will be ready
 *  to accept I/O.
 *
 * @param file A BioHDF file handle
 * @param properties Collection creation properties
 * @param path The BioHDF path to the collection
 * @param[out] aligns A BioHDF alignments handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gcreate_alignments_collection(const biohdf_file file,
                                   const bioh5g_alignments_creation_properties props,
                                   const char *path,
                                   /*OUT*/ bioh5g_alignments *aligns);



/** Open an existing alignments collection
 *
 * @param file A BioHDF file handle
 * @param path The BioHDF path to the collection
 * @param mode The access mode (read-only | read-write)
 * @param[out] alignments A BioHDF alignments handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gopen_alignments_collection(const biohdf_file file,
                                 const char *path,
                                 biohdf_open_mode mode,
                                 /*OUT*/ bioh5g_alignments *aligns);



/** Close an open alignments collection
 *
 *  This function will set the collection handle to NULL after freeing it.
 *
 * @param[in,out] alignments A BioHDF alignments handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gclose_alignments_collection(/*IN-OUT*/ bioh5g_alignments *aligns);



/** Get the path to the associated reads.
 *
 * Each BioHDF alignments collection includes a link the associated
 * reads collection.
 *
 * An empty string ("\0") is returned if the link is not present.
 *
 * @param alignments A BioHDF alignments handle
 * @param[out] reads_path The BioHDF path to the associated reads
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gget_reads_path(const bioh5g_alignments aligns,
                     /*OUT*/ char **reads_path);



/*****************************************************************************
 * Append and read data                                                      *
 *****************************************************************************/



/** Get the number of stored alignments in a collection
 *
 * @param aligns A BioHDF alignments handle
 * @param[out] count The number of alignments in the collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_alignments_count(const bioh5g_alignments aligns,
                           /*OUT*/ int64_t *count);



/** Create an iterator for an alignments collection
 *
 * @param aligns A BioHDF alignments handle
 * @param[out] iter An iterator for an alignments collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_iterator(const bioh5g_alignments aligns,
                                 /*OUT*/ bioh5g_alignments_iterator *iter);


/** Add a reference region filter to an alignments iterator
 *
 * Add reference region filters one at a time.  If no reference
 * regions are specified, all alignments are returned.
 *
 * @param iter An iterator for an alignments collection
 * @param reference The reference name
 * @param start The start point of the region (1-based, inclusive)
 * @param end The end point of the region (1-based, inclusive)
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_range_filter(bioh5g_alignments_iterator iter,
                                           const char *reference,
                                           int32_t start,
                                           int32_t end);



/** Add a SAM MAPQ filter to an alignments iterator
 *
 *  Set a minimum MAPQ level on an iterator.  When set, only 
 *  alignments which have a MAPQ score above or equal to the minimum
 *  will be returned.
 *
 *  NOTE:  Allowable MAPQ values are from 0 to 255.
 *
 * @param iter An iterator for an alignments collection
 * @param min_mapq The minimum acceptable MAPQ value (inclusive)
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_mapq_filter(bioh5g_alignments_iterator iter,
                                          unsigned char min_mapq);



/** Add a SAM FLAGS filter to an alignments iterator
 *
 * Only alignments which have all the bits in the mask set (an AND mask)
 * will be returned.
 *
 * See the SAM spec for the meanings of the individual flags.
 *
 * @param iter An iterator for an alignments collection
 * @param mask A SAM flags mask.  Note that adding a subsequent mask to
 *   an iterator clobbers the old one.  They are NOT combined with logical
 *   OR.
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gadd_alignments_iterator_flags_filter(bioh5g_alignments_iterator iter,
                                           uint32_t mask);



/** Destroy an iterator for an alignments collection
 *
 *  The iterator is set to NULL as a part of deletion.
 *
 * @param[in,out] iter An iterator for an alignments collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gdestroy_alignments_iterator(/*IN-OUT*/ bioh5g_alignments_iterator *iter);



/** Add an alignment to a collection
 *
 * @param aligns A BioHDF alignments handle
 * @param data A BioHDF alignment
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gadd_alignment(const bioh5g_alignments aligns,
                    const bioh5g_alignment_data *data);



/** Get the index of the last alignment that was added
 *
 * Useful for higher level structures where links must be created.
 *
 * @param aligns A BioHDF alignments handle
 * @param[out] index The index of the last alignment that was added
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_index_of_last_added_alignment(const bioh5g_alignments aligns,
                                        /*OUT*/ int64_t *index);



/** Get the next alignment from an alignments collection
 *
 * @param iter An iterator for an alignments collection
 * @param[out] index The index of this alignment
 * @param[out] data The BioHDF alignment
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_next_alignment(bioh5g_alignments_iterator iter,
                         /*OUT*/ int64_t *index,
                         /*OUT*/ bioh5g_alignment_data **data);



/** Given an alignment index, get the alignment from an alignment collection
 *
 * @param aligns A BioHDF alignments handle
 * @param index The index of this alignment
 * @param[out] data The BioHDF alignment
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_alignment(const bioh5g_alignments aligns,
                    int64_t index,
                    /*OUT*/ bioh5g_alignment_data **data);



/** Free alignment data that has been obtained from the library
 *
 *  The data is set to NULL as a part of deletion.
 *
 * @param[in,out] data The BioHDF alignment
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gfree_alignment_data(/*IN-OUT*/ bioh5g_alignment_data **data);



/*****************************************************************************
 * Alignment hit index functionality                                         *
 *****************************************************************************/



/** Create an index for an alignments collection
 *
 *  A pre-existing index of the same method/type will be deleted.
 *
 * @param aligns The alignments collection
 * @param method The indexing method to use
 * @param props Index creation properties
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_index(bioh5g_alignments aligns,
                              bioh5g_alignments_index_method method,
                              biohdf_index_creation_properties props);



/*****************************************************************************
 * External file header storage                                              *
 *****************************************************************************/


/** Store a file header from an alignment file (e.g. SAM)
 *
 * NOTE: Headers are stored verbatim and are not parsed.
 *
 * @param aligns The alignments collection
 * @param format The format of the header
 * @param header The header string to store
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gstore_alignment_file_header(const bioh5g_alignments aligns,
                                  bioh5g_alignments_format format,
                                  const char *header);



/** Get a stored alignment file header
 *
 * NOTE: Headers are stored verbatim and are not parsed or generated.
 * This function can only return a previously-stored header,
 * it will not generate a header from scratch.
 *
 * @param aligns The alignments collection
 * @param[out] format The format of the header
 * @param[out] header The header string
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_alignment_file_header(const bioh5g_alignments aligns,
                                /*OUT*/ bioh5g_alignments_format *format,
                                /*OUT*/ char **header);



/*****************************************************************************
 * Data formats                                                              *
 *****************************************************************************/


/** Create an alignment string in a particular format
 *
 * Read data is normally required for correct SAM output.  If no read data
 * is supplied, the QNAME will be an arbitrary integer and SEQ and QUAL will
 * both be '*'.
 *
 * @param alignment The alignment data
 * @param read The read data
 * @param format The format of the output string
 * @param[out] alignment_string The alignment string
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcreate_alignment_string(const bioh5g_alignment_data *alignment,
                              const bioh5g_read_data *read,
                              bioh5g_alignments_format format,
                              /*OUT*/ char **alignment_string);



/** Write an alignment string in a particular format to an output stream
 *
 * This saves you from having to create temp strings that will
 * just be dumped to a stream.
 *
 * Read data is normally required for correct SAM output.  If no read data
 * is supplied, the QNAME will be an arbitrary integer and SEQ and QUAL will
 * both be '*'.
 *
 * @param alignment The alignment data
 * @param read The read data
 * @param format The format of the output string
 * @param stream The output stream (can be STDOUT)
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gwrite_alignment_to_stream(const bioh5g_alignment_data *alignment,
                                const bioh5g_read_data *read,
                                bioh5g_alignments_format format,
                                FILE *stream);




/*****************************************************************************
 * Accessor functions (needed for higher-language interoperation)            *
 *****************************************************************************/

/** @name Functions: Data Accessors */
/** @{*/



BIOHDF_API biohdf_error
BIOH5Gcreate_alignment_data(/*OUT*/ bioh5g_alignment_data **data);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_read_index(bioh5g_alignment_data *data,
                               /*OUT*/ int64_t *read_index);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_read_index(bioh5g_alignment_data *data,
                               int64_t read_index);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_reference(bioh5g_alignment_data *data,
                              /*OUT*/ char **reference);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_reference(bioh5g_alignment_data *data,
                              char *reference);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_position(bioh5g_alignment_data *data,
                             /*OUT*/ int32_t *position);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_position(bioh5g_alignment_data *data,
                             int32_t position);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_length(bioh5g_alignment_data *data,
                           /*OUT*/ int32_t *length);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_length(bioh5g_alignment_data *data,
                           int32_t length);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_mapq(bioh5g_alignment_data *data,
                             /*OUT*/ unsigned char *sam_mapq);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_mapq(bioh5g_alignment_data *data,
                             unsigned char sam_mapq);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_flags(bioh5g_alignment_data *data,
                              /*OUT*/ uint32_t *sam_flags);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_flags(bioh5g_alignment_data *data,
                              uint32_t sam_flags);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_cigar(bioh5g_alignment_data *data,
                              /*OUT*/ char **sam_cigar);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_cigar(bioh5g_alignment_data *data,
                              char *sam_cigar);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_tags(bioh5g_alignment_data *data,
                             /*OUT*/ char **sam_tags);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_tags(bioh5g_alignment_data *data,
                             char *sam_tags);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_rnext(bioh5g_alignment_data *data,
                              /*OUT*/ char **sam_rnext);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_rnext(bioh5g_alignment_data *data,
                              char *sam_rnext);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_pnext(bioh5g_alignment_data *data,
                              /*OUT*/ int32_t *sam_pnext);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_pnext(bioh5g_alignment_data *data,
                              int32_t sam_pnext);



BIOHDF_API biohdf_error
BIOH5Gget_alignment_sam_tlen(bioh5g_alignment_data *data,
                             /*OUT*/ int32_t *sam_tlen);



BIOHDF_API biohdf_error
BIOH5Gset_alignment_sam_tlen(bioh5g_alignment_data *data,
                             int32_t sam_tlen);

/** @} */

/*****************************************************************************
 * Alignments properties - create, destroy, access                           *
 *****************************************************************************/

/** @name Functions: Collection Creation Properties */
/** @{*/



BIOHDF_API biohdf_error
BIOH5Gcreate_alignments_properties(/*OUT*/ bioh5g_alignments_creation_properties *props);



BIOHDF_API biohdf_error
BIOH5Gdestroy_alignments_properties(/*OUT*/ bioh5g_alignments_creation_properties *props);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_reads_path(bioh5g_alignments_creation_properties props,
                                           char *reads_path);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_refs_scheme(bioh5g_alignments_creation_properties props,
                                            biohdf_string_storage_scheme scheme);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_tags_scheme(bioh5g_alignments_creation_properties props,
                                                 biohdf_string_storage_scheme scheme);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_cigar_scheme(bioh5g_alignments_creation_properties props,
                                             biohdf_string_storage_scheme scheme);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_refs_length(bioh5g_alignments_creation_properties props,
                                            size_t length);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_tags_length(bioh5g_alignments_creation_properties props,
                                            size_t length);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_cigar_length(bioh5g_alignments_creation_properties props,
                                             size_t length);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_chunk_size(bioh5g_alignments_creation_properties props,
                                           int64_t chunk_size);



BIOHDF_API biohdf_error
BIOH5Gset_alignments_properties_compression_level(bioh5g_alignments_creation_properties props,
                                                  compression_level level);


/** @} */

/** @} */
#endif
