/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#ifndef _BIOH5G_ALIGNMENTS_INTERNAL_H
#define _BIOH5G_ALIGNMENTS_INTERNAL_H

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_vector.h"



#define BIOH5G_HIT_DATA_NAME        "DATA_alignments"
#define BIOH5G_TEMPLATE_DATA_NAME   "DATA_templates"
#define BIOH5G_REFERENCES_NAME      "DATA_references"
#define BIOH5G_SAM_CIGAR_NAME       "DATA_sam_cigar"
#define BIOH5G_SAM_TAGS_NAME        "DATA_sam_tags"
#define BIOH5G_SAM_RNEXT_NAME       "DATA_sam_rnext"

#define BIOH5G_SAM_HEADER_NAME      "HEADER_sam"

#define BIOH5G_REF_POS_INDEX_NAME        "INDEX_ref_pos"

/* TODO: This is hacky and will break if we rename the suffix */
#define BIOH5G_REF_POS_INDEX_DSET_NAME   "INDEX_ref_pos_fixed_width"



/* Internal alignments collection handle */
struct _bioh5g_alignments
{
    /* The path to the alignments collection.
     * Used for debugging and error output
     */
    char *path;

    /* The HDF5 id of the group where the alignments are stored. */
    hid_t group;

    /* The associated reads path */
    char *reads_path;

    /* The HDF5 type for basic alignment data */
    hid_t basic_data_hdf5_type;

    /* The HDF5 type for template data */
    hid_t template_hdf5_type;

    /* How the alignments are indexed */
    bioh5g_alignments_index_method index_method;

    /* Alignments storage handles */
    vector *basic_data;
    vector *templates;
    vector *references;
    vector *sam_header;
    vector *sam_cigar;
    vector *sam_tags;
    vector *sam_rnext;

    /* Alignments index handle */
    vector *aligns_index;

    /* The HDF5 offset/index of the last added element */
    hsize_t last_added_elem_idx;
};



/* Basic alignment data for HDF5 I/O */
typedef struct
{
    /* basic alignment data */
    int64_t read_index;
    int32_t position;
    int32_t length;

    /* SAM data (stored here for now) */
    unsigned char   sam_mapq;       /* mapping quality */
    uint32_t        sam_flags;      /* SAM bitwise flags */

} basic_alignment_data;



/* Alignment "template" data for HDF5 I/O
 * (used for paired-end read data, etc.)
 * See the SAM spec v1.3 for explanations of the members.
 */
typedef struct
{
    int32_t sam_pnext;      /* position of next mate */
    int32_t sam_tlen;       /* template length */
} template_data;



/* BioHDF alignments collection creation properties */
struct _bioh5g_alignments_creation_properties
{
    /* The path to the reads collection */
    char *reads_path;

    /* hybrid, 1D concat or 2D fixed */
    biohdf_string_storage_scheme    refs_scheme;
    biohdf_string_storage_scheme    tags_scheme;
    biohdf_string_storage_scheme    cigar_scheme;

    /* fixed length for 2D or hybrid */
    size_t  refs_length;
    size_t  tags_length;
    size_t  cigar_length;

    /* Chunk size and compression level
     * Currently we use the same values for all datasets.
     */
    int64_t             chunk_size;
    compression_level   compression;
} ;



/* Internal alignments iterator handle */
struct _bioh5g_alignments_iterator
{
    /* A reference to the collection over which this iterator will scan */
    struct _bioh5g_alignments *alignments;

    /* NOTE: The following indexes are used for ALL vectors */

    /* The element which will be accessed next (0-based)
     *
     * This is initialized to zero when the iterator is created.
     * It can be either an index into the alignments collection
     * (unsorted data) or into the alignments index dataset (sorted data).
     */
    int64_t current_index;

    /* The last valid index in the collection (0-based)
     *
    * It can be either the cardinality of the alignments collection
    * (unsorted data) or of the alignments index dataset (sorted data).
     */
    int64_t max_index;

    /* FILTERS */

    /* The regions we'll be searching */
    int                         has_region_filters;
    bioh5g_reference_region    *regions;
    int64_t                     n_regions;

    /* Set when we're searching a new region in range queries */
    int                         is_new_region;

    /* The current region that is being traversed in the index */
    int                         current_region;

    /* Minimum SAM MAPQ value */
    int             has_mapq_filter;
    unsigned char   min_mapq;

    /* SAM FLAGS mask */
    int         has_sam_flags_filter;
    uint32_t    sam_flags_mask;
};



/* The basic sorting unit for BioHDF */
typedef struct
{
    int64_t     original_index;
    char       *reference;
    int32_t     position;
} ref_pos_data;


#endif

