/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_file.h"
#include "biohdf_file_internal.h"
#include "biohdf_group.h"
#include "biohdf_group_internal.h"
#include "biohdf_utility_internal.h"
#include "biohdf_vector.h"

#include "bioh5g_api.h"
#include "bioh5g_reads.h"
#include "bioh5g_reads_internal.h"



/* Emit a standard error message */
static void
_print_error_info(const char *collection_name)
{
    fprintf(stderr, "             \tBioHDF object: reads collection\n");
    if(NULL == collection_name)
    {
        fprintf(stderr, "             \treads name: NULL");
    }
    else
    {
        fprintf(stderr, "             \treads name: %s", collection_name);
    }
    fprintf(stderr, "\n\n");

    return;
}



/*****************************************************************************
 * Create, open, close                                                       *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcheck_reads_presence(biohdf_file file,
                           const char *path,
                           /*OUT*/ int *presence)
{
    htri_t check;            /* Output of H5Lexists() */

    char **groups = NULL;    /* Returned set of groups as strings */
    int n_groups;            /* Number of groups returned */
    int i;                   /* iterator */

    hid_t new_location = -1; /* The new group to be tested as we traverse
                              * the path.
                              */
    hid_t old_location = -1; /* The group in which the new group is tested.
                              * This starts at the file id on the first
                              * and is replaced with group IDs as we traverse
                              * the path.
                              */
    char *type_name;          /* The BioHDF type name from the collection */

    char *        bad_group        = NULL;           /* The group on which an operation failed */

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Get a list of the groups in the path */
    error = _parse_groups_from_path(path, &groups, &n_groups);
    CHECK_BIO(error);

    /* Iterate over the groups, checking to see if they exist */
    old_location = file->file_id;

    for(i = 0; i < n_groups; i++)
    {
        /* Does this group exist? */
        check = H5Lexists(old_location, groups[i], H5P_DEFAULT);
        if(check > 0)
        {
            /* Present. Move on to the next group */
            *presence = 1;
            new_location = H5Gopen(old_location, groups[i], H5P_DEFAULT);
            CHECK_HID_T(new_location);
        }
        else if (0 == check)
        {
            /* Not present.  Return. */
            *presence = 0;
            break;
        }
        else
        {
            /* A negative (error) value was returned  Return.*/
            *presence = 0;
            bad_group = groups[i];
            BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
            goto FAIL;
        }

        /* For the last group, we need to test if it is a collection of the
         * correct type.
         */
        if(n_groups - 1 == i)
        {
            error = _read_biohdf_type_attribute(new_location, &type_name);
            if(error != BIOHDF_NO_ERROR || 0 != strcmp(type_name, BIOH5G_READS_TYPE))
            {
                *presence = 0;
                BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
                goto FAIL;
            }

            hdf_error = H5Gclose(new_location);
            CHECK_HERR_T(hdf_error);
        }

        /* Close the old "containing" group, except on the first iteration
         * (we don't want to close the file, which is old_location on
         * the first iteration).
         */
        if(H5I_GROUP == H5Iget_type(old_location))
        {
            hdf_error = H5Gclose(old_location);
            CHECK_HERR_T(hdf_error);
        }

        /* The old old/containing group is replaced with the new group */
        old_location = new_location;
    }

    /**SUCCESS***************************************************************/

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);

    if(1 == H5Iis_valid(new_location))
    {
        H5Gclose(new_location);
    }

    /* Don't close the old location if it's the file_id that was passed in! */
    if(1 == H5Iis_valid(old_location) && H5I_GROUP == H5Iget_type(old_location))
    {
        H5Gclose(old_location);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gcreate_reads_collection(biohdf_file file,
                           bioh5g_reads_properties properties,
                           const char *path,
                           /*OUT*/ bioh5g_reads *reads)
{
    vector_storage_scheme sequences_scheme;
    vector_storage_scheme ids_scheme;

    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *reads = NULL;

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Create handle. */
    *reads = (bioh5g_reads)calloc(1, sizeof(struct _bioh5g_reads));
    CHECK_PTR(reads);

    /* Create path string. */
    (*reads)->path = (char *)calloc(strlen(path) + 1, sizeof(char));
    CHECK_PTR((*reads)->path);
    strcpy((*reads)->path, path);
    
    /* Create groups */
    error = _create_or_open_group(file, path, &((*reads)->group));
    CHECK_BIO(error);


    /* Add BioHDF type attribute to the collection */
    error = _write_biohdf_type_attribute((*reads)->group, BIOH5G_READS_TYPE);
    CHECK_BIO(error);


    /* Create data containers */

    switch(properties->sequences_scheme)
    {
    case FIXED_LENGTH_STRING:
        sequences_scheme = FIXED_SIZE_ARRAY;
        break;
    case VARIABLE_LENGTH_STRING:
        sequences_scheme = VARIABLE_SIZE_ARRAY;
        break;
    case HYBRID_STRING:
        sequences_scheme = HYBRID_ARRAY;
        break;
    default:
        return BIOHDF_BADNESS;
        break;
    }

    switch(properties->ids_scheme)
    {
    case FIXED_LENGTH_STRING:
        ids_scheme = FIXED_SIZE_ARRAY;
        break;
    case VARIABLE_LENGTH_STRING:
        ids_scheme = VARIABLE_SIZE_ARRAY;
        break;
    case HYBRID_STRING:
        ids_scheme = HYBRID_ARRAY;
        break;
    default:
        return BIOHDF_BADNESS;
        break;
    }

    /* Reads */
    error = _create_vector((*reads)->group,
        BIOH5G_SEQUENCES_NAME,
        H5T_NATIVE_CHAR,
        sequences_scheme,
        properties->sequences_length,
        properties->chunk_size,
        properties->compression,
        &((*reads)->sequences));
    CHECK_BIO(error);

    /* Qualities */
    error = _create_vector((*reads)->group,
        BIOH5G_QUALS_NAME,
        H5T_NATIVE_CHAR,
        sequences_scheme,
        properties->sequences_length,
        properties->chunk_size,
        properties->compression,
        &((*reads)->qualities));
    CHECK_BIO(error);

    /* Identifiers - may have a different storage type than R & Q */
    error = _create_vector((*reads)->group,
        BIOH5G_IDS_NAME,
        H5T_NATIVE_CHAR,
        ids_scheme,
        properties->ids_length,
        properties->chunk_size,
        properties->compression,
        &((*reads)->identifiers));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);

    if(*reads != NULL)
    {
        /* Free the path string */
        free((*reads)->path);

        /* close HDF5 objects */
        if(H5Iis_valid((*reads)->group)) { H5Gclose((*reads)->group); }

        /* close vectors */
        _close_vector(&((*reads)->sequences));
        _close_vector(&((*reads)->qualities));
        _close_vector(&((*reads)->identifiers));
    }

    free(*reads);
    *reads = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gopen_reads_collection(biohdf_file file,
                         const char *path,
                         biohdf_open_mode mode,
                         /*OUT*/ bioh5g_reads *reads)
{
    int is_read_only;

    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *reads = NULL;

    CHECK_PTR(file);
    CHECK_HID_T(file->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Create handle. */
    *reads = (bioh5g_reads)calloc(1, sizeof(struct _bioh5g_reads));
    CHECK_PTR(*reads);

    /* Create path string. */
    (*reads)->path = (char *)calloc(strlen(path) + 1, sizeof(char));
    CHECK_PTR((*reads)->path);
    strcpy((*reads)->path, path);

    /* Open groups */
    error = _create_or_open_group(file, path, &((*reads)->group));
    CHECK_BIO(error);

    /* Open datasets */

    if (READ_ONLY == mode)
    {
        is_read_only = 1;
    }
    else
    {
        is_read_only = 0;
    }

    /* Reads */
    error = _open_vector((*reads)->group,
        BIOH5G_SEQUENCES_NAME,
        is_read_only,
        &((*reads)->sequences));
    CHECK_BIO(error);

    /* Qualities */
    error = _open_vector((*reads)->group,
        BIOH5G_QUALS_NAME,
        is_read_only,
        &((*reads)->qualities));
    CHECK_BIO(error);

    /* Identifiers - may have a different storage type than R & Q */
    error = _open_vector((*reads)->group,
        BIOH5G_IDS_NAME,
        is_read_only,
        &((*reads)->identifiers));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);

    if(*reads != NULL)
    {
        /* Free the path string */
        free((*reads)->path);

        /* close HDF5 objects */
        if(H5Iis_valid((*reads)->group)) { H5Gclose((*reads)->group); }

        /* close vectors */
        _close_vector(&((*reads)->sequences));
        _close_vector(&((*reads)->qualities));
        _close_vector(&((*reads)->identifiers));
    }

    free(*reads);
    *reads = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error 
BIOH5Gclose_reads_collection(/*IN-OUT*/ bioh5g_reads *reads)
{
    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* None - NULL is ok */

    /**CODE*****************************************************************/

    if(NULL == *reads){ return BIOHDF_NO_ERROR; }

    /* Free the path string */
    free((*reads)->path);

    /* Close BioHDF storage */
    error = _close_vector(&((*reads)->sequences));
    CHECK_BIO(error);

    error = _close_vector(&((*reads)->qualities));
    CHECK_BIO(error);
    
    error = _close_vector(&((*reads)->identifiers));
    CHECK_BIO(error);

    /* Close the collection group. */
    hdf_error = H5Gclose((*reads)->group);
    CHECK_HERR_T(hdf_error);

    free(*reads);
    *reads = NULL;
    
    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*reads) {
        _print_error_info((*reads)->path);
    }
    else {
        _print_error_info(NULL);
    }

    /* Try again, but don't report errors
     * This may seem redundant, but I WANT close errors to be reported since
     * they may indicate underlying problems.
     */
    if(*reads != NULL)
    {
        /* close HDF5 objects */
        if(H5Iis_valid((*reads)->group)) { H5Gclose((*reads)->group); }

        /* close vectors */
        _close_vector(&((*reads)->sequences));
        _close_vector(&((*reads)->qualities));
        _close_vector(&((*reads)->identifiers));
    }

    return BIOHDF_BADNESS;
}


/*****************************************************************************
 * Append and read data                                                      *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gget_reads_count(const bioh5g_reads reads,
                      /*OUT*/ int64_t *count)
{
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(reads);

    /**CODE*****************************************************************/

    error = _get_vector_count(reads->sequences, (hsize_t *)count);
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(reads) {
        _print_error_info(reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gcreate_reads_iterator(const bioh5g_reads reads,
                            /*OUT*/ bioh5g_reads_iterator *iter)
{
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    *iter = NULL;

    CHECK_PTR(reads);

    /**CODE*****************************************************************/

    *iter = (bioh5g_reads_iterator)calloc(1, sizeof(struct _bioh5g_reads_iterator));
    CHECK_PTR(iter);

    /* Set the alignments reference */
    (*iter)->reads = reads;

    (*iter)->current_index   = 0;

    error = BIOH5Gget_reads_count(reads, &((*iter)->max_index));
    CHECK_BIO(error);
    (*iter)->max_index -= 1;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(reads) {
        _print_error_info(reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    free(*iter);

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gdestroy_reads_iterator(/*IN-OUT*/ bioh5g_reads_iterator *iter)
{
    /* Can't really fail here */

    free(*iter);
    *iter = NULL;

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gadd_read(const bioh5g_reads reads,
               const bioh5g_read_data *data)
{
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(reads);

    /**CODE*****************************************************************/

    if(data->quality_values != NULL
        && strlen(data->sequence) != strlen(data->quality_values))
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    error = _add_vector_element(reads->sequences,
        data->sequence,
        strlen(data->sequence));
    CHECK_BIO(error);

    error = _add_vector_element(reads->qualities,
        data->quality_values,
        strlen(data->quality_values));
    CHECK_BIO(error);

    error = _add_vector_element(reads->identifiers,
        data->identifier,
        strlen(data->identifier));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(reads) {
        _print_error_info(reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_index_of_last_added_read(const bioh5g_reads reads,
                                   /*OUT*/ int64_t *index)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(reads);

    /**CODE*****************************************************************/

    *index = (int64_t)(reads->sequences->index_of_last_added_element);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(reads) {
        _print_error_info(reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_next_read(bioh5g_reads_iterator iter,
                    /*OUT*/ int64_t *index,
                    /*OUT*/ bioh5g_read_data **data)
{
    int append_nul = 1;  /* Append ASCII nul to strings */

    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(iter);

    /**CODE*****************************************************************/

    /* If the next index is > the max, output null and return "no error" */
    if(iter->current_index > iter->max_index)
    {
        *data = NULL;
        *index = -1;
        return BIOHDF_NO_ERROR;
    }

    /* Set the current index */
    *index = iter->current_index;

    /* Set up output data structure */
    *data = (bioh5g_read_data *)calloc(1, sizeof(bioh5g_read_data));
    CHECK_PTR(*data);

    error = _get_vector_element(iter->reads->sequences, 
        *index, 
        append_nul, 
        (void **)&((*data)->sequence));
    CHECK_BIO(error);

    error = _get_vector_element(iter->reads->qualities, 
        *index,
        append_nul, 
        (void **)&((*data)->quality_values));
    CHECK_BIO(error);

    error = _get_vector_element(iter->reads->identifiers, 
        *index, 
        append_nul, 
        (void **)&((*data)->identifier));
    CHECK_BIO(error);

    /* Increment the iterator */
    iter->current_index += 1;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(iter && iter->reads) {
        _print_error_info(iter->reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gget_read(const bioh5g_reads reads,
               int64_t index,
               /*OUT*/ bioh5g_read_data **data)
{
    biohdf_error error;
    int append_nul = 1;

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(reads);

    /**CODE*****************************************************************/

    *data = (bioh5g_read_data *)calloc(1, sizeof(bioh5g_read_data));
    CHECK_PTR(*data);

    error = _get_vector_element(reads->sequences, 
        (hsize_t)index, 
        append_nul, 
        (void **)&((*data)->sequence));
    CHECK_BIO(error);

    error = _get_vector_element(reads->qualities, 
        (hsize_t)index,
        append_nul, 
        (void **)&((*data)->quality_values));
    CHECK_BIO(error);
        
    error = _get_vector_element(reads->identifiers, 
        (hsize_t)index, 
        append_nul, 
        (void **)&((*data)->identifier));
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    if(reads) {
        _print_error_info(reads->path);
    }
    else {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gfree_read(/*IN-OUT*/ bioh5g_read_data **data)
{

    /* There is very little error checking that this function could
     * reasonably perform.
     */

    if(*data != NULL)
    {
        free((*data)->sequence);
        free((*data)->quality_values);
        free((*data)->identifier);
        free(*data);
        (*data) = NULL;
    }

    return BIOHDF_NO_ERROR;
}



/*****************************************************************************
 * Data formats                                                              *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_read_string(const bioh5g_read_data *read,
                         bioh5g_reads_format format,
                         /*OUT*/ char **read_string)
{
    size_t total_length = 0;
    size_t id_length    = 0;
    size_t seq_length   = 0;
    size_t qual_length  = 0;
    size_t pos          = 0;

    /**CHECK*PARAMETERS*****************************************************/

    *read_string = NULL;

    CHECK_PTR(read);
    CHECK_PTR(read->identifier);
    CHECK_PTR(read->sequence);

    if(FASTQ_FORMAT == format)
    {
        CHECK_PTR(read->quality_values);
    }

    /**CODE*****************************************************************/

    seq_length  = strlen(read->sequence);
    id_length   = strlen(read->identifier);
    qual_length = strlen(read->quality_values);

    if(seq_length != qual_length)
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    /* determine total string length */
    total_length = 1;                        /* start with nul */
    total_length += (seq_length + 1);        /* sequence length + \n */
    total_length += (id_length + 2);         /* id length, (@ or >) and newline */
    if (FASTQ_FORMAT == format)
    {
        total_length += (qual_length + 3);   /* quality line, '+' and two newlines */
    }

    /* allocate memory */
    (*read_string) = (char *)calloc(total_length, sizeof(char));
    CHECK_PTR(*read_string);

    /* copy string data */
    pos = 0;

    /* identifier */
    if(FASTQ_FORMAT == format)
    {
        strcpy((*read_string) + pos, "@");
    }
    else if (FASTA_FORMAT == format)
    {
        strcpy((*read_string) + pos, ">");
    }
    pos += 1;
    
    strcpy((*read_string) + pos, read->identifier);
    pos += id_length;
    
    strcpy((*read_string) + pos, "\n");
    pos += 1;

    /* sequence */
    strcpy((*read_string) + pos, read->sequence);
    pos += seq_length;

    strcpy((*read_string) + pos, "\n");
    pos += 1;

    /* quality values - only written for FASTQ format */
    if (FASTQ_FORMAT == format)
    {
        strcpy((*read_string) + pos, "+\n");
        pos += 2;

        strcpy((*read_string) + pos, read->quality_values);
        pos += qual_length;

        strcpy((*read_string) + pos, "\n");
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    free(*read_string);
    *read_string = NULL;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOH5Gwrite_read_to_stream(const bioh5g_read_data *read,
                           bioh5g_reads_format format,
                           FILE *stream)
{
    int           printf_error     = -1;             /* fprintf error value */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(read);
    CHECK_PTR(read->identifier);
    CHECK_PTR(read->sequence);

    if(FASTQ_FORMAT == format)
    {
        CHECK_PTR(read->quality_values);
    }

    CHECK_PTR(stream);

    if(strlen(read->sequence) != strlen(read->quality_values))
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    /**CODE*****************************************************************/

    if(FASTQ_FORMAT == format)
    {
        /* identifier */
        printf_error = fprintf(stream, "@%s\n", read->identifier);
        CHECK_INT_GT(printf_error, -1);

        /* sequence */
        printf_error = fprintf(stream, "%s\n", read->sequence);
        CHECK_INT_GT(printf_error, -1);

        /* quality values */
        printf_error = fprintf(stream, "+\n%s\n", read->quality_values);
        CHECK_INT_GT(printf_error, -1);
    }
    else if(FASTA_FORMAT == format)
    {
        /* identifier */
        printf_error = fprintf(stream, ">%s\n", read->identifier);
        CHECK_INT_GT(printf_error, -1);

        /* sequence */
        printf_error = fprintf(stream, "%s\n", read->sequence);
        CHECK_INT_GT(printf_error, -1);
    }
    else
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Accessor functions (needed for higher-language interoperation)            *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_read_data(/*OUT*/ bioh5g_read_data **data)
{
    (*data) = (bioh5g_read_data *)calloc(1, sizeof(bioh5g_read_data));
    (*data)->identifier = NULL;
    (*data)->quality_values = NULL;
    (*data)->sequence = NULL;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_read_identifier(bioh5g_read_data *data, /*OUT*/ char **identifier)
{
    (*identifier) = (char *)calloc(strlen(data->identifier) + 1, sizeof(char));
    strcpy(*identifier, data->identifier);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_read_identifier(bioh5g_read_data *data, char *identifier)
{
    if(NULL != data->identifier)
    { 
        free(data->identifier);
    }
    data->identifier = (char *)calloc(strlen(identifier) + 1, sizeof(char));
    strcpy(data->identifier, identifier);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_read_sequence(bioh5g_read_data *data, /*OUT*/ char **sequence)
{
    (*sequence) = (char *)calloc(strlen(data->sequence) + 1, sizeof(char));
    strcpy(*sequence, data->sequence);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_read_sequence(bioh5g_read_data *data, char *sequence)
{
    if(NULL != data->sequence)
    { 
        free(data->sequence);
    }
    data->sequence = (char *)calloc(strlen(sequence) + 1, sizeof(char));
    strcpy(data->sequence, sequence);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gget_read_quality_values(bioh5g_read_data *data, /*OUT*/ char **quality_values)
{
    (*quality_values) = (char *)calloc(strlen(data->quality_values) + 1, sizeof(char));
    strcpy(*quality_values, data->quality_values);
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_read_quality_values(bioh5g_read_data *data, char *quality_values)
{
    if(NULL != data->quality_values)
    { 
        free(data->quality_values);
    }
    data->quality_values = (char *)calloc(strlen(quality_values) + 1, sizeof(char));
    strcpy(data->quality_values, quality_values);
    return BIOHDF_NO_ERROR;
}



/*****************************************************************************
 * Reads properties - create, destroy, access                                *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOH5Gcreate_reads_properties(/*OUT*/ bioh5g_reads_properties *props)
{
    *props = (bioh5g_reads_properties)calloc(1, sizeof(struct _bioh5g_reads_properties));

    (*props)->reads_type        = BASE_SPACE;
    (*props)->chunk_size        = BIOHDF_DEFAULT_CHUNK_SIZE;
    (*props)->compression       = BIOHDF_DEFAULT_COMPRESSION_LEVEL;
    (*props)->ids_scheme        = VARIABLE_LENGTH_STRING;
    (*props)->sequences_scheme  = VARIABLE_LENGTH_STRING;
    (*props)->ids_length        = 0;
    (*props)->sequences_length  = 0;

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gdestroy_reads_properties(/*OUT*/ bioh5g_reads_properties *props)
{
    if(*props)
    {
        free(*props);
        (*props) = NULL;
    }

    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_reads_type(bioh5g_reads_properties props,
                                      bioh5g_reads_type reads_type)
{
    props->reads_type = reads_type;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_chunk_size(bioh5g_reads_properties props,
                                      int64_t chunk_size)
{
    props->chunk_size = chunk_size;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_compression_level(bioh5g_reads_properties props,
                                             compression_level level)
{
    props->compression = level;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_sequences_scheme(bioh5g_reads_properties props,
                                            biohdf_string_storage_scheme scheme)
{
    props->sequences_scheme = scheme;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_identifiers_scheme(bioh5g_reads_properties props,
                                              biohdf_string_storage_scheme scheme)
{
    props->ids_scheme = scheme;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_sequences_length(bioh5g_reads_properties props,
                                            size_t length)
{
    props->sequences_length = length;
    return BIOHDF_NO_ERROR;
}



BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_identifiers_length(bioh5g_reads_properties props,
                                              size_t length)
{
    props->ids_length = length;
    return BIOHDF_NO_ERROR;
}


