/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/
 
/** @defgroup bioh5g_reads bioh5g_reads
 *
 *  Represents NGS reads (FASTQ, etc. entries).
 *
 *  @{
 */

#ifndef _BIOH5G_READS_H
#define _BIOH5G_READS_H

#include <stdio.h>

#include "biohdf_api.h"
#include "biohdf_error.h"




/*****************************************************************************
 * Type definitions                                                          *
 *****************************************************************************/



/** BioHDF reads collection handle */
typedef struct _bioh5g_reads *bioh5g_reads;



/** BioHDF reads iterator handle */
typedef struct _bioh5g_reads_iterator *bioh5g_reads_iterator;



/** BioHDF reads creation properties */
typedef struct _bioh5g_reads_properties *bioh5g_reads_properties;



/*****************************************************************************
 * Common structs and enums                                                  *
 *****************************************************************************/



/** Describes the data "space" of the reads (base, color, etc.) 
 *
 *  Reads are stored verbatim and not converted between spaces.
 */
typedef enum
{
    BASE_SPACE,       /**< Base/sequence space reads */
    COLOR_SPACE       /**< Color space (SOLiD) reads */
} bioh5g_reads_type;



/** The text file format for read I/O
 *
 * BioHDF stores the FASTQ quality values verbatim and does no
 * attempt to convert or normalize them between quality score
 * schemes.
 */
typedef enum
{
    FASTQ_FORMAT,      /**< FASTQ format */
    FASTA_FORMAT       /**< FASTA (or CSFASTA) format */
} bioh5g_reads_format;



/*****************************************************************************
 * Data container                                                            *
 *****************************************************************************/



/** Read data container
 *
 *  All data is stored verbatim, in the same form as the
 *  source file.  BioHDF does no conversions between sequence
 *  spaces or quality value schemes.
 */
typedef struct
{
    char *sequence;          /**< Sequence */
    char *quality_values;    /**< Quality values */
    char *identifier;        /**< Identifier or comment ('>') line */
} bioh5g_read_data;



/*****************************************************************************
 * Create, open, close                                                       *
 *****************************************************************************/



/** Test if a reads collection exists
 *
 * This function will return TRUE if a collection of the same type exists
 * at the named location.  If any other HDF5 or BioHDF object with that
 * same name exists, TRUE will be returned as well as an error code, the
 * assumption being that Bad Code(tm) that does not check return values
 * will be more likely to attempt to open code (and fail), rather than
 * create things (which my partially succeed, making a mess).
 *
 * @param file A BioHDF file handle
 * @param path The BioHDF path to the collection
 * @param[out] presence TRUE if the collection exists, FALSE if it does not.
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcheck_reads_presence(biohdf_file file,
                           const char *path,
                           /*OUT*/ int *presence);



/** Create (and open) a new reads collection
 *
 *  The collection handle returned by this function will be ready
 *  to accept I/O.
 *
 * @param file A BioHDF file handle
 * @param properties Collection creation properties
 * @param path The BioHDF path to the collection
 * @param[out] reads A BioHDF reads handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gcreate_reads_collection(biohdf_file file,
                              bioh5g_reads_properties properties,
                              const char *path,
                              /*OUT*/ bioh5g_reads *reads);



/** Open an existing reads collection
 *
 * @param file A BioHDF file handle
 * @param path The BioHDF path to the collection
 * @param mode The access mode (read-only | read-write)
 * @param[out] reads A BioHDF reads handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gopen_reads_collection(biohdf_file file,
                            const char *path,
                            biohdf_open_mode mode,
                            /*OUT*/ bioh5g_reads *reads);



/** Close an open reads collection
 *
 *  This function will set the collection handle to NULL after freeing it.
 *
 * @param[in,out] reads A BioHDF reads handle
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error 
BIOH5Gclose_reads_collection(/*IN-OUT*/ bioh5g_reads *reads);



/*****************************************************************************
 * Append and read data                                                      *
 *****************************************************************************/



/** Get the number of stored reads in a collection
 *
 * @param reads A BioHDF reads handle
 * @param[out] count The number of reads in the collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_reads_count(const bioh5g_reads reads,
                      /*OUT*/ int64_t *count);



/** Create an iterator for a reads collection
 *
 * @param reads A BioHDF reads handle
 * @param[out] iter An iterator for a reads collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcreate_reads_iterator(const bioh5g_reads reads,
                            /*OUT*/ bioh5g_reads_iterator *iter);



/** Destroy an iterator for a reads collection
 *
 *  The iterator is set to NULL as a part of deletion.
 *
 * @param[in,out] iter An iterator for a reads collection
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gdestroy_reads_iterator(/*OUT*/ bioh5g_reads_iterator *iter);



/** Add a read to a collection
 *
 * @param reads A BioHDF reads handle
 * @param data A BioHDF read
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gadd_read(const bioh5g_reads reads,
               const bioh5g_read_data *data);



/** Get the index of the last read that was added
 *
 * Useful for adding SAM lines where a link to a given read must be
 * created.
 *
 * @param reads A BioHDF reads handle
 * @param[out] index The index of the last read that was added
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_index_of_last_added_read(const bioh5g_reads reads,
                                   /*OUT*/ int64_t *index);



/** Get the next read from a reads collection
 *
 * When the iterator has finished traversing the collection, data will
 * be NULL, index will be -1 and the return value will be BIOHDF_NO_ERROR.
 *
 * @param iter An iterator for a reads collection
 * @param[out] index The index of this read
 * @param[out] data The BioHDF read
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_next_read(bioh5g_reads_iterator iter,
                    /*OUT*/ int64_t *index,
                    /*OUT*/ bioh5g_read_data **data);



/** Given a read index, get the read from a reads collection
 *
 * @param reads A BioHDF reads handle
 * @param index The index of this read
 * @param[out] data The BioHDF read
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gget_read(const bioh5g_reads reads,
               int64_t index,
               /*OUT*/ bioh5g_read_data **data);



/** Free read data that has been obtained from the library
 *
 *  The data is set to NULL as a part of deletion.
 *
 * @param[in,out] data The BioHDF read
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gfree_read(/*OUT*/ bioh5g_read_data **data);



/*****************************************************************************
 * Data formats                                                              *
 *****************************************************************************/


/** Create a read string in a given format (FASTQ, etc.)
 *
 * The read sequence and identifier must not be NULL.  If the format is FASTQ,
 * the quality values must not be NULL.  The sequence and quality values must
 * also have the same length for FASTQ output.
 *
 * @param read The BioHDF read
 * @param format The format for the output string
 * @param[out] read_string The output string
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gcreate_read_string(const bioh5g_read_data *read,
                         bioh5g_reads_format format,
                         /*OUT*/ char **read_string);


/** Output a read to a stream in a given format
 *
 * This saves you from having to create temp strings that will
 * just be dumped to a stream.
 *
 * The read sequence and identifier must not be NULL.  If the format is FASTQ,
 * the quality values must not be NULL.  The sequence and quality values must
 * also have the same length for FASTQ output.
 *
 * @param read The BioHDF read
 * @param format The format for the output string
 * @param stream The output stream (can be STDOUT)
 *
 * @returns BIOHDF_NO_ERROR on success, a BioHDF error code on failure
 */
BIOHDF_API biohdf_error
BIOH5Gwrite_read_to_stream(const bioh5g_read_data *read,
                           bioh5g_reads_format format,
                           FILE *stream);



/*****************************************************************************
 * Accessor functions (needed for higher-language interoperation)            *
 *****************************************************************************/

/** @name Functions: Data Accessors */
/** @{*/

/** Create a read data container */
BIOHDF_API biohdf_error
BIOH5Gcreate_read_data(/*OUT*/ bioh5g_read_data **data);



/** Get the read identifier */
BIOHDF_API biohdf_error
BIOH5Gget_read_identifier(bioh5g_read_data *data, /*OUT*/ char **identifier);



/** Set the read identifier */
BIOHDF_API biohdf_error
BIOH5Gset_read_identifier(bioh5g_read_data *data, char *identifier);



/** Get the read sequence */
BIOHDF_API biohdf_error
BIOH5Gget_read_sequence(bioh5g_read_data *data, /*OUT*/ char **sequence);



/** Set the read sequence */
BIOHDF_API biohdf_error
BIOH5Gset_read_sequence(bioh5g_read_data *data, char *sequence);



/** Get the read quality values */
BIOHDF_API biohdf_error
BIOH5Gget_read_quality_values(bioh5g_read_data *data, /*OUT*/ char **quality_values);



/** Set the read quality values */
BIOHDF_API biohdf_error
BIOH5Gset_read_quality_values(bioh5g_read_data *data, char *quality_values);

/** @}*/



/*****************************************************************************
 * Reads properties - create, destroy, access                                *
 *****************************************************************************/

/** @name Functions: Collection Creation Properties */
/** @{*/



/** Create a reads properties container */
BIOHDF_API biohdf_error
BIOH5Gcreate_reads_properties(/*OUT*/ bioh5g_reads_properties *props);



/** Destroy a reads properties container */
BIOHDF_API biohdf_error
BIOH5Gdestroy_reads_properties(/*OUT*/ bioh5g_reads_properties *props);



/** Set reads properties reads type */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_reads_type(bioh5g_reads_properties props,
                                      bioh5g_reads_type reads_type);



/** Set reads properties chunk size */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_chunk_size(bioh5g_reads_properties props,
                                      int64_t chunk_size);



/** Set reads properties compression level */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_compression_level(bioh5g_reads_properties props,
                                             compression_level level);



/** Set reads properties sequences storage scheme */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_sequences_scheme(bioh5g_reads_properties props,
                                            biohdf_string_storage_scheme scheme);



/** Set reads properties identifiers storage scheme */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_identifiers_scheme(bioh5g_reads_properties props,
                                              biohdf_string_storage_scheme scheme);



/** Set reads properties sequences length */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_sequences_length(bioh5g_reads_properties props,
                                            size_t length);



/** Set reads properties identifiers length */
BIOHDF_API biohdf_error
BIOH5Gset_reads_properties_identifiers_length(bioh5g_reads_properties props,
                                              size_t length);

/** @} */

/** @} */
#endif
