/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

/** @defgroup biohdf_api biohdf_api
 *
 *  Basic BioHDF data structures
 *
 *  @{
 */

#ifndef _BIOHDF_API_H
#define _BIOHDF_API_H

#include <stdint.h>
#include <inttypes.h>


/*****************************************************************************
 * API header                                                                *
 *****************************************************************************/



/** Platform-specific library prefix 
 *
 *  extern "C", __declspec(dllexport), etc.
 */
#if defined(_WIN32)
    #if defined(LIBBIOHDF_EXPORTS)
        /* When the header file is used in the context of creating the BioHDF
         * dll, we need to label all exported functions "dllexport" so they
         * can be accessed by other software.
         */
        #if defined(__cplusplus)
            #define BIOHDF_API extern "C" __declspec(dllexport)
        #else
            #define BIOHDF_API __declspec(dllexport)
        #endif
    #elif defined(LIBBIOHDF_DLL_IMPORT)
        /* When the header file is used in the context of importing the
         * functions from a dll (the default case), we label them with
         * dllimport.  This is entirely a performance feature (it saves
         * one level of indirection) since I'm not exporting variables,
         * which do require the dllimport.
         */
        #if defined(__cplusplus)
            #define BIOHDF_API extern "C" __declspec(dllimport)
        #else
            #define BIOHDF_API __declspec(dllimport)
        #endif
    #else
        /* Everything else uses undecorated functions */
        #if defined(__cplusplus)
            #define BIOHDF_API extern "C" 
        #else
            #define BIOHDF_API
        #endif
    #endif
#else /* Not Windows */
    #if defined(__cplusplus)
        #define BIOHDF_API extern "C"
    #else
        #define BIOHDF_API
    #endif
#endif



/*****************************************************************************
 * BioHDF Defaults                                                           *
 *****************************************************************************/



/** The max string size in some fixed data structures */
#define BIOHDF_MAX_STRING_SIZE 4096



/** The default compression level (zlib compression) */
#define BIOHDF_DEFAULT_COMPRESSION_LEVEL 5



/** The default HDF5 chunk size for all datasets (KB) */
#define BIOHDF_DEFAULT_CHUNK_SIZE 1024



/*****************************************************************************
 * BioHDF Type Attribute                                                     *
 *****************************************************************************/



/** The name of the BioHDF type attribute */
#define BIOHDF_TYPE_ATTR_NAME "BIOHDF_TYPE"



/*****************************************************************************
 * BioHDF Common Structs and Enums                                           *
 *****************************************************************************/



/** The BioHDF string storage scheme
 *
 * BioHDF can use several storage schemes to hold string data.  The default
 * is variable length storage since there is no risk of data truncation.
 * However, there are performance advantages (in both size and I/O speed)
 * to using the fixed-length scheme (assuming your strings are roughly
 * uniform in length).  Unfortunately, profiling is required to determine
 * which scheme will work best in a given situation.
 *
 * NOTE: The hybrid type is unimplemented at this time.  It should
 * eventually combine the flexibility of the variable-length scheme
 * with the performance benefits of the fixed-length scheme.
 */
typedef enum
{
    FIXED_LENGTH_STRING,     /**< All strings are below a given length */
    VARIABLE_LENGTH_STRING,  /**< Strings can be any length */
    HYBRID_STRING,           /**< UNIMPLEMENTED - DO NOT USE */
} biohdf_string_storage_scheme;



/** Modes for opening collections.
 */
typedef enum _biohdf_open_mode
{
    READ_ONLY = 0,      /**< Read-only access (DEFAULT) */
    READ_WRITE          /**< Read and write access */
} biohdf_open_mode;



/** GZIP compression level.
 *
 * The same compression level is used for all data in the collection.
 */
typedef enum
{
    NO_COMPRESSION = 0,   /**< No compression */
    GZIP_1 = 1,           /**< compression level 1 */
    GZIP_2 = 2,           /**< compression level 2 */
    GZIP_3 = 3,           /**< compression level 3 */
    GZIP_4 = 4,           /**< compression level 4 */
    GZIP_5 = 5,           /**< compression level 5 */
    GZIP_6 = 6,           /**< compression level 6 */
    GZIP_7 = 7,           /**< compression level 7 */
    GZIP_8 = 8,           /**< compression level 8 */
    GZIP_9 = 9            /**< compression level 9 */
}
compression_level;



/** @} */
#endif
