/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

/** @defgroup biohdf_error biohdf_error
 *
 *  BioHDF error functions and data structures.
 *
 *  @{
 */

#ifndef _BIOHDF_ERROR_H
#define _BIOHDF_ERROR_H



/*****************************************************************************
 * Error codes and definitions                                               *
 *****************************************************************************/



/** The BioHDF error return type */
typedef int biohdf_error;



/** Error code for "no error" */
#define BIOHDF_NO_ERROR 0



/** Generic "bad" error code */
#define BIOHDF_BADNESS -1



/*****************************************************************************
 * Error macros                                                              *
 *****************************************************************************/

/* NOTE: Any function which uses these macros must have a FAIL target        */

#if defined(_MSC_VER)
    #define __func__ __FUNCTION__
#endif

/* Check to ensure a BioHDF error (biohdf_error) is BIOHDF_NO_ERROR. */
#define CHECK_BIO(val)                                                      \
do                                                                          \
{                                                                           \
    if((val) != BIOHDF_NO_ERROR){                                           \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an HDF5 herr_t is not an error value. */
#define CHECK_HERR_T(val)                                                   \
do                                                                          \
{                                                                           \
    if((val) < 0){                                                          \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an HDF5 hid_t is not an error value. */
#define CHECK_HID_T(val)                                                    \
do                                                                          \
{                                                                           \
    if((val) < 0){                                                          \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure a pointer is not NULL. */
#define CHECK_PTR(val)                                                      \
do                                                                          \
{                                                                           \
    if(NULL == (val)){                                                      \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer matches a test value. */
#define CHECK_INT_EQ(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) != (test)){                                                    \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer does not match a test value. */
#define CHECK_INT_NE(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) == (test)){                                                    \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer is not >= a particular value */
#define CHECK_INT_LT(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) >= (test)){                                                     \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer is not <= a particular value */
#define CHECK_INT_GT(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) <= (test)){                                                     \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer is not > a particular value */
#define CHECK_INT_LE(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) > (test)){                                                     \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)

/* Check to ensure an integer is not < a particular value */
#define CHECK_INT_GE(val, test)                                             \
do                                                                          \
{                                                                           \
    if((val) < (test)){                                                     \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)


/*****************************************************************************
 * Set/unset verbose STDERR output                                           *
 *****************************************************************************/



/** Set whether or not error information is sent to STDERR.
 *
 * If this flag is not set, errors are only reported via the returned
 * error codes.
 *
 * @param use_stderr FALSE if no STDERR output, TRUE if output will go to STDERR.
 */
BIOHDF_API void
BIOHDFset_stderr_setting(int use_stderr);



/** Determine whether or not error information is sent to STDERR.
 *
 * @param[out] use_stderr FALSE if no STDERR output, TRUE if output will go to STDERR.
 */
BIOHDF_API void
BIOHDFget_stderr_setting(int *use_stderr);



/*****************************************************************************
 * Emit an error message                                                     *
 *****************************************************************************/



/** Write an error message.
 *
 */
BIOHDF_API void
BIOHDFemit_error_message(const char *filename, const char *function, int line);


/** @} */
#endif
