/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_file.h"
#include "biohdf_file_internal.h"



/* Write an error message in a standard format */
static void
_print_error_info(const char *filename)
{
    fprintf(stderr, "             \tBioHDF object: file\n");
    if(NULL == filename)
    {
        fprintf(stderr, "             \tfile name: NULL");
    }
    else
    {
        fprintf(stderr, "             \tfile name: %s", filename);
    }
    fprintf(stderr, "\n\n");

    return;
}



BIOHDF_API biohdf_error
BIOHDFcreate_biohdf_file(const char* filename,
                         /*OUT*/ biohdf_file *f)
{

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(filename);
    CHECK_INT_NE(filename, 0);

    /**CODE*****************************************************************/

    /* Get memory for the file handle */
    *f = (biohdf_file)calloc(1, sizeof(struct _biohdf_file));
    CHECK_PTR(*f);

    /* Create the HDF5 file
     * H5F_ACC_EXCL means that the create function will NOT overwrite
     * existing files. This avoids inadvertent data destruction.
     */
    (*f)->file_id 
        = H5Fcreate(filename, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT);
    CHECK_HERR_T((*f)->file_id);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(filename);

    if(*f != NULL && 1 == H5Iis_valid((*f)->file_id))
    {
        H5Fclose((*f)->file_id);
    }

    if(*f != NULL)
    {
        free(*f);
        f = NULL;
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOHDFopen_biohdf_file(const char* filename,
                       biohdf_open_mode mode,
                       /*OUT*/ biohdf_file *f)
{
    unsigned int   hdf5_mode; /* HDF5 access mode: read-only or read/write */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(filename);
    CHECK_INT_NE(filename, 0);

    /**CODE*****************************************************************/

    hdf5_mode = (READ_ONLY == mode) ? H5F_ACC_RDONLY : H5F_ACC_RDWR;

    /* Get memory for the file handle */
    *f = (biohdf_file)calloc(1, sizeof(struct _biohdf_file));
    CHECK_PTR(*f);

    /* Open or create file */
    (*f)->file_id = H5Fopen(filename, hdf5_mode, H5P_DEFAULT);
    CHECK_HERR_T((*f)->file_id);


    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(filename);

    if(*f != NULL && 1 == H5Iis_valid((*f)->file_id))
    {
        H5Fclose((*f)->file_id);
    }

    if(*f != NULL)
    {
        free(*f);
        f = NULL;
    }

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
BIOHDFclose_biohdf_file(/*IN-OUT*/ biohdf_file *f)
{
    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(*f);

    /**CODE*****************************************************************/

    if(1 == H5Iis_valid((*f)->file_id))
    {
        hdf_error = H5Fclose((*f)->file_id);
        CHECK_HERR_T(hdf_error);
    }

    free(*f);
    *f = NULL;
    
    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    free(*f);
    *f = NULL;

    return BIOHDF_BADNESS;
}

