/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_file.h"
#include "biohdf_file_internal.h"
#include "biohdf_group.h"
#include "biohdf_group_internal.h"



/* Write an error message in a standard format */
static void
_print_error_info(const char *group_name)
{
    fprintf(stderr, "             \tBioHDF object: group\n");
    if(NULL == group_name)
    {
        fprintf(stderr, "             \tgroup name: NULL");
    }
    else
    {
        fprintf(stderr, "             \tgroup name: %s", group_name);
    }
    fprintf(stderr, "\n\n");

    return;
}



BIOHDF_API biohdf_error
BIOHDFcreate_group(biohdf_file f,
                   const char *path)
{
    hid_t         group_id = -1;                     /* group ID (dummy) */

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(f->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    error = _create_or_open_group(f, path, &group_id);
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    hdf_error = H5Gclose(group_id);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);

    if(1 == H5Iis_valid(group_id))
    {
        H5Gclose(group_id);
    }

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Internal functions                                                        *
 *****************************************************************************/



BIOHDF_API biohdf_error
_parse_groups_from_path(const char *path,
                       /*OUT*/ char ***groups,
                       /*OUT*/ int *n_groups)
{
    size_t path_length  = 0;   /* length of the full path in characters */
    int    i            = 0;   /* iterator */

    /* Used when copying groups out of the path with strcpy() */
    size_t start        = 0;   /* The "element 0" index into the current group name in the path we're parsing */
    size_t length       = 0;   /* The length of the current group name in the path we're parsing */

    char **arr_ptr   = NULL;   /* A pointer to the array of group names */
    char *grp_ptr    = NULL;   /* A pointer to a specific group in the array of group names */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(path);

    /**CODE*****************************************************************/

    /* Uses a two-pass parse of the path.
     *    1) count the number of groups
     *    2) extract group names
     *
     * Also note that what we're parsing is basically units of 'group/' out
     * of the path, not units of '/group'.
     */

    path_length = strlen(path);
    CHECK_INT_NE(path_length, 0);

    /* We don't support relative paths */
    if('/' != path[0])
    {
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
        goto FAIL;
    }

    *n_groups   = 0;

    /* Walk the path, counting groups */
    /* path[0] is '/' so we don't need to test it */
    for(i = 1; i < (int)path_length; i++)
    {
        if('/' == path[i])
        {
            (*n_groups) += 1;
        }
    }

    /* +1 group if the terminal group is missing the last '/' */
    if(path[path_length - 1] != '/')
    {
        (*n_groups) += 1;
    }


    /* Create memory for group names array */
    (*groups) = (char **)calloc(*n_groups, sizeof(char *));
    CHECK_PTR(*groups);

    /* Initialize */
    start = 1;
    arr_ptr = *groups;

    
    /* Walk the path, assigning strings */
    /* path[0] is '/' so we don't need to test it */
    for(i = 1; i < (int)path_length; i++)
    {
        if('/' == path[i])
        {
            /* Store the current group */
            length = ((i-1) - start) + 1;

            grp_ptr = (char *)calloc(length + 1, sizeof(char));
            CHECK_PTR(grp_ptr);

            strncpy(grp_ptr, path + start, length);
            *arr_ptr = grp_ptr;

            /* Reset */
            start = i + 1;
            arr_ptr++;
        }
    }
    /* Handle the terminal group missing the last '/' */
    if(i == path_length && path[i - 1] != '/')
    {
        /* Store the current group */
        length = ((i-1) - start) + 1;
        
        grp_ptr = (char *)calloc(length + 1, sizeof(char));
        CHECK_PTR(grp_ptr);

        strncpy(grp_ptr, path + start, length);
        *arr_ptr = grp_ptr;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(path);


    if(*groups != NULL && *n_groups > 0)
    {
        for(i = 0; i < *n_groups; i++)
        {
            free((*groups)[i]);
            (*groups)[i] = NULL;
        }
        free(*groups);
    }

    *groups     = NULL;
    *n_groups   = -1;

    return BIOHDF_BADNESS;
}



BIOHDF_API biohdf_error
_free_groups_list(/*OUT*/ char ***groups,
                  int n_groups)
{
    int i;        /* iterator */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(*groups);
    CHECK_INT_GT(n_groups, -1);

    /**CODE*****************************************************************/

    for(i = 0; i < n_groups; i++)
    {
        free((*groups)[i]);
        (*groups)[i] = NULL;
    }

    free(*groups);
    *groups = NULL;

    /**SUCCESS***************************************************************/
    
    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    return BIOHDF_BADNESS;
}



biohdf_error
_create_or_open_group(biohdf_file f,
                      const char *path,
                      /*OUT*/ hid_t *group_id)
{

    char **groups = NULL;    /* Returned set of groups as strings */
    int n_groups;            /* Number of groups returned */
    int i;                   /* iterator */

    hid_t new_location = -1; /* The new group created/opened as we traverse
                              * the path.
                              */
    hid_t old_location = -1; /* The group in which the new group is created
                              * This starts at the file id on the first
                              * and is replaced with group IDs as we traverse
                              * the path.
                              */
    htri_t group_exists;     /* Whether or not a group exists */

    char *        bad_group        = NULL;           /* The group on which an operation failed */

    herr_t        hdf_error        = -1;             /* HDF5 error code */
    biohdf_error  error            = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(f->file_id);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    *group_id = -1;

    /* Get a list of the groups in the path */
    error = _parse_groups_from_path(path, &groups, &n_groups);
    CHECK_BIO(error);

    /* Iterate over the groups, creating them if they don't exist */
    old_location = f->file_id;

    for(i = 0; i < n_groups; i++)
    {
        /* Create or open the group */
        group_exists = H5Lexists(old_location, groups[i], H5P_DEFAULT);
        if(group_exists > 0)
        {
            new_location = H5Gopen(old_location, groups[i], H5P_DEFAULT);
            CHECK_HID_T(new_location);
        }
        else if(0 == group_exists)
        {
            new_location = H5Gcreate(old_location, groups[i],
                H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
            CHECK_HID_T(new_location);
        }
        else
        {
            bad_group = groups[i];
            BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);
            goto FAIL;
        }

        /* Close the old "containing" group, except on the first iteration
         * since we don't want to close the file.
         */
        if (H5I_GROUP == H5Iget_type(old_location))
        {
            hdf_error = H5Gclose(old_location);
            CHECK_HERR_T(hdf_error);
        }

        /* The old old/containing group is replaced with the new group */
        old_location = new_location;
    }

    /* Return the last group */
    *group_id = new_location;


    /**SUCCESS***************************************************************/

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(bad_group);

    if(1 == H5Iis_valid(old_location) && H5I_GROUP == H5Iget_type(old_location))
    {
        H5Gclose(old_location);
    }

    if(1 == H5Iis_valid(new_location))
    {
        H5Gclose(new_location);
    }
    
    *group_id = -1;

    /* Try our best to clean up memory */
    _free_groups_list(&groups, n_groups);

    return error;
}

