/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_utility.h"
#include "biohdf_utility_internal.h"




/*****************************************************************************
 * Free BioHDF memory (utility function)                                     *
 *****************************************************************************/



BIOHDF_API biohdf_error
BIOHDFfree_memory(/*IN-OUT*/ void **buffer)
{
    free(*buffer);
    (*buffer) = NULL;

    return BIOHDF_NO_ERROR;
}



/*****************************************************************************
 * Index properties                                                          *
 *****************************************************************************/



/** Create an index properties container */
BIOHDF_API biohdf_error
BIOHDFcreate_index_creation_properties(/*OUT*/ biohdf_index_creation_properties *props)
{
    /**CHECK*PARAMETERS*****************************************************/

    /* N/A */

    /**CODE*****************************************************************/

    *props = (biohdf_index_creation_properties)calloc(1, sizeof(struct _biohdf_index_creation_properties));
    CHECK_PTR(*props);

    (*props)->max_memory_to_use_mb = DEFAULT_INDEX_MAX_MEMORY_TO_USE_MB;

    (*props)->temp_file_path = (char *)calloc(strlen(DEFAULT_INDEX_TEMP_FILE_PATH) + 1, sizeof(char));
    CHECK_PTR((*props)->temp_file_path);

    strcpy((*props)->temp_file_path, DEFAULT_INDEX_TEMP_FILE_PATH);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    return BIOHDF_BADNESS;
}



/** Destroy an index properties container */
BIOHDF_API biohdf_error
BIOHDFdestroy_index_creation_properties(/*OUT*/ biohdf_index_creation_properties *props)
{
    if(*props)
    {
        if((*props)->temp_file_path)
        {
            free((*props)->temp_file_path);
        }
        free(*props);
        *props = NULL;
    }

    return BIOHDF_NO_ERROR;
}



/** Set index properties max memory */
BIOHDF_API biohdf_error
BIOHDFset_index_creation_properties_max_memory_mb(biohdf_index_creation_properties props,
                                                  int64_t max_memory_to_use_mb)
{
    props->max_memory_to_use_mb = max_memory_to_use_mb;

    /* Can't fail */
    return BIOHDF_NO_ERROR;
}



/** Set index properties path to temp files */
BIOHDF_API biohdf_error
BIOHDFset_index_creation_properties_temp_path(biohdf_index_creation_properties props,
                                              char *path)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(props);
    CHECK_PTR(path);

    /**CODE*****************************************************************/

    free(props->temp_file_path);
    props->temp_file_path = (char *)calloc(strlen(path) + 1, sizeof(char));
    CHECK_PTR(props->temp_file_path);

    strcpy(props->temp_file_path, path);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Create and read BioHDF type attributes                                    *
 *****************************************************************************/



biohdf_error
_write_biohdf_type_attribute(hid_t obj, const char *type_name)
{
    htri_t   attribute_exists    = -1;
    hid_t    space_id            = -1;
    hid_t    attr_id             = -1;
    hsize_t  dims[1];

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(obj);
    CHECK_PTR(type_name);
    CHECK_INT_NE(0, strlen(type_name));

    /**CODE*****************************************************************/

    /* Does the attribute exist?  Do NOT clobber if it does. */
    attribute_exists = H5Aexists(obj, BIOHDF_TYPE_ATTR_NAME);
    if(attribute_exists < 0)
    {
        /*HDF5 ERROR*/
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }
    if(attribute_exists > 0)
    {
        /* Attribute exists */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /* Write the attribute out */

    dims[0] = strlen(type_name); /* don't store the trailing \0 */
    space_id = H5Screate_simple (1, dims, NULL);
    CHECK_HID_T(space_id);

    attr_id = H5Acreate(obj, BIOHDF_TYPE_ATTR_NAME, H5T_NATIVE_CHAR, space_id, H5P_DEFAULT, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    hdf_error = H5Awrite(attr_id, H5T_NATIVE_CHAR, type_name);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Sclose(space_id);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(NULL == type_name)
    {
        fprintf(stderr, "             \tType name: NULL\n");
    }
    else
    {
        fprintf(stderr, "             \tType name: %s\n", type_name);
    }

    if(1 == H5Iis_valid(attr_id))
    {
        H5Aclose(attr_id);
    }

    if(1 == H5Iis_valid(space_id))
    {
        H5Aclose(space_id);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_read_biohdf_type_attribute(hid_t obj, /*OUT*/ char **type_name)
{
    htri_t  attribute_exists   = 0;
    hid_t   attr_id            = -1;
    size_t  size;

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(obj);

    /**CODE*****************************************************************/

    /* Does the attribute exist? */
    attribute_exists = H5Aexists(obj, BIOHDF_TYPE_ATTR_NAME);
    if(attribute_exists < 0)
    {
        *type_name = NULL;

        /*HDF5 ERROR*/
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }
    if(0 == attribute_exists)
    {
        *type_name = NULL;

        /* Attribute does not exist */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    attr_id = H5Aopen(obj, BIOHDF_TYPE_ATTR_NAME, H5P_DEFAULT);
    CHECK_HID_T(attr_id);

    size = (size_t)H5Aget_storage_size(attr_id);
    CHECK_INT_NE(size, 0);

    /* The calloc() is a little weird, but doing it this way avoids
     * errors if the data type is not a byte.
     */
    *type_name = (char *)calloc(1, size + 1);
    CHECK_PTR(*type_name);

    hdf_error = H5Aread(attr_id, H5T_NATIVE_CHAR, (void *)(*type_name));
    CHECK_HERR_T(hdf_error);


    /**SUCCESS***************************************************************/

    hdf_error = H5Aclose(attr_id);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(NULL == *type_name)
    {
        fprintf(stderr, "             \tType name: NULL\n");
    }
    else
    {
        fprintf(stderr, "             \tType name: %s\n", *type_name);
    }

    if(1 == H5Iis_valid(attr_id))
    {
        H5Aclose(attr_id);
    }

    return BIOHDF_BADNESS;
}
