/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "biohdf_vector.h"



/*****************************************************************************
 * Internal function declarations                                            *
 *****************************************************************************/



/* Assigns dataset names for vector part storage */
static biohdf_error
_add_dataset_names(const char *name, vector *v);



/* Write an error message in a standard format */
static void
_print_error_info(const char *vector_name)
{
    fprintf(stderr, "             \tBioHDF object: vector\n");
    if(NULL == vector_name)
    {
        fprintf(stderr, "             \tvector name: NULL");
    }
    else
    {
        fprintf(stderr, "             \tvector name: %s", vector_name);
    }
    fprintf(stderr, "\n\n");

    return;
}



/*****************************************************************************
 * Lifetime functions - create, open, close                                  *
 *****************************************************************************/



biohdf_error
_create_vector(hid_t parent,
               const char *name,
               hid_t hdf5_type,
               vector_storage_scheme scheme,
               size_t fixed_length_size,
               hsize_t chunk_size,
               compression_level compression,
               /*OUT*/ vector **v)
{

    hsize_t   size[1];           /* array size */
    hid_t     tid     = -1;      /* array type */

    herr_t        hdf_error = -1;             /* HDF5 error code */
    biohdf_error  error     = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(parent);
    CHECK_HID_T(hdf5_type);
    CHECK_PTR(name);
    CHECK_INT_NE(chunk_size, 0);

    if(FIXED_SIZE_ARRAY == scheme && 0 == fixed_length_size)
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**CODE*****************************************************************/

    /* Create the vector handle. */
    (*v) = NULL;
    (*v) = (vector *)calloc(1, sizeof(vector));
    CHECK_PTR((*v));

    /* Add the type size */
    (*v)->element_size = H5Tget_size(hdf5_type);
    CHECK_INT_GT((*v)->element_size, 0);

    /* Create dataset names */
    error = _add_dataset_names(name, *v);
    CHECK_BIO(error);

    /* Initialize other vector data members */
    (*v)->scheme             = scheme;
    (*v)->fixed_length_size  = fixed_length_size;
    (*v)->next_concat_index  = 0;

    (*v)->variable_length    = NULL;
    (*v)->fixed_length       = NULL;
    (*v)->index              = NULL;
    (*v)->overflow_flags     = NULL;
    
    /* Create vector storage */
    switch((*v)->scheme)
    {
    case FIXED_SIZE_ARRAY:

        /* Create the type */
        size[0] = fixed_length_size;
        tid     = H5Tarray_create(hdf5_type, 1, size);
        CHECK_HID_T(tid);

        /* Create fixed-length vector */
        error = _create_vpart(parent,
               (*v)->fixed_length_s,
               tid,
               chunk_size,
               compression,
               &((*v)->fixed_length));
        CHECK_BIO(error);

        hdf_error = H5Tclose(tid);
        CHECK_HERR_T(hdf_error);

        break;

    case VARIABLE_SIZE_ARRAY:

        /* Create concatenated vector */
        error = _create_vpart(parent,
               (*v)->variable_length_s,
               hdf5_type,
               chunk_size,
               compression,
               &((*v)->variable_length));
        CHECK_BIO(error);

        /* Create the index type */
        size[0] = fixed_length_size;
        
        tid = H5Tcreate(H5T_COMPOUND, sizeof(concat_indexes));
        CHECK_HID_T(tid);
        
        hdf_error = H5Tinsert(tid, "start", HOFFSET(concat_indexes, start), H5T_STD_I64LE);
        CHECK_HERR_T(hdf_error);
        
        hdf_error = H5Tinsert(tid, "end",   HOFFSET(concat_indexes, end),   H5T_STD_I64LE);
        CHECK_HERR_T(hdf_error);

        /* Create index vector */
        error = _create_vpart(parent,
               (*v)->index_s,
               tid,
               chunk_size,
               compression,
               &((*v)->index));
        CHECK_BIO(error);

        hdf_error = H5Tclose(tid);
        CHECK_HERR_T(hdf_error);

        break;

    case HYBRID_ARRAY:

        error = BIOHDF_BADNESS;
        goto FAIL;

        break;

    default:

        /* Shouldn't get here */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(name);

    if(H5Iis_valid(tid))
    {
        H5Tclose(tid);
    }

    if(NULL != *v)
    {
        _close_vector(v);
    }

    free(*v);
    *v = NULL;

    return BIOHDF_BADNESS;
}



biohdf_error
_open_vector(hid_t parent,
             const char *name,
             int is_read_only,
             /*OUT*/ vector **v)
{
    hid_t         hdf5_type = -1;

    biohdf_error  error     = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(parent);
    CHECK_PTR(name);

    /**CODE*****************************************************************/

    /* Create the vector handle. */
    (*v) = NULL;
    (*v) = (vector *)calloc(1, sizeof(vector));
    CHECK_PTR((*v));

    /* Create dataset names */
    error = _add_dataset_names(name, *v);
    CHECK_BIO(error);

    /* Initialize other vector data members */
    (*v)->variable_length  = NULL;
    (*v)->fixed_length     = NULL;
    (*v)->index            = NULL;
    (*v)->overflow_flags   = NULL;

    /* Determine vector storage scheme */
    if(H5Lexists(parent, (*v)->variable_length_s, H5P_DEFAULT) 
        && H5Lexists(parent, (*v)->index_s, H5P_DEFAULT))
    {
        /************************
         * variable-length data *
         ************************/

        (*v)->scheme = VARIABLE_SIZE_ARRAY;

        error = _open_vpart(parent, (*v)->variable_length_s,
            is_read_only, &((*v)->variable_length));
        CHECK_BIO(error);

        error = _open_vpart(parent, (*v)->index_s,
            is_read_only, &((*v)->index));
        CHECK_BIO(error);


        /* Add the type size */
        hdf5_type          = H5Dget_type(((*v)->variable_length)->dataset);
        CHECK_HID_T(hdf5_type);

        (*v)->element_size = H5Tget_size(hdf5_type);
        if((*v)->element_size < 1)
        {
            BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
            goto FAIL;
        }

        /* Set the dataset size */
        error = _get_vpart_count((*v)->variable_length, 
            &((*v)->next_concat_index));
        CHECK_BIO(error);
    }
    else if(H5Lexists(parent, (*v)->fixed_length_s,    H5P_DEFAULT) 
         && H5Lexists(parent, (*v)->variable_length_s, H5P_DEFAULT)
         && H5Lexists(parent, (*v)->overflow_flags_s,  H5P_DEFAULT))
    {
        /***************
         * hybrid data *
         ***************/

        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }
    else if(H5Lexists(parent, (*v)->fixed_length_s, H5P_DEFAULT))
    {
        /*********************
         * fixed-length data *
         *********************/

        (*v)->scheme = FIXED_SIZE_ARRAY;

        error = _open_vpart(parent, (*v)->fixed_length_s, 
            is_read_only, &((*v)->fixed_length));
        CHECK_BIO(error);

        /* Get the array width */
        error = _get_vpart_width((*v)->fixed_length, &((*v)->fixed_length_size));
        CHECK_BIO(error);
    }
    else
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(name);

    if(H5Iis_valid(hdf5_type))
    {
        H5Tclose(hdf5_type);
    }

    if(NULL != *v)
    {
        _close_vector(v);
    }

    free(*v);
    *v = NULL;

    return BIOHDF_BADNESS;
}



biohdf_error
_close_vector(vector **v)
{
    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* None - NULL input is ok to simplify upstream error processing */

    /**CODE*****************************************************************/

    if(NULL == *v){ return BIOHDF_NO_ERROR; }

    /* vector names */
    free((*v)->variable_length_s);
    free((*v)->fixed_length_s);
    free((*v)->index_s);
    free((*v)->overflow_flags_s);

    /* vector handles */
    error = _close_vpart(&(*v)->variable_length);
    CHECK_BIO(error);

    error = _close_vpart(&(*v)->fixed_length);
    CHECK_BIO(error);

    error = _close_vpart(&(*v)->index);
    CHECK_BIO(error);

    error = _close_vpart(&(*v)->overflow_flags);
    CHECK_BIO(error);

    free((*v)->name);

    /* And the vector handle */
    free(*v);
    *v = NULL;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*v) {
        _print_error_info((*v)->name);
    }
    else{
        _print_error_info(NULL);
    }

    /* Try again, but don't report errors
     * This may seem redundant, but I WANT close errors to be reported since
     * they may indicate underlying problems.
     */

    free((*v)->name);

    /* vector handles */
    _close_vpart(&(*v)->variable_length);
    _close_vpart(&(*v)->fixed_length);
    _close_vpart(&(*v)->index);
    _close_vpart(&(*v)->overflow_flags);

    /* And the vector handle */
    free(*v);
    *v = NULL;

    return error;
}



/*****************************************************************************
 * Data functions - read, write, count                                       *
 *****************************************************************************/



biohdf_error
_get_vector_count(const vector *v,
                  /*OUT*/ hsize_t *count)
{
    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);

    /**CODE*****************************************************************/

    switch(v->scheme)
    {
    case FIXED_SIZE_ARRAY:
        error = _get_vpart_count(v->fixed_length, count);
        CHECK_BIO(error);
        break;

    case VARIABLE_SIZE_ARRAY:
        error = _get_vpart_count(v->index, count);
        CHECK_BIO(error);
        break;

    case HYBRID_ARRAY:
        error = _get_vpart_count(v->fixed_length, count);
        CHECK_BIO(error);
        break;

    default:
        /* Shouldn't get here */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    *count = 0;

    return BIOHDF_BADNESS;
}



biohdf_error
_add_vector_element(vector *v,
                    const void* element,
                    hsize_t length)
{
    hssize_t      index_data[2];

    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(element);

    /* NOTE: Zero length elements are allowed! */

    /**CODE*****************************************************************/

    switch(v->scheme)
    {
    case FIXED_SIZE_ARRAY:

        /* Check for potential truncation */
        if(length > v->fixed_length_size)
        {
            BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
            goto FAIL;
        }

        error = _add_vpart_element(v->fixed_length, element);
        CHECK_BIO(error);

        /* Set the index of the last added element */
        v->index_of_last_added_element = v->fixed_length->data_count - 1;

        break;

    case VARIABLE_SIZE_ARRAY:

        if (0 == length)
        {
            index_data[0] = -1;
            index_data[1] = -1;
            error = _add_vpart_element(v->index, &index_data);
            CHECK_BIO(error);
        }
        else
        {
            /* Add index data */
            index_data[0] = v->next_concat_index;
            index_data[1] = (index_data[0] + length) - 1;
            error = _add_vpart_element(v->index, &index_data);
            CHECK_BIO(error);

            v->next_concat_index += length;

            /* Add data to concat vector */
            error = _add_vpart_block(v->variable_length, element, length);
            CHECK_BIO(error);
        }
        /* Set the index of the last added element */
        v->index_of_last_added_element = v->index->data_count - 1;

        break;

    case HYBRID_ARRAY:
        /* Not implemented */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;

    default:
        /* Shouldn't get here */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_get_vector_element(const vector *v,
                    hsize_t index,
                    int is_c_string,
                    /*OUT*/ void **element)
{
    concat_indexes *indexes = NULL; /* Returned start/end indexes from
                                     * the index dataset
                                     */

    biohdf_error  error = BIOHDF_BADNESS; /* BioHDF error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);

    /**CODE*****************************************************************/

    (*element) = NULL;

    switch(v->scheme)
    {
    case FIXED_SIZE_ARRAY:

        error = _get_vpart_block(v->fixed_length, index, index,
            is_c_string, element);
        CHECK_BIO(error);
        break;

    case VARIABLE_SIZE_ARRAY:
        
        /* Query index dataset to get indexes in the variable length dataset.
         * The index dataset is not a C string so it gets a value of zero (false).
         */
        error = _get_vpart_block(v->index, index, index,
             0, (void**)&indexes);
        CHECK_BIO(error);

        /* Read from concat vector block.
         * "empty" entries are indicated with start and end coordinates of -1.
         */
        if (indexes->start >= 0)
        {
            error = _get_vpart_block(v->variable_length, indexes->start,
                indexes->end, is_c_string, element);
            CHECK_BIO(error);
        }
        else if (is_c_string)
        {
            /* special case for zero length strings */
            (*element) = calloc(1, sizeof(char));
            CHECK_PTR(*element);

            strcpy((char *)*element, "\0");
        }
        else
        {
            /* The best we can do is return a NULL pointer for an "empty"
             * entry.
             */
            (*element) = NULL;
        }

        break;

    case HYBRID_ARRAY:
        
        /* Not implemented yet */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
        
        break;

    default:

        /* Shouldn't get here */
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    free(indexes);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    free(indexes);

    free(*element);
    *element = NULL;

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Internal function definitions                                             *
 *****************************************************************************/



static biohdf_error
_add_dataset_names(const char *name, vector *v)
{
    size_t name_length = 0;

    /* Standardized dataset suffixes */
    char concat_ext[]   = "_concatenated";
    char fixed_ext[]    = "_fixed_width";
    char index_ext[]    = "_concat_index";
    char overflow_ext[] = "_overflow_flags";

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(name);

    /**CODE*****************************************************************/

    v->name              = NULL;
    v->variable_length_s = NULL;
    v->fixed_length_s    = NULL;
    v->index_s           = NULL;
    v->overflow_flags_s  = NULL;

    name_length = strlen(name) + 1;
    v->name = (char *)calloc(name_length, sizeof(char));
    CHECK_PTR(v->name);
    strcpy(v->name, name);

    /* Create dataset names */
    name_length = strlen(name) + strlen(concat_ext) + 1;
    v->variable_length_s = (char *)calloc(name_length, sizeof(char));
    CHECK_PTR(v->variable_length_s);
    strcpy(v->variable_length_s, name);
    strcat(v->variable_length_s, concat_ext);

    name_length = strlen(name) + strlen(fixed_ext) + 1;
    v->fixed_length_s = (char *)calloc(name_length, sizeof(char));
    CHECK_PTR(v->fixed_length_s);
    strcpy(v->fixed_length_s, name);
    strcat(v->fixed_length_s, fixed_ext);
    
    name_length = strlen(name) + strlen(index_ext) + 1;
    v->index_s = (char *)calloc(name_length, sizeof(char));
    CHECK_PTR(v->index_s);
    strcpy(v->index_s, name);
    strcat(v->index_s, index_ext);
    
    name_length = strlen(name) + strlen(overflow_ext) + 1;
    v->overflow_flags_s = (char *)calloc(name_length, sizeof(char));
    CHECK_PTR(v->overflow_flags_s);
    strcpy(v->overflow_flags_s, name);
    strcat(v->overflow_flags_s, overflow_ext);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    if(NULL != v && NULL != name)
    {
        free(v->name);
        free(v->variable_length_s);
        free(v->fixed_length_s);
        free(v->index_s);
        free(v->overflow_flags_s);
    }

    return BIOHDF_BADNESS;
}

