/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#ifndef _BIOHDF_VECTOR_H
#define _BIOHDF_VECTOR_H

#include <stdlib.h>

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"
#include "biohdf_vpart.h"



/*****************************************************************************
 * A vector is a 1D growable arrays of data that can handle variable-size    *
 * elements.                                                                 *
 *****************************************************************************/



/* Indexes (start, end) into the 1D concatenated data */
typedef struct _concat_indexes
{
    hssize_t start;
    hssize_t end;
} concat_indexes;



/* The various variable-length storage schemes that we use:
 * - A fixed-size 2D array.
 * - Variable-length concatenated 1D array w/ separate index array
 * - Hybrid structure that primarily stores in 2D but can overflow to 1D
 */
typedef enum
{
    FIXED_SIZE_ARRAY,
    VARIABLE_SIZE_ARRAY,
    HYBRID_ARRAY,
} vector_storage_scheme;



/*****************************************************************************
 * Vector handle                                                             *
 *****************************************************************************/



typedef struct _vector
{
    /* The 'name' of the vector - used in error reporting */
    char *name;

    /* Size of the data element */
    size_t element_size;

    /* Fixed-length, variable-length or hybrid */
    vector_storage_scheme scheme;

    /* Width of the fixed-length data, if it exists */
    hsize_t fixed_length_size;

    /* Next position to be used in the concatenated dataset, if it exists */
    hsize_t next_concat_index;

    /* Internal vector part handles where the data is actually stored */
    vpart *variable_length;   /* Concatenated data */
    vpart *fixed_length;      /* Fixed-width data */
    vpart *index;             /* Indexes into the concatenated data */
    vpart *overflow_flags;    /* Overflow flags for the hybrid scheme */

    /* Internal vpart names */
    char *variable_length_s;   /* Concatenated data */
    char *fixed_length_s;      /* Fixed-width data */
    char *index_s;             /* Indexes into the concatenated data */
    char *overflow_flags_s;    /* Overflow flags for the hybrid scheme */

    /* The HDF5 offset/index of the last added or read element */
    /* TODO: come up with a better solution without the concurrency issues */
    hsize_t index_of_last_added_element;

} vector;



/*****************************************************************************
 * Lifetime functions - create, open, close                                  *
 *****************************************************************************/



/* Create vector storage of a given type - 1D, 2D or hybrid. */
biohdf_error
_create_vector(hid_t parent,
               const char *name,
               hid_t hdf5_type,
               vector_storage_scheme scheme,
               size_t fixed_length_size,
               hsize_t chunk_size,
               compression_level compression,
               /*OUT*/ vector **v);



/* Open vector storage */
biohdf_error
_open_vector(hid_t parent,
             const char *name,
             int is_read_only,
             /*OUT*/ vector **v);



/* Close vector storage. */
biohdf_error
_close_vector(vector **v);



/*****************************************************************************
 * Data functions - read, write, count                                       *
 *****************************************************************************/



/* Get the number of elements contained in a vector */
biohdf_error
_get_vector_count(const vector *v,
                  /*OUT*/ hsize_t *count);



/* Add an element to a vector. */
biohdf_error
_add_vector_element(vector *v,
                    const void* element,
                    hsize_t length);



/* Read an element from a vector */
biohdf_error
_get_vector_element(const vector *v,
                    hsize_t index,
                    int is_c_string,
                    /*OUT*/ void **element);



#endif
