/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "biohdf_vpart.h"



/*****************************************************************************
 * Internal function declarations                                            *
 *****************************************************************************/


/* Write an error message in a standard format */
static void
_print_error_info(const char *vpart_name)
{
    fprintf(stderr, "             \tBioHDF object: vector part (vpart)\n");
    if(NULL == vpart_name)
    {
        fprintf(stderr, "             \tvpart name: NULL");
    }
    else
    {
        fprintf(stderr, "             \tvpart name: %s", vpart_name);
    }
    fprintf(stderr, "\n\n");

    return;
}



/*****************************************************************************
 * Lifetime functions - create, open, close                                  *
 *****************************************************************************/



biohdf_error
_create_vpart(hid_t parent,
              const char *vpart_name,
              hid_t hdf5_type,
              hsize_t chunk_size,
              compression_level compression,
              /*OUT*/ vpart **v)
{

    hid_t   creation_space     = -1;          /* Dataset creation dataspace */
    hid_t   link_props         = H5P_DEFAULT; /* Link creation properties */
    hid_t   dset_create_props  = -1;          /* Dataset creation props */
    hid_t   dset_access_props  = H5P_DEFAULT; /* Dataset access properties */

    /* We're creating a 1D dataset with an initial size of 0 that can grow
     * to unlimited size.
     */
    int     rank = 1;                 /* # dimensions */
    hsize_t dims = 0;                 /* initial size */
    hsize_t maxdims = H5S_UNLIMITED;  /* max size */

    H5T_class_t type_class;           /* The class of the HDF5 data type - array, etc. */

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(parent);
    CHECK_HID_T(hdf5_type);
    CHECK_PTR(vpart_name);
    CHECK_INT_NE(chunk_size, 0);

    /* Check to make sure the dataset does not exist */
    if(H5Lexists(parent, vpart_name, H5P_DEFAULT))
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**CODE*****************************************************************/

    /* Create the vpart handle. */
    (*v) = NULL;
    (*v) = (vpart *)calloc(1, sizeof(vpart));
    CHECK_PTR((*v));

    (*v)->is_readonly            = 0;    /* ready for writing */
    (*v)->dataset                = -1;
    (*v)->file_dataspace         = -1;
    (*v)->hdf5_type              = -1;

    (*v)->data_count = 0;


    /* Create the dataset access properties list */
    dset_create_props = H5Pcreate(H5P_DATASET_CREATE);
    CHECK_HID_T(dset_create_props);

    /* HAS to be chunked due to unlimited size. */
    hdf_error = H5Pset_chunk(dset_create_props, rank, &chunk_size);
    CHECK_HERR_T(hdf_error);

    /* Set the compression level. */
    if(compression > 0)
    {
        hdf_error = H5Pset_deflate(dset_create_props, compression);
        CHECK_HERR_T(hdf_error);
    }

    /* Create the dataset creation dataspace */
    creation_space = H5Screate_simple(rank, &dims, &maxdims);
    CHECK_HID_T(creation_space);

    /* Create the dataset */
    (*v)->dataset = H5Dcreate(parent, 
        vpart_name, 
        hdf5_type, 
        creation_space, 
        link_props, 
        dset_create_props, 
        dset_access_props);
    CHECK_HID_T((*v)->dataset);

    /* Add the data type.  Copying it makes it our responsibility so we
     * can close it with H5Tclose() later.
     */
    (*v)->hdf5_type = H5Tcopy(hdf5_type);
    CHECK_HID_T((*v)->hdf5_type);

    (*v)->type_size_bytes = H5Tget_size(hdf5_type);
    CHECK_INT_NE((*v)->type_size_bytes, 0);

    /* Add the file dataspace */
    (*v)->file_dataspace = H5Dget_space((*v)->dataset);
    CHECK_HID_T((*v)->file_dataspace);


    /* Set the data width */
    type_class = H5Tget_class(hdf5_type);
    CHECK_INT_NE(type_class, H5T_NO_CLASS);

    if(H5T_ARRAY == type_class)
    {
        hdf_error = H5Tget_array_dims(hdf5_type, &((*v)->data_width));
        CHECK_HERR_T(hdf_error);
    }
    else
    {
        (*v)->data_width = 1;
    }

    /* Store the vpart name */
    (*v)->name = (char *)calloc(strlen(vpart_name) + 1, sizeof(char));
    CHECK_PTR((*v)->name);
    strcpy((*v)->name, vpart_name);


    /**SUCCESS***************************************************************/


    hdf_error = H5Sclose(creation_space);
    CHECK_HERR_T(hdf_error);

    hdf_error = H5Pclose(dset_create_props);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(vpart_name);

    /* We don't check HDF5 error codes in the cleanup section */

    if(1 == H5Iis_valid(creation_space))
    {
        H5Sclose(creation_space);
    }

    if(1 == H5Iis_valid(dset_create_props))
    {
        H5Pclose(dset_create_props);
    }

    if(NULL != *v && 1 == H5Iis_valid((*v)->dataset))
    {
        H5Dclose((*v)->dataset);
    }
    
    if(NULL != *v && 1 == H5Iis_valid((*v)->hdf5_type))
    {
        H5Tclose((*v)->hdf5_type);
    }

    if(NULL != *v && 1 == H5Iis_valid((*v)->file_dataspace))
    {
        H5Sclose((*v)->file_dataspace);
    }

    if(*v != NULL)
    {
        free((*v)->name);
    }
    free(*v);
    (*v) = NULL;

    return BIOHDF_BADNESS;
}



biohdf_error
_open_vpart(hid_t parent,
            char *vpart_name,
            int is_readonly,
            /*OUT*/ vpart **v)
{
    hid_t         type_id = 0;      /* The dataset type ID */
    H5T_class_t   type_class;       /* The class of the HDF5 data type - array, etc. */

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_HID_T(parent);
    CHECK_PTR(vpart_name);

    /* Check to make sure the dataset exists */
    if(!H5Lexists(parent, vpart_name, H5P_DEFAULT))
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**CODE*****************************************************************/

    (*v) = NULL;  /* So we can check for NULL when freeing in FAIL: */

    /* Create the vpart handle. */
    (*v) = (vpart *)calloc(1, sizeof(vpart));
    CHECK_PTR((*v));

    (*v)->is_readonly            = is_readonly;
    (*v)->dataset                = -1;
    (*v)->file_dataspace         = -1;
    (*v)->hdf5_type              = -1;

    /* Open the dataset. */
    (*v)->dataset = H5Dopen(parent,
        vpart_name,
        H5P_DEFAULT);
    CHECK_HID_T((*v)->dataset);

    /* Add the HDF5 dataset type and size to the vpart. */
    (*v)->hdf5_type = H5Dget_type((*v)->dataset);
    CHECK_HID_T((*v)->hdf5_type);

    (*v)->type_size_bytes = H5Tget_size((*v)->hdf5_type);
    CHECK_INT_NE((*v)->type_size_bytes, 0);

    /* Add the file dataspace */
    (*v)->file_dataspace = H5Dget_space((*v)->dataset);
    CHECK_HID_T((*v)->file_dataspace);

    /* Store the current size
     * There is no way to determine an error condition from this call.
     */
    (*v)->data_count = H5Sget_simple_extent_npoints((*v)->file_dataspace);

    /* Set the data width */
    type_id = H5Dget_type((*v)->dataset);
    CHECK_HID_T(type_id);

    type_class = H5Tget_class(type_id);
    CHECK_INT_NE(type_class, H5T_NO_CLASS);

    if(H5T_ARRAY == type_class)
    {
        hdf_error = H5Tget_array_dims(type_id, &((*v)->data_width));
        CHECK_HERR_T(hdf_error);
    }
    else
    {
        (*v)->data_width = 1;
    }

    /* Store the vpart name */
    (*v)->name = (char *)calloc(strlen(vpart_name) + 1, sizeof(char));
    CHECK_PTR((*v)->name);
    strcpy((*v)->name, vpart_name);


    /**SUCCESS***************************************************************/

    hdf_error = H5Tclose(type_id);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    _print_error_info(vpart_name);

    /* We don't check HDF5 error codes in the cleanup section */

    if(NULL != *v && H5Iis_valid((*v)->dataset))
    {
        H5Dclose((*v)->dataset);
    }
    
    if(NULL != *v && H5Iis_valid((*v)->hdf5_type))
    {
        H5Tclose((*v)->hdf5_type);
    }

    if(NULL != *v && H5Iis_valid((*v)->file_dataspace))
    {
        H5Sclose((*v)->file_dataspace);
    }

    if(H5Iis_valid(type_id))
    {
        H5Tclose(type_id);
    }

    if(*v != NULL)
    {
        free((*v)->name);
    }
    free(*v);
    (*v) = NULL;

    return BIOHDF_BADNESS;
}



biohdf_error
_close_vpart(/*OUT*/ vpart **v)
{
    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    /* None - NULL input is ok to simplify upstream error processing */

    /**CODE*****************************************************************/

    if(NULL == *v){ return BIOHDF_NO_ERROR; }

    if(H5Iis_valid((*v)->dataset)) 
    {
        hdf_error = H5Dclose((*v)->dataset);
        CHECK_HERR_T(hdf_error);
    }

    if(H5Iis_valid((*v)->file_dataspace))
    {
        hdf_error = H5Sclose((*v)->file_dataspace);
        CHECK_HERR_T(hdf_error);
    }

    if(H5Iis_valid((*v)->hdf5_type))
    {
        hdf_error = H5Tclose((*v)->hdf5_type);
        CHECK_HERR_T(hdf_error);
    }

    free((*v)->name);

    free(*v);
    (*v) = NULL;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(*v) {
        _print_error_info((*v)->name);
    }
    else
    {
        _print_error_info(NULL);
    }


    /* Try again, but don't report errors
     * This may seem redundant, but I WANT close errors to be reported since
     * they may indicate underlying problems.
     */
    if(H5Iis_valid((*v)->dataset))        { H5Dclose((*v)->dataset); }
    if(H5Iis_valid((*v)->file_dataspace)) { H5Sclose((*v)->file_dataspace); }
    if(H5Iis_valid((*v)->hdf5_type))      { H5Tclose((*v)->hdf5_type); }

    free((*v)->name);

    free(*v);
    (*v) = NULL;

    return BIOHDF_BADNESS;
}



/*****************************************************************************
 * Data functions - read, append                                             *
 *****************************************************************************/



biohdf_error
_get_vpart_width(const vpart *v, /*OUT*/ hsize_t *width)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(width);

    /**CODE*****************************************************************/

    *width = v->data_width;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_get_vpart_count(const vpart *v, /*OUT*/ hsize_t *count)
{
    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(count);

    /**CODE*****************************************************************/

    *count = v->data_count;

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_add_vpart_element(vpart *v,
                   const void *element)
{
    biohdf_error    error = BIOHDF_BADNESS;

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(element);

    /**CODE*****************************************************************/

    error = _add_vpart_block(v, element, 1);
    CHECK_BIO(error);

    /**SUCCESS***************************************************************/

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_add_vpart_block(vpart *v,
                 const void *block,
                 hsize_t n_elements)
{
    int      rank;
    hsize_t  start;
    hid_t    memory_space = -1;

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_PTR(block);
    CHECK_INT_NE(v->is_readonly, 1);    /* Can't be read-only */
    CHECK_INT_NE(n_elements, 0);        /* No point in adding 0 elements */

    /**CODE*****************************************************************/

    /* Set where we'll start writing */
    start = v->data_count;

    /* Adjust the dataset to the new size */
    v->data_count += n_elements;
    hdf_error = H5Dset_extent(v->dataset, &(v->data_count));
    CHECK_HERR_T(hdf_error);

    /* Replace the file dataspace with one that reflects the new size */
    hdf_error = H5Sclose(v->file_dataspace);
    CHECK_HERR_T(hdf_error);

    v->file_dataspace = H5Dget_space(v->dataset);
    CHECK_HID_T(v->file_dataspace);

    /* Select hyperslab from file dataspace */
    hdf_error = H5Sselect_hyperslab(v->file_dataspace,
        H5S_SELECT_SET,       /* new selection (not adjusting an old one) */
        &start,               /* start */
        NULL,                 /* stride, NULL implies all elements */
        &n_elements,          /* count */
        NULL);                /* block, NULL implies all elements */
    CHECK_HERR_T(hdf_error);

    /* Create memory dataspace */
    rank = 1;
    memory_space = H5Screate_simple(rank, &n_elements, &n_elements);
    CHECK_HID_T(memory_space);

    /* Write data to the dataset */
    hdf_error = H5Dwrite(v->dataset,
        v->hdf5_type,
        memory_space,
        v->file_dataspace,
        H5P_DEFAULT,
        block);
    CHECK_HERR_T(hdf_error);

    /**SUCCESS***************************************************************/

    hdf_error = H5Sclose(memory_space);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }

    if(H5Iis_valid(memory_space))
    {
        H5Sclose(memory_space);
    }

    return BIOHDF_BADNESS;
}



biohdf_error
_get_vpart_block(const vpart *v,
                 hsize_t start,
                 hsize_t end,
                 int is_c_string,
                 /*OUT*/ void **block)
{
    int rank;                       /* # vpart dimensions (one) */

    hsize_t selection_start[1];     /* HDF5 dataspace variables */
    hsize_t *selection_stride;
    hsize_t selection_count[1];
    hsize_t *selection_block;

    hid_t memory_space = -1;        /* HDF5 memory dataspace */

    size_t buffer_size;             /* size of the results in memory */

    herr_t        hdf_error        = -1;             /* HDF5 error code */

    /**CHECK*PARAMETERS*****************************************************/

    CHECK_PTR(v);
    CHECK_INT_LE(start, end);
    if (start > end)
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**CODE*****************************************************************/

    (*block) = NULL;
    rank = 1;

    selection_start[0] = start ;
    selection_stride   = NULL;  /* NULL implies we're selecting all */
    selection_count[0] = (end - start) + 1;
    selection_block    = NULL;   /* NULL means 1 element in all directions */

    /* Select a subset of the dataset for reading
     * H5S_SELECT_SET implies a new selection (not adjusting an old one)
     */
    hdf_error = H5Sselect_hyperslab(v->file_dataspace,
        H5S_SELECT_SET,
        selection_start,
        selection_stride,
        selection_count,
        selection_block);
    CHECK_HERR_T(hdf_error);

    /* Describe how the dataset selection will be mapped into memory
     * We know the size so both initial and max size are the same
     */
    memory_space = H5Screate_simple(rank,
        selection_count,
        selection_count);
    CHECK_HID_T(memory_space);

    /* Create buffer for data results */
    buffer_size = v->type_size_bytes;
    if(is_c_string)
    {
        buffer_size += 1;
    }
    *block = calloc((size_t)(selection_count[0]), buffer_size);
    CHECK_PTR(*block);

    /* Read the data from the dataset into memory */
    hdf_error = H5Dread(v->dataset,
        v->hdf5_type,
        memory_space,
        v->file_dataspace,
        H5P_DEFAULT,          /* default xfer properties */
        *block);
    CHECK_HERR_T(hdf_error);

    /* If this is a bunch of chars that represents a C string, it needs
     * to be terminated with a nul.
     */
    if(is_c_string)
    {
        ((char *)(*block))[(buffer_size * selection_count[0]) - 1] = '\0';
    }

    /**SUCCESS***************************************************************/

    hdf_error = H5Sclose(memory_space);
    CHECK_HERR_T(hdf_error);

    return BIOHDF_NO_ERROR;

    /**FAILURE***************************************************************/

    FAIL:

    if(v) {
        _print_error_info(v->name);
    }
    else
    {
        _print_error_info(NULL);
    }


    (*block) = NULL;

    if(H5Iis_valid(memory_space))
    {
        H5Sclose(memory_space);
    }

    return BIOHDF_BADNESS;
}

