/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#ifndef _BIOHDF_VPART_H
#define _BIOHDF_VPART_H

#include "hdf5.h"

#include "biohdf_api.h"
#include "biohdf_error.h"



/*****************************************************************************
 * A vpart is a 1D growable arrays of fixed-size elements.                   *
 * It is the underlying unit of a BioHDF vector.                             *
 *****************************************************************************/



/*****************************************************************************
 * Vector part handle                                                        *
 *****************************************************************************/



typedef struct _vpart
{
    int       is_readonly;

    hid_t     dataset;
    
    hid_t     file_dataspace;
    
    hid_t     hdf5_type;

    size_t    type_size_bytes;

    hsize_t   data_count;

    hsize_t   data_width;  /* 1 for atomic types, n for arrays of size n */

    char      *name;   /* Used for debugging output */
} vpart;




/*****************************************************************************
 * Lifetime functions - create, open, close                                  *
 *****************************************************************************/



/* Create 1D HDF storage */
biohdf_error
_create_vpart(hid_t parent,
              const char *vpart_name,
              hid_t hdf5_type,
              hsize_t chunk_size,
              compression_level compression,
              /*OUT*/ vpart **v);



/* Open vpart storage. */
biohdf_error
_open_vpart(hid_t parent,
            char *vpart_name,
            int is_readonly,
            /*OUT*/ vpart **v);



/* Close vpart storage. */
biohdf_error
_close_vpart(/*OUT*/ vpart **v);




/*****************************************************************************
 * Data functions - read, append                                             *
 *****************************************************************************/



/* Get the width of the vpart - 1 for atomic types, n for arrays of width n */
biohdf_error
_get_vpart_width(const vpart *v, /*OUT*/ hsize_t *width);



/* Get the size of the vpart */
biohdf_error
_get_vpart_count(const vpart *v, /*OUT*/ hsize_t *count);



/* Append a single element to a vpart. */
biohdf_error
_add_vpart_element(vpart *v,
                   const void *element);



/* Append a block of elements to a vpart. */
biohdf_error
_add_vpart_block(vpart *v,
                 const void *block,
                 hsize_t n_elements);



/* Read a block of elements from a vpart */
biohdf_error
_get_vpart_block(const vpart *v,
                 hsize_t start,
                 hsize_t end,
                 int is_c_string,
                 /*OUT*/ void **block);




#endif
