/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "biohdf.h"
#include "bioh5g.h"



/*****************************************************************************
 * Main test program                                                         *
 *****************************************************************************/


int main(int argc, char **argv) 
{
    biohdf_file         file    = NULL;
    bioh5g_alignments   aligns  = NULL;
    bioh5g_reads        reads   = NULL;

    bioh5g_alignments_iterator  iter    = NULL;

    bioh5g_alignments_creation_properties   a_props = NULL;
    bioh5g_reads_properties                 r_props = NULL;
    biohdf_index_creation_properties        i_props = NULL;

    bioh5g_alignments_format    header_format;

    bioh5g_alignment_data   *a_data     = NULL;

    int n_alignments = 0; /* # alignments in collection */

    //int32_t mask = 0; /* SAM flags mask */

    int64_t count   = -1;
    int64_t index   = -1;

    char filename[]     = "test_alignments.h5";
    char reads_path[]   = "/R";
    char aligns_path[]  = "/A";
    char header_in[]    = "This is a header";

    char *header_out    = NULL;
    char *test_path     = NULL;

    FILE            *test   = NULL;           /* For testing file existence */

    int             stdio_error = -1;             /* C stdio error code */
    biohdf_error    error       = BIOHDF_BADNESS; /* BioHDF error code */

    /**CODE*****************************************************************/

    /************************
     * Create the test file *
     ************************/

    if((test = fopen(filename, "r")))
    {
        fclose(test);
        remove(filename);
    }

    error = BIOHDFcreate_biohdf_file(filename, &file);
    CHECK_BIO(error);

    /************************************
     * Test creating alignments storage *
     ************************************/

    /* Create reads properties */
    error = BIOH5Gcreate_reads_properties(&r_props);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_sequences_scheme(r_props, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_identifiers_scheme(r_props, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_reads_type(r_props, BASE_SPACE);
    CHECK_BIO(error);

    /* Create the collection */
    error = BIOH5Gcreate_reads_collection(file,
        r_props,
        reads_path,
        &reads);
    CHECK_BIO(error);

    error = BIOH5Gdestroy_reads_properties(&r_props);
    CHECK_BIO(error);

    /* Create alignments properties */
    error = BIOH5Gcreate_alignments_properties(&a_props);
    CHECK_BIO(error);
    error = BIOH5Gset_alignments_properties_cigar_scheme(a_props, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_alignments_properties_refs_scheme(a_props, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_alignments_properties_tags_scheme(a_props, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_alignments_properties_reads_path(a_props, reads_path);
    CHECK_BIO(error);

    /* Create the collection */
    error = BIOH5Gcreate_alignments_collection(file, a_props, aligns_path, &aligns);
    CHECK_BIO(error);

    error = BIOH5Gdestroy_alignments_properties(&a_props);
    CHECK_BIO(error);


    /***********************************
     * Test closing alignments storage *
     ***********************************/

    error = BIOH5Gclose_reads_collection(&reads);
    CHECK_BIO(error);

    error = BIOH5Gclose_alignments_collection(&aligns);
    CHECK_BIO(error);


    /***********************************
     * Test opening alignments storage *
     ***********************************/

    error = BIOH5Gopen_reads_collection(file, reads_path, READ_WRITE, &reads);
    CHECK_BIO(error);

    error = BIOH5Gopen_alignments_collection(file, aligns_path, READ_WRITE, &aligns);
    CHECK_BIO(error);

    /*******************************
     * Test getting the reads path *
     *******************************/

    error = BIOH5Gget_reads_path(aligns, &test_path);
    CHECK_BIO(error);
    CHECK_PTR(test_path);
    if(strcmp(test_path, reads_path))
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }
    free(test_path);


    /******************************
     * Get the number of elements *
     ******************************/

    error = BIOH5Gget_alignments_count(aligns, &count);
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 0);


    /**************************************************
     * Attempt to index when there are no stored data *
     **************************************************/

    error = BIOHDFcreate_index_creation_properties(&i_props);
    CHECK_BIO(error);
    error = BIOHDFset_index_creation_properties_max_memory_mb(i_props, 1024);
    CHECK_BIO(error);
    error = BIOHDFset_index_creation_properties_temp_path(i_props, ".");
    CHECK_BIO(error);

    error = BIOH5Gcreate_alignments_index(aligns, REF_POS_SECONDARY, i_props);
    CHECK_BIO(error);

    /* NOTE: Don't destroy properties - needed below... */


    /***************************************
     * Test iterators on empty collections *
     ***************************************/

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;
    }

    /* Zero alignments should have been returned */
    CHECK_INT_EQ(count, 0);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /*******************************
     * Test adding some alignments *
     *******************************/

    /* Create an empty alignment */
    a_data = (bioh5g_alignment_data *)calloc(1, sizeof(bioh5g_alignment_data));
    a_data->length = 0;
    a_data->position = 0;
    a_data->read_index = 0;
    a_data->reference = (char *)calloc(9, sizeof(char));
    a_data->sam_cigar = (char *)calloc(1, sizeof(char));
    a_data->sam_flags = 0;
    a_data->sam_mapq = 0;
    a_data->sam_pnext = 0;
    a_data->sam_rnext = (char *)calloc(1, sizeof(char));
    a_data->sam_tags = (char *)calloc(1, sizeof(char));
    a_data->sam_tlen = 0;


    /* Add test alignments.  It takes at least six to cover all the spanning
     * cases.  Remember that all positions are 1-indexed as far as BioHDF
     * is concerned, but that HDF5 offsets are zero based.
     *
     *                   1         2         3
     *         01234567890123456789012345678901234  <-- HDF5 offset
     *         ---------[---INTERVAL---]----------
     * 1        (*****)
     * 2              (*****)
     * 3                     (*****)
     * 15                    (****)
     * 4                            (*****)
     * 5                                   (*****)
     * 6            (***********************)
     *
     * Predicted index order:
     *    1,6,2,3 and 15,4,5    (above numbering scheme)
     *    11,13,1,12,2,8        (by HDF5 offset as inserted, below)
     *
     * We also need eight alignments to test endpoints:
     *
     *                1         2         3
     *      0123456789012345678901234567890    <-- HDF5 offset
     *      ---------[------------]--------
     * 7      (*****)
     * 8       (*****)
     * 9             (*****)
     * 10             (*****)
     *
     * 11                  (*****)
     * 12                   (*****)
     * 13                         (*****)
     * 14                          (*****)
     *
     * Predicted index order:
     *     7,8,9,10,11,12,13,14    (above numbering scheme)
     *     6,4,3,9,10,0,5,7        (by HDF5 offset as inserted, below)
     *
     * The test interval will cover positions 10-20, inclusive
     * All alignments are of length 5 (except #6 & #15) to simplify things.
     *
     * The data are inserted out of order so they can be sorted later.
     *
     * These also have a smattering of SAM flags and MAPQ values
     * for later queries.
     *
     * I'm borrowing the tlen field to use as a flag for elements
     * that are returned by a range query over the indicated
     * interval.
     */

    /* Set the number of alignments for later use */
    n_alignments = 15;

    /* Set the length for all the alignments (except #6) */
    a_data->length = 5;

    /* Add alignments in random order */

    /* 12 */
    a_data->position = 16;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 255;
    a_data->sam_flags = 0x4;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 2 */
    a_data->position = 8;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 123;
    a_data->sam_flags = 0x4 | 0x8;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 4 */
    a_data->position = 18;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 122;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 9 */
    a_data->position = 10;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 8 */
    a_data->position = 6;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 13 */
    a_data->position = 20;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 7 */
    a_data->position = 5;
    a_data->sam_tlen = 0;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 14 */
    a_data->position = 21;
    a_data->sam_tlen = 0;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 5 */
    a_data->position = 23;
    a_data->sam_tlen = 0;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 10 */
    a_data->position = 11;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 11 */
    a_data->position = 15;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "7-14");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 1 */
    a_data->position = 3;
    a_data->sam_tlen = 0;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 3 */
    a_data->position = 12;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 6 - special size, so it goes last */
    a_data->length = 18;
    a_data->position = 7;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    /* 15 - special size, so it goes last */
    a_data->length = 4;
    a_data->position = 12;
    a_data->sam_tlen = 1;
    a_data->sam_mapq = 0;
    a_data->sam_flags = 0;
    memset(a_data->reference, 0, 9 * sizeof(char));
    strcpy(a_data->reference, "1-6");
    error = BIOH5Gadd_alignment(aligns, a_data);
    CHECK_BIO(error);

    error = BIOH5Gfree_alignment_data(&a_data);
    CHECK_BIO(error);


    /************************************
     * Test iterators on unindexed data *
     ************************************/

    /* NOTE: The previous creation of an index on an empty dataset will leave
     * the index type as "unsorted".  If this changes in the future, these tests
     * will not adequately evaluate iteration over unsorted data.
     */


    /*****************************************
     * Check range filter for alignments 1-6 *
     *****************************************/

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_range_filter(iter, "1-6", 10, 20);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);

        count++;

        /* Alignments that should be returned have sam_tlen set to 1 */
        if(0 == a_data->sam_tlen)
        {
            BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
            goto FAIL;
        }

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* Five alignments should have been returned */
    CHECK_INT_EQ(count, 5);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /******************************************
     * Check range filter for alignments 7-14 *
     ******************************************/

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_range_filter(iter, "7-14", 10, 20);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);

        count++;

        /* Alignments that should be returned have sam_tlen set to 1 */
        if(0 == a_data->sam_tlen)
        {
            BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
            goto FAIL;
        }

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* Six alignments should have been returned */
    CHECK_INT_EQ(count, 6);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /********************************
     * Check multiple range filters *
     ********************************/

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_range_filter(iter, "1-6", 10, 20);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_range_filter(iter, "7-14", 10, 20);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);

        count++;

        /* Alignments that should be returned have sam_tlen set to 1 */
        if(0 == a_data->sam_tlen)
        {
            BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
            goto FAIL;
        }

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* Eleven alignments should have been returned */
    CHECK_INT_EQ(count, 11);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /* *******************
     * Check MAPQ filter *
     *********************/

    /* Min MAPQ = 123 */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_mapq_filter(iter, 123);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* Two alignments should have been returned */
    CHECK_INT_EQ(count, 2);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /* Min MAPQ = 255 */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_mapq_filter(iter, 255);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* One alignment should have been returned */
    CHECK_INT_EQ(count, 1);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /* Min MAPQ = 0 (should return all alignments */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_mapq_filter(iter, 0);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* All alignments should have been returned */
    CHECK_INT_EQ(count, n_alignments);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /**************************
     * Check SAM flags filter *
     **************************/

    /* Check 0x4 set */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_flags_filter(iter, 0x4);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* Two alignments should have been returned */
    CHECK_INT_EQ(count, 2);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /* Check 0x4 and 0x8 set */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_flags_filter(iter, 0x4 | 0x8);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* One alignment should have been returned */
    CHECK_INT_EQ(count, 1);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);

    /* Check 0x10 set (should not return anything */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_flags_filter(iter, 0x10);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* No alignments should have been returned */
    CHECK_INT_EQ(count, 0);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /* Check mask of 0 (return all alignments) */

    error = BIOH5Gcreate_alignments_iterator(aligns, &iter);
    CHECK_BIO(error);
    error = BIOH5Gadd_alignments_iterator_flags_filter(iter, 0);
    CHECK_BIO(error);

    count = 0;
    while((error = BIOH5Gget_next_alignment(iter, &index, &a_data)) != BIOHDF_BADNESS
        && a_data != NULL)
    {
        CHECK_BIO(error);
        count++;

        error = BIOH5Gfree_alignment_data(&a_data);
        CHECK_BIO(error);
    }

    /* All alignments should have been returned */
    CHECK_INT_EQ(count, n_alignments);

    error = BIOH5Gdestroy_alignments_iterator(&iter);
    CHECK_BIO(error);


    /*************************************
     * Create an index on the alignments *
     *************************************/

    /* Re-use the properties from before */

    error = BIOH5Gcreate_alignments_index(aligns, REF_POS_SECONDARY, i_props);
    CHECK_BIO(error);

    error = BIOHDFdestroy_index_creation_properties(&i_props);
    CHECK_BIO(error);

    /******************************
     * Get the number of elements *
     ******************************/

    error = BIOH5Gget_alignments_count(aligns, &count);
    CHECK_BIO(error);
    CHECK_INT_EQ(count, n_alignments);

    /**********************************************
     * Get some elements by index to test sorting *
     **********************************************/


    /********************
     * Store/get header *
     ********************/

    error = BIOH5Gstore_alignment_file_header(aligns, SAM_FORMAT, header_in);
    CHECK_BIO(error);
    error = BIOH5Gget_alignment_file_header(aligns, &header_format, &header_out);
    CHECK_BIO(error);
    CHECK_INT_EQ((int)header_format, (int)SAM_FORMAT);
    if(strcmp(header_in, header_out))
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    /**SUCCESS***************************************************************/

    error = BIOH5Gclose_reads_collection(&reads);
    CHECK_BIO(error);
    error = BIOH5Gclose_alignments_collection(&aligns);
    CHECK_BIO(error);
    error = BIOHDFclose_biohdf_file(&file);
    CHECK_BIO(error);

    stdio_error = remove(filename);
    CHECK_INT_EQ(stdio_error, 0);

    return 0;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** TEST FAILURE ***   Test: Alignments (test/test_alignment.c)\n");

    BIOH5Gdestroy_alignments_properties(&a_props);
    BIOH5Gdestroy_reads_properties(&r_props);
    BIOHDFdestroy_index_creation_properties(&i_props);

    BIOH5Gdestroy_alignments_iterator(&iter);

    BIOH5Gfree_alignment_data(&a_data);

    BIOH5Gclose_reads_collection(&reads);
    BIOH5Gclose_alignments_collection(&aligns);
    BIOHDFclose_biohdf_file(&file);

    free(header_out);
    free(test_path);

    /* The temp file is left in place for debugging on errors.  It will
     * get cleaned up when the program is run again.
     */

    return 1;
}
