/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf.h"
#include "biohdf_file_internal.h"
#include "biohdf_group_internal.h"


/* Check to ensure a C string is not NULL and matches a test value. */
#define CHECK_STRING(val, test)                                             \
do                                                                          \
{                                                                           \
    if(NULL == (val) || NULL == (test) || strcmp((val), (test))){           \
        BIOHDFemit_error_message(__FILE__, __func__, (int)__LINE__);        \
        goto FAIL;                                                          \
    }                                                                       \
}while(0)



/*****************************************************************************
 * Main test program                                                         *
 *****************************************************************************/



int
main(int argc, char **argv) 
{
    char         **groups          = NULL;               /* array of group strings */
    int          n_groups          = 0;                  /* # of groups found in a path */
    htri_t       exists            = -1;                 /* Existence marker for groups */

    char         filename[]        = "test_groups.h5";   /* Test file name */
    biohdf_file  file              = NULL;               /* BioHDF file handle */
    FILE         *test             = NULL;               /* For testing file existence */

    /* Various test paths */
    char         group_path[]      = "/path/to/a/group/";
    char         collection_path[] = "/path/to/a/collection";

    biohdf_error error             = BIOHDF_BADNESS; /* BioHDF error code */

    /**CODE*****************************************************************/

    /**********************
     * Test group parsing *
     **********************/

    error = _parse_groups_from_path("/foo/", &groups, &n_groups);
    CHECK_BIO(error);
    CHECK_INT_EQ(n_groups, 1);
    CHECK_STRING(*(groups + 0), "foo");

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    error = _parse_groups_from_path("/bar", &groups, &n_groups);
    CHECK_BIO(error);
    CHECK_INT_EQ(n_groups, 1);
    CHECK_STRING(*(groups + 0), "bar");

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    error = _parse_groups_from_path("/foo/bar", &groups, &n_groups);
    CHECK_BIO(error);
    CHECK_INT_EQ(n_groups, 2);
    CHECK_STRING(*(groups + 0), "foo");
    CHECK_STRING(*(groups + 1), "bar");

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    error = _parse_groups_from_path("/path/to/a/group/", &groups, &n_groups);
    CHECK_BIO(error);
    CHECK_INT_EQ(n_groups, 4);
    CHECK_STRING(*(groups + 0), "path");
    CHECK_STRING(*(groups + 1), "to");
    CHECK_STRING(*(groups + 2), "a");
    CHECK_STRING(*(groups + 3), "group");

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    error = _parse_groups_from_path("/path/to/a/collection", &groups, &n_groups);
    CHECK_BIO(error);
    CHECK_INT_EQ(n_groups, 4);
    CHECK_STRING(*(groups + 0), "path");
    CHECK_STRING(*(groups + 1), "to");
    CHECK_STRING(*(groups + 2), "a");
    CHECK_STRING(*(groups + 3), "collection");

    error = _free_groups_list(&groups, n_groups);
    CHECK_BIO(error);

    /************************
     * Create the test file *
     ************************/

    if((test = fopen(filename, "r")))
    {
        fclose(test);
        remove(filename);
    }

    error = BIOHDFcreate_biohdf_file(filename, &file);
    CHECK_BIO(error);


    /***********************
     * Test group creation *
     ***********************/

    error = BIOHDFcreate_group(file, group_path);
    CHECK_BIO(error);
    error = BIOHDFcreate_group(file, group_path);           /* on purpose! (should pass) */
    CHECK_BIO(error);
    error = BIOHDFcreate_group(file, collection_path);
    CHECK_BIO(error);


    /************************
     * Test group existence *
     ************************/

    exists = H5Lexists(file->file_id, "/path/", H5P_DEFAULT);
    if (exists < 1)
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    exists = H5Lexists(file->file_id, "/path/to/a/group/", H5P_DEFAULT);
    if (exists < 1)
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }

    exists = H5Lexists(file->file_id, "/path/to/a/collection/", H5P_DEFAULT);
    if (exists < 1)
    {
        BIOHDFemit_error_message(__FILE__, __func__, __LINE__);
        goto FAIL;
    }



    /**SUCCESS***************************************************************/

    error = BIOHDFclose_biohdf_file(&file);
    CHECK_BIO(error);

    remove(filename);

    return 0;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** TEST FAILURE ***   Test: Groups (test/test_groups.c)\n");

    if(file != NULL)
    {
        BIOHDFclose_biohdf_file(&file);
    }

    /* The temp file is left in place for debugging on errors.  It will
     * get cleaned up when the program is run again.
     */

    /* The group list memory may have gotten corrupted in a spectacular way
     * so we just let it get cleaned up when the process exits instead of
     * potentially making a bigger mess when we try to free it.
     */

    return 1;
}
