/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>


#include "biohdf.h"
#include "bioh5g.h"



/*****************************************************************************
 * Main test program                                                         *
 *****************************************************************************/



int
main(int argc, char **argv) 
{
    biohdf_file               file;
    bioh5g_reads              reads;
    bioh5g_reads              var_reads;
    bioh5g_reads_properties   properties;
    bioh5g_reads_properties   var_properties;

    bioh5g_read_data          in;
    bioh5g_read_data          *out = NULL;
    int64_t                   count;
    int64_t                   i;

    bioh5g_reads_iterator     iter = NULL;

    char filename[]      = "test_reads.h5";
    char fixed_path[]    = "/fixed/";
    char variable_path[] = "/variable/";

    FILE         *test             = NULL;               /* For testing file existence */

    biohdf_error error             = BIOHDF_BADNESS; /* BioHDF error code */

    /**CODE*****************************************************************/

    /************************
     * Create the test file *
     ************************/

    if((test = fopen(filename, "r")))
    {
        fclose(test);
        remove(filename);
    }

    error = BIOHDFcreate_biohdf_file(filename, &file);
    CHECK_BIO(error);


    /***********************************
     * Test fixed-length reads storage *
     ***********************************/

    /* Create reads properties for fake data */
    error = BIOH5Gcreate_reads_properties(&properties);
    CHECK_BIO(error);

    error = BIOH5Gset_reads_properties_sequences_scheme(properties, FIXED_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_sequences_length(properties, 10);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_identifiers_scheme(properties, FIXED_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_identifiers_length(properties, 10);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_reads_type(properties, BASE_SPACE);
    CHECK_BIO(error);

    /* Create reads storage */
    error = BIOH5Gcreate_reads_collection(file,
        properties,
        fixed_path,
        &reads);
    CHECK_BIO(error);

    error = BIOH5Gdestroy_reads_properties(&properties);
    CHECK_BIO(error);

    /* Store some reads */
    in.sequence        = "0123456789";
    in.quality_values  = "9876543210";
    in.identifier      = "abcdefghij";
    error = BIOH5Gadd_read(reads, &in);
    CHECK_BIO(error);

    in.sequence        = "abcdefghij";
    in.quality_values  = "jihgfedcba";
    in.identifier      = "0123456789";
    error = BIOH5Gadd_read(reads, &in);
    CHECK_BIO(error);

    in.sequence        = "ABCDEFGHIJ";
    in.quality_values  = "JIHGFEDCBA";
    in.identifier      = "0123456789";
    error = BIOH5Gadd_read(reads, &in);
    CHECK_BIO(error);

    error = BIOH5Gclose_reads_collection(&reads);
    CHECK_BIO(error);


    /**************************************
     * Test variable-length reads storage *
     **************************************/

    error = BIOH5Gcreate_reads_properties(&var_properties);
    CHECK_BIO(error);

    error = BIOH5Gset_reads_properties_sequences_scheme(var_properties, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_identifiers_scheme(var_properties, VARIABLE_LENGTH_STRING);
    CHECK_BIO(error);
    error = BIOH5Gset_reads_properties_reads_type(var_properties, BASE_SPACE);
    CHECK_BIO(error);

    /* Create reads storage */
    error = BIOH5Gcreate_reads_collection(file,
        var_properties,
        variable_path,
        &var_reads);
    CHECK_BIO(error);

    error = BIOH5Gdestroy_reads_properties(&var_properties);

    /* Store some reads */
    in.sequence        = "Hi there!";
    in.quality_values  = "42       ";
    in.identifier      = "This is the identifier";
    error = BIOH5Gadd_read(var_reads, &in);
    CHECK_BIO(error);

    in.sequence   = "";
    in.quality_values  = "";
    in.identifier = "";
    error = BIOH5Gadd_read(var_reads, &in);
    CHECK_BIO(error);

    in.sequence        = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    in.quality_values  = "0123456789                ";
    in.identifier      = "This is an identifier that is very long";
    error = BIOH5Gadd_read(var_reads, &in);
    CHECK_BIO(error);

    error = BIOH5Gclose_reads_collection(&var_reads);
    CHECK_BIO(error);

    /**********************************
     * Test fixed-length reads output *
     **********************************/

    /* Open reads storage */
    error = BIOH5Gopen_reads_collection(file,
        fixed_path,
        READ_ONLY,
        &reads);
    CHECK_BIO(error);

    /* Get reads by index */
    error = BIOH5Gget_reads_count(reads, &count);
    CHECK_BIO(error);

    for(i = 0; i < count; i++)
    {
        error = BIOH5Gget_read(reads, i, &out);
        CHECK_BIO(error);
        error = BIOH5Gfree_read(&out);
        CHECK_BIO(error);
    }

    /* Get reads using an iterator */
    error = BIOH5Gcreate_reads_iterator(reads, &iter);
    CHECK_BIO(error);

    count = 0;
    while(BIOHDF_NO_ERROR == (error = BIOH5Gget_next_read(iter, &i, &out)) && out != NULL)
    {
        count++;
        error = BIOH5Gfree_read(&out);
        CHECK_BIO(error);
    }
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 3);

    error = BIOH5Gdestroy_reads_iterator(&iter);
    CHECK_BIO(error);

    error = BIOH5Gclose_reads_collection(&reads);
    CHECK_BIO(error);


    /*************************************
     * Test variable-length reads output *
     *************************************/

    /* Open reads storage */
    error = BIOH5Gopen_reads_collection(file,
        variable_path,
        READ_ONLY,
        &var_reads);
    CHECK_BIO(error);

    /* Get reads by index */
    error = BIOH5Gget_reads_count(var_reads, &count);
    CHECK_BIO(error);
    for(i = 0; i < count; i++)
    {
        error = BIOH5Gget_read(var_reads, i, &out);
        CHECK_BIO(error);
        error = BIOH5Gfree_read(&out);
        CHECK_BIO(error);
    }

    /* Get reads using an iterator */
    error = BIOH5Gcreate_reads_iterator(var_reads, &iter);
    CHECK_BIO(error);

    count = 0;
    while(BIOHDF_NO_ERROR == (error = BIOH5Gget_next_read(iter, &i, &out)) && out != NULL)
    {
        count++;
        error = BIOH5Gfree_read(&out);
        CHECK_BIO(error);
    }
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 3);

    error = BIOH5Gdestroy_reads_iterator(&iter);
    CHECK_BIO(error);

    error = BIOH5Gclose_reads_collection(&var_reads);
    CHECK_BIO(error);


    /**SUCCESS***************************************************************/

    error = BIOHDFclose_biohdf_file(&file);
    CHECK_BIO(error);

    remove(filename);

    return 0;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** TEST FAILURE ***   Test: Reads (test/test_reads.c)\n");

    BIOH5Gclose_reads_collection(&reads);
    BIOH5Gclose_reads_collection(&var_reads);
    BIOHDFclose_biohdf_file(&file);

    /* The temp file is left in place for debugging on errors.  It will
     * get cleaned up when the program is run again.
     */

    return 1;
}
