/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"

#include "biohdf.h"

#include "biohdf_file.h"
#include "biohdf_file_internal.h"
#include "biohdf_vpart.h"
#include "biohdf_vector.h"


/* Used to test compound types */
typedef struct
{
    int     i;
    char    c;
    double  d;
} compound_type;




/*****************************************************************************
 * Main test program                                                         *
 *****************************************************************************/


/* Test basic BioHDF vector storage */
int
main(int argc, char **argv) 
{
    biohdf_file        file          = NULL;   /* BioHDF file handle */

    vpart              *vp           = NULL;   /* Vector handle      */
    vector             *v            = NULL;   /* Vector part handle */
    
    hid_t              tid           = -1;     /* HDF5 type id for the data
                                                * to be stored.
                                                */

    hsize_t            size[1];                /* length of the string data 
                                                * to be stored 
                                                */

    char               buffer[500];            /* A buffer to hold data that
                                                * will be stored in a vector
                                                * or vector part.
                                                */

    char               *data          = NULL;  /* A buffer filled by the
                                                * vector and vpart iterators
                                                */

    int                i;                      /* iterator */
    
    int                is_read_only   = 0;
    int                is_c_string    = 0;
    hsize_t            count;
    compound_type      cdata;

    long int           chunk_size = 1024;
    compression_level  compression = 5;

    
    int                fixed_length;

    char   filename[]      = "test_vectors.h5";  /* test file name */

    char   vp_name[]       = "vpart";
    char   fixed_name[]    = "FIXED";
    char   variable_name[] = "VARIABLE";
    char   compound_name[] = "FIXED_COMPOUND";

    FILE         *test             = NULL;           /* For testing file existence */

    herr_t       hdf_error         = -1;             /* HDF5 error code */
    biohdf_error error             = BIOHDF_BADNESS; /* BioHDF error code */

    /**CODE*****************************************************************/

    /* Create the test BioHDF file */

    if((test = fopen(filename, "r")))
    {
        fclose(test);
        remove(filename);
    }

    error = BIOHDFcreate_biohdf_file(filename, &file);
    CHECK_BIO(error);


    /***********************************************************************/
    /**VECTOR PART TESTS****************************************************/
    /***********************************************************************/


    /*******************************************************
     * Test adding blocks of data to a vector part (vpart) *
     *******************************************************/

    /* Tests _add_vpart_block() */

    /* Create the HDF5 data type - a fixed-length "string" of length 50.
     * The 1 for the second parameter is '# dimensions'
     */
    size[0] = 50;
    tid = H5Tarray_create(H5T_NATIVE_CHAR, 1, size);
    CHECK_HID_T(tid);

    /* Create the vector part */
    error = _create_vpart(file->file_id, vp_name, tid, 1024, GZIP_5, &vp);
    CHECK_BIO(error);

    /* Add some data to a buffer for later storage */
    for(i = 0; i < 50; i++)
    {
        buffer[i] = '0';
    }
    for(i = 50; i < 100; i++)
    {
        buffer[i] = '1';
    }
    for(i = 100; i < 150; i++)
    {
        buffer[i] = '2';
    }
    for(i = 150; i < 200; i++)
    {
        buffer[i] = '3';
    }
    for(i = 200; i < 250; i++)
    {
        buffer[i] = '4';
    }
    for(i = 250; i < 300; i++)
    {
        buffer[i] = '5';
    }
    for(i = 300; i < 350; i++)
    {
        buffer[i] = '6';
    }
    for(i = 350; i < 400; i++)
    {
        buffer[i] = '7';
    }
    for(i = 400; i < 450; i++)
    {
        buffer[i] = '8';
    }
    error = _add_vpart_block(vp, &buffer, 9);
    CHECK_BIO(error);

    for(i = 0; i < 50; i++)
    {
        buffer[i] = '9';
    }
    error = _add_vpart_block(vp, &buffer, 1);
    CHECK_BIO(error);
    
    /* Close everything */
    hdf_error = H5Tclose(tid);
    tid = -1;
    CHECK_HERR_T(hdf_error);
    
    error = _close_vpart(&vp);
    vp = NULL;
    CHECK_BIO(error);



    /******************************************************
     * Test adding data elements to a vector part (vpart) *
     ******************************************************/

    /* Tests _add_vpart_element() */

    /* Open the vpart
     * This also tests opening as opposed to creating vparts.
     */
    is_read_only = 0;
    error = _open_vpart(file->file_id, vp_name, is_read_only, &vp);
    CHECK_BIO(error);

    /* Create and add some single vector parts */
    for(i = 0; i < 50; i++)
    {
        buffer[i] = 'A';
    }
    error = _add_vpart_element(vp, &buffer);
    CHECK_BIO(error);

    for(i = 0; i < 50; i++)
    {
        buffer[i] = 'B';
    }
    error = _add_vpart_element(vp, &buffer);
    CHECK_BIO(error);

    /* Close up */
    error = _close_vpart(&vp);
    vp = NULL;
    CHECK_BIO(error);



    /******************************************************************
     * Test reading data from a vector part (vpart)                   *
     ******************************************************************/

    /* Open the vector part */
    is_read_only = 1;
    error = _open_vpart(file->file_id, vp_name, is_read_only, &vp);
    CHECK_BIO(error);

    /* Get the number of items stored in the vector part */
    error = _get_vpart_count(vp, &count);
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 12);

    /* Read all the data */
    is_c_string = 1;
    for(i = 0; i < count; i++)
    {
        error = _get_vpart_block(vp, (hsize_t)i, (hsize_t)i, is_c_string, (void **)&data);
        CHECK_BIO(error);
        CHECK_PTR(data);
        free(data);
        data = NULL;
    }

    /* Close the vector part */
    error = _close_vpart(&vp);
    CHECK_BIO(error);


    /***********************************************************************/
    /**VECTOR TESTS*********************************************************/
    /***********************************************************************/


    /***********************************************************
     * Test creating and reading data from a FIXED-SIZE vector *
     ***********************************************************/


    /* Create the vector */
    fixed_length = 35;
    error = _create_vector(file->file_id, fixed_name, H5T_NATIVE_CHAR,
        FIXED_SIZE_ARRAY, fixed_length, chunk_size, compression, &v);
    CHECK_BIO(error);

    /* Add some data to it */
    for(i = 0; i < fixed_length; i++)
    {
        buffer[i] = 'C';
    }
    error = _add_vector_element(v, &buffer, fixed_length);
    CHECK_BIO(error);

    for(i = 0; i < fixed_length; i++)
    {
        buffer[i] = 'D';
    }
    error = _add_vector_element(v, &buffer, fixed_length);
    CHECK_BIO(error);

    for(i = 0; i < fixed_length; i++)
    {
        buffer[i] = 'E';
    }
    error = _add_vector_element(v, &buffer, fixed_length);
    CHECK_BIO(error);

    error = _close_vector(&v);
    v = NULL;
    CHECK_BIO(error);


    /* Can I reopen it properly?  Will it guess the scheme correctly? */
    is_read_only = 1;
    error = _open_vector(file->file_id, fixed_name, is_read_only, &v);
    CHECK_BIO(error);

    /* Check the count */
    error = _get_vector_count(v, &count);
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 3);


    /* Read the data in the vector */
    is_c_string = 1;
    for(i = 0; i < count; i++)
    {
        error = _get_vector_element(v, i, is_c_string, (void**)&data);
        CHECK_BIO(error);
        CHECK_PTR(data);

        free(data);
    }

    /* Close everything */
    error = _close_vector(&v);
    v = NULL;
    CHECK_BIO(error);


    /**************************************************************
     * Test creating and reading data from a VARIABLE-SIZE vector *
     **************************************************************/

    /* Create the vector */
    fixed_length = 0;
    error = _create_vector(file->file_id, variable_name, H5T_NATIVE_CHAR,
        VARIABLE_SIZE_ARRAY, fixed_length, chunk_size, compression, &v);
    CHECK_BIO(error);

    /* Add some data to it */
    for(i = 0; i < 42; i++)
    {
        buffer[i] = 'M';
    }
    error = _add_vector_element(v, &buffer, 42);
    CHECK_BIO(error);

    /* We need to support zero length vectors! */
    error = _add_vector_element(v, &buffer, 0);
    CHECK_BIO(error);

    for(i = 0; i < 23; i++)
    {
        buffer[i] = 'N';
    }
    error = _add_vector_element(v, &buffer, 23);
    CHECK_BIO(error);

    error = _close_vector(&v);
    CHECK_BIO(error);


    /* Can I reopen it properly?  Will it guess the scheme correctly? */
    error = _open_vector(file->file_id, variable_name, is_read_only, &v);
    CHECK_BIO(error);

    /* Check correct number of elements */
    error = _get_vector_count(v, &count);
    CHECK_BIO(error);
    CHECK_INT_EQ(count, 3);

    /* Read back the data */
    is_c_string = 1;
    for(i = 0; i < count; i++)
    {
        error = _get_vector_element(v, i, is_c_string, (void**)&data);
        CHECK_BIO(error);
        CHECK_PTR(data);

        free(data);
    }

    /* Close everything. */
    error = _close_vector(&v);
    CHECK_BIO(error);


    /*************************************************************
     * Test creating and reading data using a compound data type *
     *************************************************************/

    /* Create compound type */
    tid = H5Tcreate(H5T_COMPOUND, sizeof(compound_type));
    CHECK_HID_T(tid);

    hdf_error = H5Tinsert(tid, "integer",    HOFFSET(compound_type, i),  H5T_NATIVE_INT);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Tinsert(tid, "character",  HOFFSET(compound_type, c),  H5T_NATIVE_CHAR);
    CHECK_HERR_T(hdf_error);
    hdf_error = H5Tinsert(tid, "double",     HOFFSET(compound_type, d),  H5T_NATIVE_DOUBLE);
    CHECK_HERR_T(hdf_error);

    /* Create the vector */
    fixed_length = 1;
    error = _create_vector(file->file_id, compound_name, tid,
        FIXED_SIZE_ARRAY, fixed_length, chunk_size, compression, &v);
    CHECK_BIO(error);

    /* Write some data to the vector */
    cdata.c = 'z';
    cdata.d = 3.14159;
    cdata.i = 42;
    error = _add_vector_element(v, (void*)&cdata, 1);
    CHECK_BIO(error);

    cdata.c = 'q';
    cdata.d = 2.71;
    cdata.i = 1234;
    error = _add_vector_element(v, (void*)&cdata, 1);
    CHECK_BIO(error);


    /* Close everything.  Again. */
    error = H5Tclose(tid);
    CHECK_BIO(error);
    tid = -1;

    error = _close_vector(&v);
    CHECK_BIO(error);
    v = NULL;


    /**SUCCESS***************************************************************/

    error = BIOHDFclose_biohdf_file(&file);
    CHECK_BIO(error);

    remove(filename);

    return 0;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** TEST FAILURE ***   Test: Vectors (test/test_vectors.c)\n");

    if(tid >= 0)
    {
        H5Tclose(tid);
    }

    if(vp != NULL)
    {
        _close_vpart(&vp);
    }

    if(v != NULL)
    {
        _close_vector(&v);
    }

    if(data != NULL)
    {
        free(data);
    }

    if(file != NULL)
    {
        BIOHDFclose_biohdf_file(&file);
    }

    return 1;
}

