/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(HAVE_UNISTD_H)
    #include <unistd.h>    /* getopt processing */
    #include <getopt.h>
#elif defined(_WIN32)
    #include "bsd_getopt.h"
#endif

#include "biohdf.h"
#include "bioh5g.h"



static void _parse_options(int argc,
                           char *argv[],
                           /*OUT*/ char                       **biohdf_filename,
                           /*OUT*/ char                       **output_filename,
                           /*OUT*/ char                       **path_to_alignments,
                           /*OUT*/ bioh5g_alignments_format   *format,
                           /*OUT*/ int64_t               *n_elements,
                           /*OUT*/ int                        *min_quality,
                           /*OUT*/ uint32_t              *sam_flag_mask,
                           /*OUT*/ int                        *n_regions,
                           /*OUT*/ bioh5g_reference_region    **regions,
                           /*OUT*/ int                        *write_header);


int main(int argc, char** argv)
{
    biohdf_file                 f           = NULL;
    bioh5g_reads                r           = NULL;
    bioh5g_alignments           a           = NULL;
    bioh5g_read_data           *r_data      = NULL;
    bioh5g_alignment_data      *a_data      = NULL;
    bioh5g_alignments_format    format      = SAM_FORMAT;
    int64_t                     n_elements  = 0;

    FILE               *out                 = NULL;
    char               *out_filename        = NULL;
    char               *out_mode            = "a";
    char               *in_filename         = NULL;
    biohdf_open_mode    in_mode             = READ_ONLY;
    char               *path_to_alignments  = NULL;
    char               *path_to_reads       = NULL;
    char               *header              = NULL;
    int                 n_regions           = 0;

    bioh5g_reference_region *regions = NULL;

    bioh5g_alignments_iterator iter = NULL;

    int64_t     count   = 0;
    int64_t     start   = 0;
    int64_t     end     = 0;
    int64_t     i       = 0;
    int64_t     index   = 0; /* Dummy OUT variable */

    int         min_quality     = INT_MAX;
    uint32_t    sam_flag_mask   = 0;

    int             write_header    = 1;

    biohdf_error    error       = BIOHDF_NO_ERROR;  /* BioHDF error code */
    int             stdio_err   = 0;                /* stdio error code */

    /**CODE*****************************************************************/

    /* Parse command line options */
    _parse_options(argc, argv,
        &in_filename,
        &out_filename,
        &path_to_alignments,
        &format,
        &n_elements,
        &min_quality,
        &sam_flag_mask,
        &n_regions,
        &regions,
        &write_header);

    if(NULL == in_filename)
    {
        printf("ERROR: must specify a BioHDF filename with -f\n");
        exit(EXIT_FAILURE);
    }

    if(NULL == path_to_alignments)
    {
        printf("ERROR: must specify an alignments collection path with -A\n");
        exit(EXIT_FAILURE);
    }


    /* Open export file or stream */
    if(out_filename != NULL)
    {
        out = fopen(out_filename, out_mode);
        CHECK_PTR(out);
    }
    else
    {
        out = stdout;
    }

    /* Open BioHDF alignment hits and reads storage */
    error = BIOHDFopen_biohdf_file(in_filename, in_mode, &f);
    CHECK_BIO(error);
    error = BIOH5Gopen_alignments_collection(f, path_to_alignments, in_mode, &a);
    CHECK_BIO(error);
    error = BIOH5Gget_reads_path(a, &path_to_reads);
    CHECK_BIO(error);
    error = BIOH5Gopen_reads_collection(f, path_to_reads, in_mode, &r);
    CHECK_BIO(error);
    error = BIOHDFfree_memory((void **)&path_to_reads);
    CHECK_BIO(error);


    if(0 == n_elements)
    {
        /* Use iterators */

        /* Emit header */
        if(1 == write_header)
        {
            error = BIOH5Gget_alignment_file_header(a, &format, &header);
            CHECK_BIO(error);
            if(header != NULL)
            {
                stdio_err = fprintf(out, "%s", header);
                CHECK_INT_GE(stdio_err, 0);
                error = BIOHDFfree_memory((void**)(&header));
                CHECK_BIO(error);
            }
        }

        /* Create alignments iterator */
        error = BIOH5Gcreate_alignments_iterator(a, &iter);
        CHECK_BIO(error);
        if(min_quality < INT_MAX)
        {
            error = BIOH5Gadd_alignments_iterator_mapq_filter(iter, min_quality);
            CHECK_BIO(error);
        }
        if(sam_flag_mask != 0)
        {
            error = BIOH5Gadd_alignments_iterator_flags_filter(iter, sam_flag_mask);
            CHECK_BIO(error);
        }
        for(i = 0; i < n_regions; i++)
        {
            error = BIOH5Gadd_alignments_iterator_range_filter(iter, 
                regions[i].reference, regions[i].start, regions[i].end);
            CHECK_BIO(error);
        }

        /* Emit alignments */
        while(BIOHDF_NO_ERROR == (error = BIOH5Gget_next_alignment(iter, &index, &a_data)) && a_data != NULL)
        {
            error = BIOH5Gget_read(r, a_data->read_index, &r_data);
            CHECK_BIO(error);
            error = BIOH5Gwrite_alignment_to_stream(a_data, r_data, SAM_FORMAT, out);
            CHECK_BIO(error);

            error = BIOH5Gfree_alignment_data(&a_data);
            CHECK_BIO(error);
            error = BIOH5Gfree_read(&r_data);
            CHECK_BIO(error);
        }
        CHECK_BIO(error);
        error = BIOH5Gdestroy_alignments_iterator(&iter);
        CHECK_BIO(error);
    }
    else
    {
        /* Return n elements */

        error = BIOH5Gget_alignments_count(a, &count);
        CHECK_BIO(error);
        if (n_elements > 0) /* HEAD */
        {
            start = 0;
            end = (n_elements < count) ? n_elements : count;
        }
        else if (n_elements < 0) /* TAIL */
        {
            start = (count + n_elements < 0) ? 0 : count + n_elements;
            end = count;
        }
        else
        {
            start = 0;
            end = count;
        }

        for (i=start; i < end; i++)
        {
            error = BIOH5Gget_alignment(a, i, &a_data);
            CHECK_BIO(error);
            error = BIOH5Gget_read(r, a_data->read_index, &r_data);
            CHECK_BIO(error);

            error = BIOH5Gwrite_alignment_to_stream(a_data, r_data, SAM_FORMAT, out);
            CHECK_BIO(error);

            error = BIOH5Gfree_alignment_data(&a_data);
            CHECK_BIO(error);
            error = BIOH5Gfree_read(&r_data);
            CHECK_BIO(error);
        }
    }

    /**SUCCESS***************************************************************/

    /* Free memory and close files */
    if(out != stdout  && out != stderr)
    {
        stdio_err = fclose(out);
        CHECK_INT_GE(0, stdio_err);
    }

    error = BIOH5Gclose_alignments_collection(&a);
    CHECK_BIO(error);
    error = BIOH5Gclose_reads_collection(&r);
    CHECK_BIO(error);
    error = BIOHDFclose_biohdf_file(&f);
    CHECK_BIO(error);

    free(in_filename);
    free(path_to_alignments);
    free(out_filename);

    return EXIT_SUCCESS;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** ERROR ***  in bioh5g_export_alignments\n");

    /* Try to close/free as much as possible */

    if(out != stdout  && out != stderr)
    {
        fclose(out);
    }

    BIOH5Gclose_alignments_collection(&a);
    BIOH5Gclose_reads_collection(&r);
    BIOHDFclose_biohdf_file(&f);

    free(in_filename);
    free(path_to_alignments);
    free(out_filename);

    return EXIT_FAILURE;
}


/*****************************************************************************
 * Usage strings                                                             *
 *****************************************************************************/

const char *cmd_description = 
    "Export alignments to stdout or file.\n";

const char *cmd_option = 
    "  [--biohdf-file|-f <filename>] (REQUIRED ARGUMENT)\n"
    "      The BioHDF file name.\n"
    "\n"
    "  [--alignments-collection|-A <collection>] (REQUIRED ARGUMENT)\n"
    "      The named location in the BioHDF file where the alignments\n"
    "      will be obtained.  There is no need to specify the reads group\n"
    "      as that will be obtained from the alignments collection.\n"
    "      \n"
    "      This should look like a filesystem path.  e.g. /path/to/alignments\n"
    "      Collection paths must start from the root group '/'.  The\n"
    "      trailing slash is optional.\n"
    "\n"
    "  [--out-file|-o <filename>]\n"
    "      Output file name.\n"
    "      Default is STDOUT if a filename is not specified.\n"
    "\n"
    "  [--out-format|-F sam]\n"
    "      Output file format for the alignments.  Default is SAM.\n"
    "\n"
    "  [--beginning|-b <number of entries>]\n"
    "      Return n entries from the start of the collection.\n"
    "      No header is emitted and no filters are used with this option.\n"
    "      Similar to the unix head command.\n"
    "\n"
    "  [--end|-e <number of entries>]\n"
    "      Return n entries from the end of the collection.\n"
    "      No header is emitted and no filters are used with this option.\n"
    "      Similar to the unix tail command.\n"
    "\n"
    "  <reference>:<start>:<end>\n"
    "      Filter and return only alignments which overlap with\n"
    "      the 1-based, inclusive interval between <start> and\n"
    "      <end> on <reference>.  Multiple intervals can be specified,\n"
    "      separated by spaces, and the intervals should come after all\n"
    "      other options.  Omit the value (but not the colon) for <start>\n"
    "      or <end> if the interval should run to the start or end of the\n"
    "      reference.\n"
    "      examples: \"chr1::\", \"chr18:10500:15000\"\n"
    "\n"
    "  [--min-quality|-Q <integer quality score>]\n"
    "      Filter and return only alignments exceed or equal a minimum\n"
    "      quality value (SAM MAPQ, et al.).\n"
    "\n"
    "  [--sam-flag-mask|-M <unsigned integer SAM FLAG mask>]\n"
    "      Filter and return only those alignments which have all the bits\n"
    "      in the mask set.\n"
    "\n"
    "  [--no-header\n"
    "      Do not emit the SAM header.\n"
    "\n"
    "  [--help|-h]\n"
    "      Command line help message.\n";

/* TODO: fixme */
const char *cmd_example =
    "bioh5g_export_alignments -f myfile.h5 -A \"/myhits\" -o output_file.sam\n"
    "bioh5g_export_alignments -f myfile.h5 -A \"/path/to/alignments/\" -F sam -b 25\n"
    "bioh5g_export_alignments -f myfile.h5 -A \"/sample1/alignments/\" -F sam\n"
    "     chr1::  chr18:10500:15000  chr14:12345:  chr13::42424\n";



static void _parse_options(int argc,
                           char *argv[],
                           /*OUT*/ char                       **biohdf_filename,
                           /*OUT*/ char                       **output_filename,
                           /*OUT*/ char                       **path_to_alignments,
                           /*OUT*/ bioh5g_alignments_format   *format,
                           /*OUT*/ int64_t               *n_elements,
                           /*OUT*/ int                        *min_quality,
                           /*OUT*/ uint32_t              *sam_flag_mask,
                           /*OUT*/ int                        *n_regions,
                           /*OUT*/ bioh5g_reference_region    **regions,
                           /*OUT*/ int                        *write_header)
{
    int i, j, k;                     /* iterators  for region parse */
    int length;                      /* string length */
    int bad_number;                  /* flag for non-num chars in a num field */
    char s[BIOHDF_MAX_STRING_SIZE];  /* temp string */
    bioh5g_reference_region r;       /* temp region */

    int c = -1;
    int option_index = -1;

    /* Short options */
    char *options = "f:o:A:F:b:e:Q:M:h";

    /* Long options */
    struct option long_options[] = {
        {"biohdf-file",            required_argument, 0, 'f'},
        {"out-file",               required_argument, 0, 'o'},
        {"alignments-collection",  required_argument, 0, 'A'},
        {"out-format",             required_argument, 0, 'F'},
        {"beginning",              required_argument, 0, 'b'},
        {"end",                    required_argument, 0, 'e'},
        {"min-quality",            required_argument, 0, 'Q'},
        {"sam-flag-mask",          required_argument, 0, 'M'},
        {"no-header",              no_argument,       write_header, 0},
        {"help",                   no_argument,       0, 'h'},
        {0, 0, 0, 0}
    };

    if(argc <= 1)
    {
        goto USAGE;
    }

    /* Parse options */
    opterr = 0;
    for (;;)
    {
        c = getopt_long(argc, argv, options, long_options, &option_index);
        if (-1 == c) { break; }

        switch (c)
        {
        case 0:
            /* If this option sets a flag, do nothing else now. */
            break;

        /* BioHDF file */
        case 'f':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*biohdf_filename) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*biohdf_filename), optarg);
            break;

        /* Path to alignments */
        case 'A':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*path_to_alignments) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*path_to_alignments), optarg);
            break;

        /* Output file */
        case 'o':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*output_filename) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*output_filename), optarg);
            break;

        /* Output format */
        case 'F':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            if(0 == strcmp("SAM", optarg) || 0 == strcmp("sam", optarg))
            {
                *format = SAM_FORMAT;
            }
            else
            {
                printf("Unrecognized format: %s\n", optarg);
                goto USAGE;
            }
            break;

        /* Number of lines at the start of the file */
        case 'b':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*n_elements) = atoi(optarg);
            break;

        /* Number of lines at the end of the file */
        case 'e':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*n_elements) = -atoi(optarg);
            break;

        /* Minimum quality value - SAM MAPQ et al. */
        case 'Q':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*min_quality) = atoi(optarg);
            break;

        /* SAM FLAG mask for filtering results */
        case 'M':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*sam_flag_mask) = atoi(optarg);
            break;

        /* Error, so show help and return error code */
        default : 
            fprintf(stderr, "\nError: Unrecognized command-line option.\n\n");
            /* fall-through to help */

            /* help, show usage and exit */
        case 'h':
            goto USAGE;
        }
    }

    /* Create region memory using argc as an upper bound of the # */
    *regions = (bioh5g_reference_region *)calloc(
        argc, sizeof(bioh5g_reference_region));
    *n_regions = 0;

    /* Find chromosome regions */
    for(i = 0; i < argc; i++)
    {
        memset(r.reference, 0, BIOHDF_MAX_STRING_SIZE);
        bad_number = 0;
        r.start = 0;
        r.end = INT_MAX;

        length = (int)strlen(argv[i]);
        

        /* Reference name */
        k = 0;
        j = 0;
        while(j < length && j < BIOHDF_MAX_STRING_SIZE && argv[i][j] != ':')
        {
            r.reference[k] = argv[i][j];
            j++;
            k++;
        }

        if(argv[i][j] != ':') { continue; }  /* Not a region */

        /* Start */
        memset(s, 0, BIOHDF_MAX_STRING_SIZE);
        k = 0;
        j++;
        while(j < length && j < BIOHDF_MAX_STRING_SIZE && argv[i][j] != ':')
        {
            if(!isdigit(argv[i][j])) { bad_number = 1; }
            s[k] = argv[i][j];
            j++;
            k++;
        }

        if(argv[i][j] != ':' || bad_number) { continue; }  /* Not a region */
        r.start = '\0' == s[0] ? 0 : atoi(s);

        /* End */
        memset(s, 0, BIOHDF_MAX_STRING_SIZE);
        k = 0;
        j++;
        while(j < length && j < BIOHDF_MAX_STRING_SIZE)
        {
            if(!isdigit(argv[i][j])) { bad_number = 1; }
            s[k] = argv[i][j];
            j++;
            k++;
        }
        r.end = '\0' == s[0] ? INT_MAX : atoi(s);

        if(bad_number) { continue; }  /* Not a region */

        /* Add the region to the list */
        strcpy((*regions)[*n_regions].reference, r.reference);
        (*regions)[*n_regions].start = r.start;
        (*regions)[*n_regions].end = r.end;
        (*n_regions)++;
    }

    return;

USAGE:
    fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
        cmd_description, cmd_option, cmd_example,
        "For more information about BioHDF visit http://www.biohdf.org/\n");
    exit(EXIT_FAILURE);
}
