/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(HAVE_UNISTD_H)
    #include <unistd.h>    /* getopt processing */
    #include <getopt.h>
#elif defined(_WIN32)
    #include "bsd_getopt.h"
#endif

#include "biohdf.h"
#include "bioh5g.h"

/* Export format */
typedef enum
{
    FASTQ,
    FASTA,
    SEQUENCES_ONLY,
    QUALITY_VALUES_ONLY
} output_format;


static void _parse_options(int argc,
                           char *argv[],
                           /*OUT*/ char **biohdf_file_name,
                           /*OUT*/ char **path_to_reads,
                           /*OUT*/ char **output_file,
                           /*OUT*/ output_format *format,
                           /*OUT*/ int64_t *n_elements);


int main(int argc, char** argv) 
{
    biohdf_file         f           = NULL;
    bioh5g_reads        r           = NULL;
    bioh5g_read_data    *data       = NULL;
    output_format       format      = FASTQ;
    int64_t             n_elements  = 0;

    FILE             *out               = NULL;
    char             *out_filename      = NULL;
    char             *out_mode          = "a";
    char             *in_filename       = NULL;
    biohdf_open_mode  in_mode           = READ_ONLY;
    char             *path_to_reads     = NULL;

    int64_t     count   = 0;
    int64_t     start   = 0;
    int64_t     end     = 0;
    int64_t     i       = 0;

    biohdf_error    error       = BIOHDF_NO_ERROR;  /* BioHDF error code */
    int             stdio_err   = 0;                /* stdio error code */

    /**CODE*****************************************************************/

    /* Parse command line options */
    _parse_options(argc, argv,
        &in_filename,
        &path_to_reads,
        &out_filename,
        &format,
        &n_elements);

    if(NULL == in_filename)
    {
        printf("ERROR: must specify a BioHDF file to export files from with -f\n");
        exit(EXIT_FAILURE);
    }

    if(NULL == path_to_reads)
    {
        printf("ERROR: must specify a reads collection path with -R\n");
        exit(EXIT_FAILURE);
    }


    /* Open export file or stream */
    if(out_filename != NULL)
    {
        out = fopen(out_filename, out_mode);
        CHECK_PTR(out);
    }
    else
    {
        out = stdout;
    }


    /* Open BioHDF reads storage */
    error = BIOHDFopen_biohdf_file(in_filename, in_mode, &f);
    CHECK_BIO(error);
    error = BIOH5Gopen_reads_collection(f, path_to_reads, in_mode, &r);
    CHECK_BIO(error);


    /* Determine start and end points of export */
    error = BIOH5Gget_reads_count(r, &count);
    CHECK_BIO(error);

    if (n_elements > 0) /* HEAD */
    {
        start = 0;
        end = (n_elements < count) ? n_elements : count;
    }
    else if (n_elements < 0) /* TAIL */
    {
        start = (count + n_elements < 0) ? 0 : count + n_elements;
        end = count;
    }
    else
    {
        start = 0;
        end = count;
    }

    /* Export the reads */
    for (i = start; i < end; i++)
    {
        error = BIOH5Gget_read(r, i, &data);
        CHECK_BIO(error);

        switch(format)
        {
            case SEQUENCES_ONLY:
                stdio_err = fprintf(out, "%s\n", data->sequence);
                CHECK_INT_GE(0, stdio_err);
                break;
            case QUALITY_VALUES_ONLY:
                stdio_err = fprintf(out, "%s\n", data->quality_values);
                CHECK_INT_GE(0, stdio_err);
                break;
            case FASTA:
                error = BIOH5Gwrite_read_to_stream(data, FASTA_FORMAT, out);
                CHECK_BIO(error);
                break;
            default:   /* FASTQ */
                error = BIOH5Gwrite_read_to_stream(data, FASTQ_FORMAT, out);
                CHECK_BIO(error);
                break;
        }
        error = BIOH5Gfree_read(&data);
        CHECK_BIO(error);
    }

    /**SUCCESS***************************************************************/

    /* Free memory and close files */
    if(out != stdout && out != stderr)
    {
        stdio_err = fclose(out);
        CHECK_INT_NE(EOF, stdio_err);
    }
    error = BIOH5Gclose_reads_collection(&r);
    CHECK_BIO(error);
    error = BIOHDFclose_biohdf_file(&f);
    CHECK_BIO(error);

    free(in_filename);
    free(path_to_reads);
    free(out_filename);

    return EXIT_SUCCESS;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** ERROR ***  in bioh5g_export_reads\n");

    /* Try to close/free as much as possible */

    if(out != stdout && out != stderr)
    {
        fclose(out);
    }

    BIOH5Gclose_reads_collection(&r);
    BIOHDFclose_biohdf_file(&f);

    free(in_filename);
    free(path_to_reads);
    free(out_filename);

    return EXIT_FAILURE;
}


/*****************************************************************************
 * Usage strings                                                             *
 *****************************************************************************/

const char *cmd_description = 
    "Export reads from a BioHDF file.\n";

const char *cmd_option = 
    "  [--biohdf-file|-f <filename>] (REQUIRED ARGUMENT)\n"
    "      The BioHDF file name.\n"
    "\n"
    "  [--reads-collection|-R <collection>] (REQUIRED ARGUMENT)\n"
    "      The named location in the BioHDF file where the reads will be\n"
    "      obtained.\n"
    "      \n"
    "      This should look like a filesystem path.  e.g. /path/to/reads\n"
    "      Collection paths must start from the root group '/'.  The\n"
    "      trailing slash is optional.\n"
    "\n"
    "  [--out-file|-o <filename>]\n"
    "      Output file name.\n"
    "      Default is STDOUT if a filename is not specified.\n"
    "\n"
    "  [--out-format|-F fasta|fastq|reads|qualities]\n"
    "      Output file format for read sequences and associated\n"
    "      ASCII-encoded qualities.  The reads and qualities options\n"
    "      return just those data, 1 entry per line.\n"
    "      Default is FASTQ.\n"
    "\n"
    "  [--beginning|-b <number of entries>]\n"
    "      Return n entries from the start of the collection.\n"
    "      Similar to the unix head command.\n"
    "\n"
    "  [--end|-e <number of entries>]\n"
    "      Return n entries from the end of the collection.\n"
    "      Similar to the unix tail command.\n"
    "\n"
    "  [--help|-h]\n"
    "      Command line help message.\n";

/* TODO: fixme */
const char *cmd_example =
    "bioh5g_export_reads -f myfile.h5 -R \"/sample1/reads/\" -o output.fastq\n"
    "bioh5g_export_reads -f myfile.h5 -R \"/complicated/data/org/reads\"\n"
    "     -F qualities\n"
    "bioh5g_export_reads -f myfile.h5 -R \"/reads\" -F fasta -b 25\n";



/*****************************************************************************
 * Parse Command-Line Options                                                *
 *****************************************************************************/

static void _parse_options(int argc,
                          char *argv[],
                          /*OUT*/ char **biohdf_file_name,
                          /*OUT*/ char **path_to_reads,
                          /*OUT*/ char **output_file,
                          /*OUT*/ output_format *format,
                          /*OUT*/ int64_t *n_elements)
{
    int c = -1;
    int option_index = -1;

    /* Short options */
    char *options = "f:R:o:F:b:e:h";

    /* Long options */
    struct option long_options[] = {
        {"biohdf-file",       required_argument, 0, 'f'},
        {"reads-collection",  required_argument, 0, 'R'},
        {"out-file",          required_argument, 0, 'o'},
        {"out-format",        required_argument, 0, 'F'},
        {"beginning",         required_argument, 0, 'b'},
        {"end",               required_argument, 0, 'e'},
        {"help",              no_argument,       0, 'h'},
        {0, 0, 0, 0}
    };

    if(argc <= 1)
    {
        goto USAGE;
    }

    /* Parse options */
    opterr = 0;
    for (;;)
    {
        c = getopt_long(argc, argv, options, long_options, &option_index);
        if (-1 == c) { break; }

        switch (c)
        {
        case 0:
            /* If this option sets a flag, do nothing else now. */
            break;

        /* BioHDF file */
        case 'f':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*biohdf_file_name) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*biohdf_file_name), optarg);
            break;

        /* Path to reads */
        case 'R':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*path_to_reads) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*path_to_reads), optarg);
            break;

        /* Output file */
        case 'o':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*output_file) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*output_file), optarg);
            break;

        /* Output format */
        case 'F':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            if(0 == strcmp("fasta", optarg))
            {
                *format = FASTA;
            }
            else if(0 == strcmp("reads", optarg))
            {
                *format = SEQUENCES_ONLY;
            }
            else if(0 == strcmp("qualities", optarg))
            {
                *format = QUALITY_VALUES_ONLY;
            }
            else
            {
                *format = FASTQ;
            }
            break;

        /* Number of lines at the start of the file */
        case 'b':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*n_elements) = atoi(optarg);
            break;

        /* Number of lines at the end of the file*/
        case 'e':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*n_elements) = -atoi(optarg);
            break;

        /* Error, so show help and return error code */
        default : 
            fprintf(stderr, "\nError: Unrecognized command-line option.\n\n");
            /* fall-through to help */

            /* help, show usage and exit */
        case 'h':
            goto USAGE;
        }
    }

    return;

    /**USAGE*****************************************************************/

    USAGE:

    fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
        cmd_description, cmd_option, cmd_example,
        "For more information about BioHDF visit http://www.biohdf.org/\n");

    exit(EXIT_FAILURE);
}
