/*****************************************************************************
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 *****************************************************************************/

#include "config.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(HAVE_UNISTD_H)
    #include <unistd.h>    /* getopt processing */
    #include <getopt.h>
#elif defined(_WIN32)
    #include "bsd_getopt.h"
#endif

#include "biohdf.h"
#include "bioh5g.h"



/* SAM data container for file parse */
typedef struct
{
    char           QNAME[BIOHDF_MAX_STRING_SIZE];
    long int       FLAGS;
    char           RNAME[BIOHDF_MAX_STRING_SIZE];
    long int       POS;
    char           MAPQ;
    char           CIGAR[BIOHDF_MAX_STRING_SIZE];
    char           RNEXT[BIOHDF_MAX_STRING_SIZE];
    long int       PNEXT;
    long int       TLEN;
    char           SEQ[BIOHDF_MAX_STRING_SIZE];
    char           QUAL[BIOHDF_MAX_STRING_SIZE];
    char           TAGS[BIOHDF_MAX_STRING_SIZE];
} sam_data;



static void 
_parse_options(int argc,
               char *argv[],
               /*OUT*/ char                      **biohdf_filename,
               /*OUT*/ char                      **input_filename,
               /*OUT*/ char                      **reads_path,
               /*OUT*/ char                      **alignments_path,
               /*OUT*/ bioh5g_alignments_format  *input_format,
               /*OUT*/ size_t                    *seqs_length,
               /*OUT*/ size_t                    *ids_length,
               /*OUT*/ size_t                    *refs_length,
               /*OUT*/ size_t                    *cigar_length,
               /*OUT*/ size_t                    *tags_length,
               /*OUT*/ int                       *compression_level,
               /*OUT*/ int64_t                   *chunk_size_bytes,
               /*OUT*/ int                       *do_not_index,
               /*OUT*/ int                       *do_not_store_header,
               /*OUT*/ char                      **path_to_index_temp_files,
               /*OUT*/ int64_t                   *max_memory_to_use_during_indexing);

static int _parse_sam_line(char *line, /*OUT*/ sam_data *data);


int main(int argc, char **argv)
{
    /* BioHDF file variables. */
    biohdf_file       f                  = NULL;
    biohdf_open_mode  out_mode           = READ_WRITE;
    char             *reads_path         = NULL;
    char             *alignments_path    = NULL;

    /* Alignment hit and reads variables. */
    bioh5g_reads                   r           = NULL;
    bioh5g_read_data              *r_data      = NULL;

    bioh5g_alignments          a           = NULL;
    bioh5g_alignment_data     *a_data      = NULL;
    bioh5g_alignments_format   a_format    = SAM_FORMAT;

    /* BioHDF string storage lengths. */
    size_t seqs_length    = 0;
    size_t ids_length     = 0;
    size_t refs_length    = 0;
    size_t tags_length    = 0;
    size_t cigar_length   = 0;

    /* Do we need to index the alignments? Default is 'yes' */
    int    do_not_index = 0;

    /* HDF5 compression (zlib) size and chunk size */
    int         compression_level   = BIOHDF_DEFAULT_COMPRESSION_LEVEL;
    int64_t     chunk_size_bytes    = BIOHDF_DEFAULT_CHUNK_SIZE;

    /* Properties are only needed if you create the HDF5 datasets. */
    bioh5g_reads_properties                  read_props   = NULL;
    bioh5g_alignments_creation_properties    align_props  = NULL;

    /* Index creation properties */
    biohdf_index_creation_properties index_props = NULL;
    char *path_to_index_temp_files = NULL;
    int64_t max_memory_to_use_during_indexing = -1;

    /* File parsing variables */
    int     presence            = 0;
    FILE   *in                  = NULL;
    FILE   *test                = NULL;
    char   *in_filename         = NULL;
    char   *biohdf_filename     = NULL;

    char              line[BIOHDF_MAX_STRING_SIZE];

    /* Header parsing variables.  The flag is used to determine when
     * the entire header has been parsed.
     */
    int        parsing_header    = 1;
    char      *header            = NULL;
    size_t     line_length       = 0;
    size_t     header_length     = 0;
    size_t     header_size       = 0;

    /* Data structures for parse and storage */
    sam_data       *sam         = NULL;
    int64_t         read_index  = -1;

    /* Whether or not to store the file header */
    int             do_not_store_header;

    biohdf_error    error       = BIOHDF_NO_ERROR;  /* BioHDF error code */
    int             stdio_err   = 0;                /* stdio error code */

    /**CODE*****************************************************************/

    /* Parse command line options */
    _parse_options(argc, argv,
        &biohdf_filename,
        &in_filename,
        &reads_path,
        &alignments_path,
        &a_format,
        &seqs_length,
        &ids_length,
        &refs_length,
        &cigar_length,
        &tags_length,
        &compression_level,
        &chunk_size_bytes,
        &do_not_index,
        &do_not_store_header,
        &path_to_index_temp_files,
        &max_memory_to_use_during_indexing);

    if(NULL == biohdf_filename)
    {
        printf("ERROR: must specify a BioHDF filename with -f\n");
        exit(EXIT_FAILURE);
    }

    if(NULL == reads_path)
    {
        printf("ERROR: must specify a reads collection path with -R\n");
        exit(EXIT_FAILURE);
    }

    if(NULL == alignments_path)
    {
        printf("ERROR: must specify an alignments collection path with -A\n");
        exit(EXIT_FAILURE);
    }

    /* Open import file or stream */
    if(in_filename != NULL)
    {
        in = fopen(in_filename, "r");
        CHECK_PTR(in);
    }
    else
    {
        in = stdin;
    }
    if(NULL == in)
    {
        fprintf(stderr, "Bad input file name: %s\n", in_filename);
        exit(EXIT_FAILURE);
    }

    /* Create or open the BioHDF file */
    test = fopen(biohdf_filename, "r");
    if(test)
    {
        fclose(test);
        error = BIOHDFopen_biohdf_file(biohdf_filename, out_mode, &f);
        CHECK_BIO(error);
    }
    else
    {
        error = BIOHDFcreate_biohdf_file(biohdf_filename, &f);
        CHECK_BIO(error);
    }

    /* Create or open the BioHDF reads storage */
    error = BIOH5Gcheck_reads_presence(f, reads_path, &presence);
    CHECK_BIO(error);
    if(1 == presence)
    {
        error = BIOH5Gopen_reads_collection(f,
            reads_path,
            out_mode,
            &r);
        CHECK_BIO(error);
    }
    else
    {
        error = BIOH5Gcreate_reads_properties(&read_props);
        CHECK_BIO(error);

        /* Set sequence / quality value string storage scheme */
        if(seqs_length > 0)
        {
            error = BIOH5Gset_reads_properties_sequences_scheme(read_props, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_reads_properties_sequences_length(read_props, seqs_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_reads_properties_sequences_scheme(read_props, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        /* Set identifier string storage scheme */
        if(ids_length > 0)
        {
            error = BIOH5Gset_reads_properties_identifiers_scheme(read_props, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_reads_properties_identifiers_length(read_props, ids_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_reads_properties_identifiers_scheme(read_props, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        error = BIOH5Gset_reads_properties_chunk_size(read_props, chunk_size_bytes);
        CHECK_BIO(error);
        error = BIOH5Gset_reads_properties_compression_level(read_props, compression_level);
        CHECK_BIO(error);
        error = BIOH5Gset_reads_properties_reads_type(read_props, BASE_SPACE);
        CHECK_BIO(error);

        error = BIOH5Gcreate_reads_collection(f,
            read_props,
            reads_path,
            &r);
        CHECK_BIO(error);
    }


    /* Create or open the BioHDF alignments storage */
    error = BIOH5Gcheck_alignments_presence(f, alignments_path, &presence);
    CHECK_BIO(error);
    if(presence > 0)
    {
        error = BIOH5Gopen_alignments_collection(f,
            alignments_path,
            out_mode,
            &a);
        CHECK_BIO(error);
    }
    else
    {
        error = BIOH5Gcreate_alignments_properties(&align_props);
        CHECK_BIO(error);

        /* Set reference string storage scheme */
        if(refs_length > 0)
        {
            error = BIOH5Gset_alignments_properties_refs_scheme(align_props, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_alignments_properties_refs_length(align_props, refs_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_alignments_properties_refs_scheme(align_props, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        /* Set CIGAR string storage scheme */
        if(cigar_length > 0)
        {
            error = BIOH5Gset_alignments_properties_cigar_scheme(align_props, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_alignments_properties_cigar_length(align_props, cigar_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_alignments_properties_cigar_scheme(align_props, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        /* Set tags string storage scheme */
        if(tags_length > 0)
        {
            error = BIOH5Gset_alignments_properties_tags_scheme(align_props, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_alignments_properties_tags_length(align_props, tags_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_alignments_properties_tags_scheme(align_props, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        error = BIOH5Gset_alignments_properties_reads_path(align_props, reads_path);
        CHECK_BIO(error);
        error = BIOH5Gset_alignments_properties_chunk_size(align_props, chunk_size_bytes);
        CHECK_BIO(error);
        error = BIOH5Gset_alignments_properties_compression_level(align_props, compression_level);
        CHECK_BIO(error);

        error = BIOH5Gcreate_alignments_collection(f,
            align_props,
            alignments_path,
            &a);
        CHECK_BIO(error);
    }


    /* Set up read data memory. We reuse this allocated memory */
    r_data = (bioh5g_read_data *)calloc(1, sizeof(bioh5g_read_data));
    CHECK_PTR(r_data);
    r_data->identifier      = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(r_data->identifier);
    r_data->sequence        = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(r_data->sequence);
    r_data->quality_values  = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(r_data->quality_values);

    /* Set up alignment data memory. We reuse this allocated memory */
    a_data = (bioh5g_alignment_data *)calloc(1, sizeof(bioh5g_alignment_data));
    CHECK_PTR(a_data);
    a_data->reference  = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(a_data->reference);
    a_data->sam_cigar  = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(a_data->sam_cigar);
    a_data->sam_tags   = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(a_data->sam_tags);
    a_data->sam_rnext  = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(a_data->sam_rnext);

    /* Set up for header parse */
    parsing_header = 1;
    header_size    = 4096;  /* Initial storage size - this can grow */
    header_length  = 0;     /* Actual length of header */
    header         = (char *)calloc(header_size, sizeof(char));
    CHECK_PTR(header);

    /* SAM data element.  Like other parse memory, this is reused. */
    sam = (sam_data *)calloc(1, sizeof(sam_data));
    CHECK_PTR(sam);

    /* Parse SAM file */
    while(fgets(line, BIOHDF_MAX_STRING_SIZE, in))
    {
        if('@' == line[0])
        {
            /* Add to the header string */
            line_length = strlen(line);
            header_length += line_length;
            if (header_length >= header_size)
            {
                while(header_length >= header_size)
                {
                    header_size *= 2;
                }
                header = (char *)realloc(header, header_size);
                CHECK_PTR(header);
            }
            strcat(header, line);
        }
        else if (' '  == line[0] || '\n' == line[0] || '\t' == line[0])
        {
            /* Ignore whitespace lines */
            continue;
        }
        else
        {
            /* Store the header string if this is the first entry */
            if(1 == parsing_header)
            {
                if(!do_not_store_header && strlen(header) > 0)
                {
                    /* Store header string */
                    error = BIOH5Gstore_alignment_file_header(a, SAM_FORMAT, header);
                    CHECK_BIO(error);
                }
                parsing_header = 0;
            }

            /* Fill fields from SAM file line.
             * Must be at least 11 fields (tags are optional)
             */
            if (_parse_sam_line(line, sam) >= 11)
            {
                /* Fill and store read data */
                strncpy(r_data->identifier,      sam->QNAME, BIOHDF_MAX_STRING_SIZE);
                strncpy(r_data->sequence,        sam->SEQ,   BIOHDF_MAX_STRING_SIZE);
                strncpy(r_data->quality_values,  sam->QUAL,  BIOHDF_MAX_STRING_SIZE);
                
                error = BIOH5Gadd_read(r, r_data);
                CHECK_BIO(error);
                error = BIOH5Gget_index_of_last_added_read(r, &read_index);
                CHECK_BIO(error);

                /* Fill and store alignment data */
                a_data->read_index = read_index;
                a_data->position   = sam->POS;
                a_data->length     = (int32_t)strlen(r_data->sequence);
                a_data->sam_mapq   = sam->MAPQ;
                a_data->sam_flags  = sam->FLAGS;
                a_data->sam_pnext  = sam->PNEXT;
                a_data->sam_tlen   = sam->TLEN;
                strncpy(a_data->reference,  sam->RNAME,  BIOHDF_MAX_STRING_SIZE);
                strncpy(a_data->sam_cigar,  sam->CIGAR,  BIOHDF_MAX_STRING_SIZE);
                strncpy(a_data->sam_rnext,  sam->RNEXT,  BIOHDF_MAX_STRING_SIZE);
                strncpy(a_data->sam_tags,   sam->TAGS,   BIOHDF_MAX_STRING_SIZE);

                error = BIOH5Gadd_alignment(a, a_data);
                CHECK_BIO(error);
            }
        }
    }

    /* Index alignments */
    if(!do_not_index)
    {
        error = BIOHDFcreate_index_creation_properties(&index_props);
        if(max_memory_to_use_during_indexing != -1)
        {
            error = BIOHDFset_index_creation_properties_max_memory_mb(index_props, max_memory_to_use_during_indexing);
            CHECK_BIO(error);
        }
        if(path_to_index_temp_files != NULL)
        {
            error = BIOHDFset_index_creation_properties_temp_path(index_props, path_to_index_temp_files);
            CHECK_BIO(error);
        }

        error = BIOH5Gcreate_alignments_index(a, REF_POS_SECONDARY, index_props);
        CHECK_BIO(error);
    }

    /**SUCCESS***************************************************************/

    /* Free memory and close files */
    if(in != stdin)
    {
        stdio_err = fclose(in);
        CHECK_INT_NE(stdio_err, EOF);
    }

    error = BIOH5Gclose_alignments_collection(&a);
    CHECK_BIO(error);
    error = BIOH5Gclose_reads_collection(&r);
    CHECK_BIO(error);

    error = BIOH5Gdestroy_reads_properties(&read_props);
    CHECK_BIO(error);
    error = BIOH5Gdestroy_alignments_properties(&align_props);
    CHECK_BIO(error);
    error = BIOHDFdestroy_index_creation_properties(&index_props);
    CHECK_BIO(error);

    error = BIOHDFclose_biohdf_file(&f);
    CHECK_BIO(error);

    free(header);
    free(sam);

    free(r_data->identifier);
    free(r_data->sequence);
    free(r_data->quality_values);
    free(r_data);

    free(a_data->reference);
    free(a_data->sam_cigar);
    free(a_data->sam_tags);
    free(a_data->sam_rnext);
    free(a_data);

    return EXIT_SUCCESS;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** ERROR ***  in bioh5g_import_alignments\n");

    /* Try to close/free as much as possible */

    if(in != stdin)
    {
        fclose(in);
    }

    BIOH5Gclose_alignments_collection(&a);
    BIOH5Gclose_reads_collection(&r);

    BIOH5Gdestroy_reads_properties(&read_props);
    BIOH5Gdestroy_alignments_properties(&align_props);
    BIOHDFdestroy_index_creation_properties(&index_props);

    BIOHDFclose_biohdf_file(&f);

    free(header);
    free(sam);

    free(r_data->identifier);
    free(r_data->sequence);
    free(r_data->quality_values);
    free(r_data);

    free(a_data->reference);
    free(a_data->sam_cigar);
    free(a_data->sam_tags);
    free(a_data->sam_rnext);
    free(a_data);

    return EXIT_FAILURE;
}


/*****************************************************************************
 * Usage strings                                                             *
 *****************************************************************************/

const char *cmd_description = 
    "Import alignments into a BioHDF file.\n";


const char *cmd_option = 
    "  [--biohdf-file|-f <filename>] (REQUIRED ARGUMENT)\n"
    "      The BioHDF file name.  We recommend using the .h5 extension.\n"
    "\n"
    "  [--alignments-collection|-A <collection>] (REQUIRED ARGUMENT)\n"
    "      The named location in the BioHDF file where the alignments will\n"
    "      be stored.  Groups which do not exist will be created.\n"
    "      \n"
    "      This should look like a filesystem path.  e.g. /path/to/alignments\n"
    "      Collection paths must start from the root group '/'.  The\n"
    "      trailing slash is optional.\n"
    "\n"
    "  [--reads-collection|-R <collection>] (REQUIRED ARGUMENT)\n"
    "      The named location in the BioHDF file where the reads will be\n"
    "      stored.  Groups which do not exist will be created.\n"
    "      \n"
    "      This should look like a filesystem path.  e.g. /path/to/reads\n"
    "      Collection paths must start from the root group '/'.  The\n"
    "      trailing slash is optional.\n"
    "\n"
    "  [--in-file|-i <filename>]\n"
    "      Input SAM file name or the string 'stdin'.\n"
    "      If -i is not specified, the input defaults to standard input.\n"
    "\n"
    "  [--in-format|-F sam]\n"
    "      Input file format for the alignments.  Default is SAM.\n"
    "\n"
    "  [--seqs-length |-s <length>]\n"
    "  [--ids-length  |-d <length>]\n"
    "  [--refs-length |-r <length>]\n"
    "  [--cigar-length|-c <length>]\n"
    "  [--tags-length |-t <length>]\n"
    "      The max length of various strings.  These should be specified if it\n"
    "      can be determined as there are significant performance gains in file\n"
    "      size and access speed when this is known.\n"
    "\n"
    "      *** WARNING ***\n"
    "      Specifying too large of a value wastes space as the empty trailing.\n"
    "      space is stored in the BioHDF file.\n"
    "\n"
    "      *** WARNING ***\n"
    "      If too small a value is chosen and the program attempts to store\n"
    "      data that is too large in the BioHDF file, the program will abort\n"
    "      with an exit code of -1 and write an error message to stderr.\n"
    "\n"
    "      NOTE: The BioHDF distribution inclues a bioh5g_preparse_sam tool\n"
    "      which can be used to determine appropriate field sizes.\n"
    "\n"
    "  [--chunk-size|-C <HDF5 chunk size in bytes>]\n"
    "      The HDF5 chunk size in bytes.  Default is 1024. \n"
    "      This can have a huge effect on performance so it can be useful to\n"
    "      experiment with this value.\n"
    "\n"
    "  [--compression-level|-Z <integer compression level from 1 to 10>]\n"
    "      The zlib compression level.  Default is 5.\n"
    "\n"
    "  [--do-not-index]\n"
    "      Do not create an index on the alignments.\n"
    "\n"
    "  [--temp-file-path|-T <unsigned integer SAM FLAG mask>]\n"
    "      Filter and return only those alignments which have all the bits\n"
    "      in the mask set.  Default is the current working directory.\n"
    "\n"
    "      NOTE: This path MUST end in a file separator ('\\' or '/',\n"
    "      depending on platform).\n"
    "\n"
    "  [--max-index-memory-mb|-M <max memory to use (MB)>]\n"
    "      The maximum amount of memory (MB) to use while creating the index.\n"
    "      Default is 512 MB.\n"
    "\n"
    "  [--do-not-store-header]\n"
    "      Do not store the file header.\n"
    "\n"
    "      This is MANDATORY if you are importing multiple header-containing\n"
    "      files into a single collection.\n"
    "\n"
    "  [--help|-h]\n"
    "      Command line help message.\n";

const char *cmd_example =
    "bioh5g_import_alignments -f myfile.h5 -i inputfile.sam\n"
    "     -R /sample1/reads -A /sample1/alignments/aligner_params_1/\n"
    "     -s 100 -d 100 -r 5 -c 35\n";



static void 
_parse_options(int argc,
               char *argv[],
               /*OUT*/ char                      **biohdf_filename,
               /*OUT*/ char                      **input_filename,
               /*OUT*/ char                      **reads_path,
               /*OUT*/ char                      **alignments_path,
               /*OUT*/ bioh5g_alignments_format  *input_format,
               /*OUT*/ size_t                    *seqs_length,
               /*OUT*/ size_t                    *ids_length,
               /*OUT*/ size_t                    *refs_length,
               /*OUT*/ size_t                    *cigar_length,
               /*OUT*/ size_t                    *tags_length,
               /*OUT*/ int                       *compression_level,
               /*OUT*/ int64_t                   *chunk_size_bytes,
               /*OUT*/ int                       *do_not_index,
               /*OUT*/ int                       *do_not_store_header,
               /*OUT*/ char                      **path_to_index_temp_files,
               /*OUT*/ int64_t              *max_memory_to_use_during_indexing)
{
    int c = -1;
    int option_index = -1;

    /* Short options */
    char *options = "f:i:A:R:F:s:d:r:c:t:C:Z:T:M:h";

    /* Long options */
    struct option long_options[] = {
        {"biohdf-file",           required_argument, 0, 'f'},
        {"in-file",               required_argument, 0, 'i'},

        {"reads-collection",      required_argument, 0, 'R'},
        {"alignments-collection", required_argument, 0, 'A'},
        
        {"in-format",             required_argument, 0, 'F'},

        {"seqs-length",           required_argument, 0, 's'},
        {"ids-length",            required_argument, 0, 'd'},
        {"refs-length",           required_argument, 0, 'r'},
        {"cigar-length",          required_argument, 0, 'c'},
        {"tags-length",           required_argument, 0, 't'},

        {"chunk-size",            required_argument, 0, 'C'},
        {"compression-level",     required_argument, 0, 'Z'},

        {"temp-file-path",        required_argument, 0, 'T'},
        {"max-index-memory-mb",   required_argument, 0, 'M'},

        {"do-not-index",          no_argument,       do_not_index, 1},
        {"do-not-store-header",   no_argument,       do_not_store_header, 1},

        {"help",                  no_argument,       0, 'h'},
        {0, 0, 0, 0}
    };

    if(argc <= 1)
    {
        goto USAGE;
    }

    /* Initialize */
    biohdf_filename[0]   = '\0';
    input_filename[0]    = '\0';
    reads_path[0]        = '\0';
    alignments_path[0]         = '\0';
    *input_format        = SAM_FORMAT;
    *seqs_length         = 0;
    *ids_length          = 0;
    *refs_length         = 0;
    *cigar_length        = 0;
    *tags_length         = 0;
    *do_not_index        = 0;
    *do_not_store_header = 0;


    /* Parse options */
    opterr = 0;
    for (;;)
    {
        c = getopt_long(argc, argv, options, long_options, &option_index);
        if (-1 == c) { break; }

        switch (c)
        {
        case 0:
            /* If this option sets a flag, do nothing else now. */
            break;

        /* BioHDF file */
        case 'f':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*biohdf_filename) = (char *)calloc(strlen(optarg) + 1, sizeof(char));
            strcpy((*biohdf_filename), optarg);
            break;

        /* Input file */
        case 'i':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*input_filename) = (char *)calloc(strlen(optarg) + 1, sizeof(char));
            strcpy((*input_filename), optarg);
            break;

        /* Path to reads */
        case 'R':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*reads_path) = (char *)calloc(strlen(optarg) + 1, sizeof(char));
            strcpy((*reads_path), optarg);
            break;

        /* Path to alignments */
        case 'A':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*alignments_path) = (char *)calloc(strlen(optarg) + 1, sizeof(char));
            strcpy((*alignments_path), optarg);
            break;

        /* Input format */
        case 'F':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *input_format = SAM_FORMAT;
            break;

        /* Sequences/quality values storage length */
        case 's':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *seqs_length = atoi(optarg);
            break;

        /* IDs storage length */
        case 'd':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *ids_length = atoi(optarg);
            break;

        /* Reference strings storage length */
        case 'r':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *refs_length = atoi(optarg);
            break;

        /* CIGAR strings storage length */
        case 'c':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *cigar_length = atoi(optarg);
            break;

        /* Tag strings storage length */
        case 't':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *tags_length = atoi(optarg);
            break;

        /* Compression level */
        case 'Z':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *compression_level = atoi(optarg);
            break;

        /* HDF5 chunk size */
        case 'C':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *chunk_size_bytes = (int64_t)atoi(optarg);
            break;

        /* Temp file path for index creation */
        case 'T':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*path_to_index_temp_files) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*path_to_index_temp_files), optarg);
            break;

        /* Max amount of memory to use while creating indexes */
        case 'M':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*max_memory_to_use_during_indexing) = atol(optarg);
            break;

        /* Error, so show help and return error code */
        default : 
            fprintf(stderr, "\nError: Unrecognized command-line option.\n\n");
            /* fall-through to help */

            /* help, show usage and exit */
        case 'h':
            goto USAGE;
        }
    }

    return;

USAGE:
    fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
        cmd_description, cmd_option, cmd_example,
        "For more information about BioHDF visit http://www.biohdf.org/\n");
    exit(0);
}



static int
_parse_sam_line(char *line, /*OUT*/ sam_data *sam)
{
    int             i;
    int             j;
    int             field;
    int             len;
    char            buf[BIOHDF_MAX_STRING_SIZE];

    /**CODE*****************************************************************/

    memset(sam, 0, sizeof(sam_data));
    field   = 0;
    len     = (int)strlen(line);
    i       = 0;

    while(i <= len)
    {
        /* initialize */
        memset(buf, 0, BIOHDF_MAX_STRING_SIZE);
        j = 0;
        field++;

        /* Read the next token (tab-delimited) */
        /* When we get to the tags (field 12) we read to the end of the line */
        while(i <= len && (line[i] != '\t' || 12 == field))
        {
            buf[j] = line[i];
            i++;
            j++;
        }

        /* move past the tab */
        i++;

        /* Potentially fix up newlines if this is the last field */
        /* j is one past the end of the string, j-1 is the string terminating nul */
        if('\n' == buf[j-2])
        {
            buf[j-2] = '\0';
        }

        /* store the token */
        switch(field)
        {
        case 1:
            strcpy(sam->QNAME, buf);
            break;
        case 2:
            if('*' == buf[0])
            {
                /* FLAG can be *, which is weird */
                sam->FLAGS = 0;
            }
            else
            {
                sam->FLAGS = atoi(buf);
            }
            break;
        case 3:
            if('*' == buf[0] && 1 == strlen(buf))
            {
                /* Wasteful to store the '*' */
                strcpy(sam->RNAME, "\0");
            }
            else
            {
                strcpy(sam->RNAME, buf);
            }
            break;
        case 4:
            sam->POS = atoi(buf);
            break;
        case 5:
            sam->MAPQ = atoi(buf);
            break;
        case 6:
            if('*' == buf[0] && 1 == strlen(buf))
            {
                /* Wasteful to store the '*' */
                strcpy(sam->CIGAR, "\0");
            }
            else
            {
                strcpy(sam->CIGAR, buf);
            }
            break;
        case 7:
            if('*' == buf[0] && 1 == strlen(buf))
            {
                /* Wasteful to store the '*' */
                strcpy(sam->RNEXT, "\0");
            }
            else
            {
                strcpy(sam->RNEXT, buf);
            }
            break;
        case 8:
            sam->PNEXT = atoi(buf);
            break;
        case 9:
            sam->TLEN = atoi(buf);
            break;
        case 10:
            if('*' == buf[0] && 1 == strlen(buf))
            {
                /* Wasteful to store the '*' */
                strcpy(sam->SEQ, "\0");
            }
            else
            {
                strcpy(sam->SEQ, buf);
            }
            break;
        case 11:
            if('*' == buf[0] && 1 == strlen(buf))
            {
                /* Wasteful to store the '*' */
                strcpy(sam->QUAL, "\0");
            }
            else
            {
                strcpy(sam->QUAL, buf);
            }
            break;
        case 12:
            strcpy(sam->TAGS, buf);
            break;
        default:
            return -1;
        }
    }

    return field;
}
