/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "config.h"

#include <inttypes.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(HAVE_UNISTD_H)
    #include <unistd.h>    /* getopt processing */
    #include <getopt.h>
#elif defined(_WIN32)
    #include "bsd_getopt.h"
#endif

#include "biohdf.h"
#include "bioh5g.h"


#ifdef _WIN32
    #define LINE_MAX MAX_INPUT
#endif


typedef enum
{
    PARSING_ID,
    PARSING_SEQ,
    PARSING_PLUS,
    PARSING_QUALS
} fastq_parse_state;


static void _parse_options(int argc,
                           char *argv[],
                           /*OUT*/ char **biohdf_file_name,
                           /*OUT*/ char **reads_path,
                           /*OUT*/ char **input_file,
                           /*OUT*/ bioh5g_reads_format *input_format,
                           /*OUT*/ size_t *seqs_length,
                           /*OUT*/ size_t *ids_length,
                           /*OUT*/ int *compression_level,
                           /*OUT*/ long int *chunk_size);


int main(int argc, char **argv)
{
    biohdf_file              f          = NULL;
    bioh5g_reads             r          = NULL;
    bioh5g_read_data        *data       = NULL;
    bioh5g_reads_format      format     = FASTQ_FORMAT;
    fastq_parse_state        state      = PARSING_SEQ;

    size_t seqs_length  = 0;
    size_t ids_length   = 0;

    bioh5g_reads_properties  properties   = NULL;

    FILE               *in                = NULL;
    FILE               *test              = NULL;
    char               *in_filename       = NULL;
    char               *out_filename      = NULL;
    biohdf_open_mode    out_mode          = READ_WRITE;
    char               *reads_path        = NULL;
    int                 presence          = 0;

    /* HDF5 compression (zlib) size and chunk size */
    int             compression_level = BIOHDF_DEFAULT_COMPRESSION_LEVEL;
    long int        chunk_size        = BIOHDF_DEFAULT_CHUNK_SIZE;

    /* Used by the parser to trim lines */
    char       *head    = 0;
    size_t      len     = 0;

    char line[LINE_MAX];

    /* The line number we're reading.  Used to generate an error
     * message if we attempt to store data that would be truncated
     */
    int64_t line_number = 0;

    biohdf_error    error       = BIOHDF_NO_ERROR;  /* BioHDF error code */
    int             stdio_err   = 0;                /* stdio error code */

    /**CODE*****************************************************************/

    /* Parse command line options */
    _parse_options(argc, argv,
        &out_filename,
        &reads_path,
        &in_filename,
        &format,
        &seqs_length,
        &ids_length,
        &compression_level,
        &chunk_size);

    if(NULL == out_filename)
    {
        printf("ERROR: must specify a BioHDF filename with -f\n");
        exit(EXIT_FAILURE);
    }

    if(NULL == reads_path)
    {
        printf("ERROR: must specify a reads collection path with -R\n");
        exit(EXIT_FAILURE);
    }

    /* Open import file or stream */
    if(in_filename != NULL)
    {
        in = fopen(in_filename, "r");
        CHECK_PTR(in);
    }
    else
    {
        in = stdin;
    }

    /* Create or open the BioHDF file */
    test = fopen(out_filename, "r");
    if(test)
    {
        fclose(test);
        error = BIOHDFopen_biohdf_file(out_filename, out_mode, &f);
        CHECK_BIO(error);
    }
    else
    {
        error = BIOHDFcreate_biohdf_file(out_filename, &f);
        CHECK_BIO(error);
    }

    /* Create or open the BioHDF reads storage */
    error = BIOH5Gcheck_reads_presence(f, reads_path, &presence);
    CHECK_BIO(error);
    if(1 == presence)
    {
        error = BIOH5Gopen_reads_collection(f,
            reads_path,
            out_mode,
            &r);
        CHECK_BIO(error);
    }
    else
    {
        error = BIOH5Gcreate_reads_properties(&properties);
        CHECK_BIO(error);

        /* Set sequence / quality value string storage scheme */
        if(seqs_length > 0)
        {
            error = BIOH5Gset_reads_properties_sequences_scheme(properties, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_reads_properties_sequences_length(properties, seqs_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_reads_properties_sequences_scheme(properties, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        /* Set identifier string storage scheme */
        if(ids_length > 0)
        {
            error = BIOH5Gset_reads_properties_identifiers_scheme(properties, FIXED_LENGTH_STRING);
            CHECK_BIO(error);
            error = BIOH5Gset_reads_properties_identifiers_length(properties, ids_length);
            CHECK_BIO(error);
        }
        else
        {
            error = BIOH5Gset_reads_properties_identifiers_scheme(properties, VARIABLE_LENGTH_STRING);
            CHECK_BIO(error);
        }

        error = BIOH5Gset_reads_properties_chunk_size(properties, chunk_size);
        CHECK_BIO(error);
        error = BIOH5Gset_reads_properties_compression_level(properties, compression_level);
        CHECK_BIO(error);

        /* TODO: Handle this properly later */
        error = BIOH5Gset_reads_properties_reads_type(properties, BASE_SPACE);
        CHECK_BIO(error);

        error = BIOH5Gcreate_reads_collection(f,
            properties,
            reads_path,
            &r);
        CHECK_BIO(error);

        error = BIOH5Gdestroy_reads_properties(&properties);
        CHECK_BIO(error);
    }

    /* Set up read data memory. We reuse this allocated memory */
    data = (bioh5g_read_data *)calloc(1, sizeof(bioh5g_read_data));
    CHECK_PTR(data);
    data->identifier = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(data->identifier);
    data->sequence = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(data->sequence);
    data->quality_values = (char *)calloc(BIOHDF_MAX_STRING_SIZE, sizeof(char));
    CHECK_PTR(data->quality_values);


    /* Parse the file */
    state = PARSING_ID;


    if(FASTQ_FORMAT == format)
    {
        line_number = 0;

        while(fgets(line, LINE_MAX, in))
        {
            line_number++;

            if (PARSING_ID == state)
            {
                /* Need to remove '@', leading whitespace and newline. */
                if ('@' == line[0])
                {
                    /* Need to remove delimiter, leading whitespace and newline. */
                    len = strlen(line);
                    if('\n' == line[len - 1])
                    {
                        line[len - 1] = '\0';
                        len--;
                    }
                    head = line;
                    while('@' == *head || ' ' == *head)
                    {
                        head++;
                        len--;
                    }
                    strncat(data->identifier, head, len);

                    state = PARSING_SEQ;
                }
            }
            else if(PARSING_SEQ == state)
            {
                /* Trim newline */
                len = strlen(line);
                if('\n' == line[len - 1])
                {
                    line[len - 1] = '\0';
                }
                
                /* Add to the sequence string */
                strcat(data->sequence, line);

                state = PARSING_PLUS;
            }
            else if(PARSING_PLUS == state)
            {
                if ('+' != line[0])
                {
                    fprintf(stderr, "PARSER ERROR AT LINE %" PRId64 "\n", line_number);
                    exit(-1);
                }
                state = PARSING_QUALS;
            }
            else  /* PARSING_QUALS */
            {
                /* trim newline */
                len = strlen(line);
                if('\n' == line[len - 1])
                {
                    line[len - 1] = '\0';
                }

                /* store the quality values */
                strcat(data->quality_values, line);

                /* Commit data */
                error = BIOH5Gadd_read(r, data);
                if (BIOHDF_BADNESS == error)
                {
                    fprintf(stderr, "\tNOTE:\n");
                    fprintf(stderr, "\tATTEMPTING TO STORE DATA WHICH WOULD BE TRUNCATED\n");
                    fprintf(stderr, "\tIS A COMMON CAUSE OF THIS ERROR.\n");
                    fprintf(stderr, "\tERROR WAS AT INPUT FILE LINE: %" PRId64 "\n", line_number);
                }
                CHECK_BIO(error);

                /* Reset buffers */
                (data->identifier)[0] = '\0';
                (data->sequence)[0] = '\0';
                (data->quality_values)[0] = '\0';

                state = PARSING_ID;
            }
        }
    }
    if(FASTA_FORMAT == format)
    {
        while(fgets(line, LINE_MAX, in))
        {
            line_number++;
            if('>' == line[0])
            {
                if((data->sequence)[0] != '\0')
                {
                    /* Commit data */
                    error = BIOH5Gadd_read(r, data);
                    if (BIOHDF_BADNESS == error)
                    {
                        fprintf(stderr, "\tNOTE:\n");
                        fprintf(stderr, "\tATTEMPTING TO STORE DATA WHICH WOULD BE TRUNCATED\n");
                        fprintf(stderr, "\tIS A COMMON CAUSE OF THIS ERROR.\n");
                        fprintf(stderr, "\tERROR WAS AT INPUT FILE LINE: %" PRId64 "\n", line_number);
                    }
                    CHECK_BIO(error);

                    /* Reset data */
                    (data->identifier)[0] = '\0';
                    (data->sequence)[0]   = '\0';
                    (data->quality_values)[0]  = '\0';
                }

                /* Need to remove '>', leading whitespace and newline. */
                len = strlen(line);
                if('\n' == line[len - 1])
                {
                    line[len - 1] = '\0';
                    len--;
                }
                head = line;
                while('>' == *head || ' ' == *head)
                {
                    head++;
                    len--;
                }
                strncat(data->identifier, head, len);
                
            }
            else
            {
                /* Naive.  Don't assume MAX_LINE will hold entire sequence */
                len = strlen(line);
                if('\n' == line[len - 1])
                {
                    line[len - 1] = '\0';
                }
                strcat(data->sequence, line);
            }
        }

        /* Add the last read */
        if(strlen(data->sequence) > 0)
        {
            error = BIOH5Gadd_read(r, data);
            CHECK_BIO(error);
        }
    }


    /**SUCCESS***************************************************************/

    /* Free memory and close files */
    if(in != stdout)
    {
        stdio_err = fclose(in);
        CHECK_INT_NE(stdio_err, EOF);
    }

    error = BIOH5Gclose_reads_collection(&r);
    CHECK_BIO(error);
    error = BIOHDFclose_biohdf_file(&f);
    CHECK_BIO(error);
    error = BIOH5Gdestroy_reads_properties(&properties);
    CHECK_BIO(error);

    free(data->quality_values);
    free(data->sequence);
    free(data->identifier);
    free(data);

    free(in_filename);
    free(reads_path);
    free(out_filename);

    return EXIT_SUCCESS;

    /**FAILURE***************************************************************/

    FAIL:

    fprintf(stderr, "*** ERROR ***  in bioh5g_import_reads\n");

    /* Try to close/free as much as possible */

    if(in != stdin)
    {
        fclose(in);
    }
    BIOH5Gclose_reads_collection(&r);
    BIOHDFclose_biohdf_file(&f);
    BIOH5Gdestroy_reads_properties(&properties);

    free(data->quality_values);
    free(data->sequence);
    free(data->identifier);
    free(data);

    free(in_filename);
    free(reads_path);
    free(out_filename);

    return EXIT_FAILURE;
}


/*****************************************************************************
 * Usage strings                                                             *
 *****************************************************************************/

const char *cmd_description = 
    "Import reads into a BioHDF file from either FASTQ or FASTA format.\n";


const char *cmd_option = 
    "  [--biohdf-file|-f <filename>] (REQUIRED ARGUMENT)\n"
    "      The BioHDF file name.  We recommend using the .h5 extension.\n"
    "\n"
    "  [--reads-collection|-R <collection>] (REQUIRED ARGUMENT)\n"
    "      The named location in the BioHDF file where the reads will be\n"
    "      stored.  Groups which do not exist will be created.\n"
    "      \n"
    "      This should look like a filesystem path.  e.g. /path/to/reads\n"
    "      Collection paths must start from the root group '/'.  The\n"
    "      trailing slash is optional.\n"
    "\n"
    "  [--in-file|-i <filename>]\n"
    "      Input  FASTQ or FASTA file name.\n"
    "      Default is STDIN if a filename is not specified.\n"
    "      NOTE: Multi-line FASTQ files are NOT supported.\n"
    "\n"
    "  [--in-format|-F fasta|fastq]\n"
    "      Input file format for read sequences and associated ASCII-encoded \n"
    "      qualities. Default is FASTQ.\n"
    "\n"
    "  [--seqs-length|-s <length>]\n"
    "  [--ids-length| -d <length>]\n"
    "      The max length of a sequence / quality value or identifier line,\n"
    "      if known.  This should be specified if it can be determined as\n"
    "      there are significant performance gains in file size and access\n"
    "      speed when this is known.\n"
    "      *** WARNING ***\n"
    "      Specifying too large of a value wastes space as the empty trailing.\n"
    "      space is stored in the BioHDF file.\n"
    "      *** WARNING ***\n"
    "      If too small a value is chosen and the program attempts to store\n"
    "      data that is too large in the BioHDF file, the program will abort\n"
    "      with an exit code of -1 and write an error message to stderr.\n"
    "\n"
    "  [--chunk-size|-C <HDF5 chunk size in bytes>]\n"
    "      The HDF5 chunk size in bytes.  Default is 1024. \n"
    "      This can have a huge effect on performance so it can be useful to\n"
    "      experiment with this value.\n"
    "\n"
    "  [--compression-level|-Z <integer compression level from 1 to 10>]\n"
    "      The zlib compression level.  Default is 5.\n"
    "\n"
    "  [--help|-h]\n"
    "      Command line help message.\n";

const char *cmd_example =
    "bioh5g_import_reads -f myfile.h5 -R \"/sample1/reads\" -i input.fastq\n"
    "     -F fastq -s 100 -d 100 -Z 8\n"
    "bioh5g_import_reads -f myfile.h5 -R \"/myreads\" -F fastq\n"
    "     < input.fastq\n";



static void _parse_options(int argc,
                           char *argv[],
                           /*OUT*/ char **biohdf_file_name,
                           /*OUT*/ char **reads_path,
                           /*OUT*/ char **input_file,
                           /*OUT*/ bioh5g_reads_format *input_format,
                           /*OUT*/ size_t *seqs_length,
                           /*OUT*/ size_t *ids_length,
                           /*OUT*/ int *compression_level,
                           /*OUT*/ long int *chunk_size)
{
    int c = -1;
    int option_index = -1;

    /* Short options */
    char *options = "f:R:i:F:s:d:C:Z:h";

    /* Long options */
    struct option long_options[] = {
        {"biohdf-file",       required_argument, 0, 'f'},
        {"reads-collection",  required_argument, 0, 'R'},
        {"in-file",           required_argument, 0, 'i'},
        {"in-format",         required_argument, 0, 'F'},
        {"seqs-length",       required_argument, 0, 's'},
        {"ids-length",        required_argument, 0, 'd'},
        {"chunk-size",        required_argument, 0, 'C'},
        {"compression-level", required_argument, 0, 'Z'},
        {"help",              no_argument,       0, 'h'},
        {0, 0, 0, 0}
    };

    if(argc <= 1)
    {
        goto USAGE;
    }

    /* Parse options */
    opterr = 0;
    for (;;)
    {
        c = getopt_long(argc, argv, options, long_options, &option_index);
        if (-1 == c) { break; }

        switch (c)
        {
        case 0:
            /* If this option sets a flag, do nothing else now. */
            break;

        /* BioHDF file */
        case 'f':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*biohdf_file_name) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*biohdf_file_name), optarg);
            break;

        /* Path to reads */
        case 'R':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*reads_path) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*reads_path), optarg);
            break;

        /* Input file */
        case 'i':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            (*input_file) = (char *)calloc(strlen(optarg) + 1,
                sizeof(char));
            strcpy((*input_file), optarg);
            break;

        /* Input format */
        case 'F':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            if(0 == strcmp("fastq", optarg))
            {
                *input_format = FASTQ_FORMAT;
            }
            else
            {
                *input_format = FASTA_FORMAT;
            }
            break;

        /* Sequences/quality values storage length */
        case 's':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *seqs_length = atoi(optarg);
            break;

        /* IDs storage length */
        case 'd':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *ids_length = atoi(optarg);
            break;

        /* Compression level */
        case 'Z':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *compression_level = atoi(optarg);
            break;

        /* HDF5 chunk size */
        case 'C':
            if (optarg == NULL || optarg[0] == '-') goto USAGE;
            *chunk_size = atoi(optarg);
            break;

        /* Error, so show help and return error code */
        default : 
            fprintf(stderr, "\nError: Unrecognized command-line option.\n\n");
            /* fall-through to help */

            /* help, show usage and exit */
        case 'h':
            goto USAGE;
        }
    }

    return;

USAGE:
    fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
        cmd_description, cmd_option, cmd_example,
        "For more information about BioHDF visit http://www.biohdf.org/\n");
    exit(0);
}
