/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group                                                *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the file COPYING.  COPYING can be found at the root of the source code    *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "config.h"

#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* The max string size in some fixed data structures */
/* This is clearly overkill but will handle very large tag strings */
#define MAX_STRING_SIZE 16384


/* SAM data container for file parse */
typedef struct
{
    char            QNAME[MAX_STRING_SIZE];
    long int        FLAGS;
    char            RNAME[MAX_STRING_SIZE];
    long int        POS;
    char            MAPQ;
    char            CIGAR[MAX_STRING_SIZE];
    char            RNEXT[MAX_STRING_SIZE];
    long int        PNEXT;
    long int        TLEN;
    char            SEQ[MAX_STRING_SIZE];
    char            QUAL[MAX_STRING_SIZE];
    char            TAGS[MAX_STRING_SIZE];

    int             has_tags;
} sam_data;


static int _parse_sam_line(char *line, /*OUT*/ sam_data *data);


int main(int argc, char **argv)
{
    FILE        *f      = NULL;

    sam_data    sam;

    int         len     = 0;

    char        line[MAX_STRING_SIZE];

    /* metrics */
    int64_t     n_sam_lines     = 0;
    int64_t     n_lines         = 0;
    int64_t     n_header_lines  = 0;

    int64_t     n_header_bytes  = 0;

    int         qname_min       = MAX_STRING_SIZE;
    int         qname_max       = 0;
    int         rname_min       = MAX_STRING_SIZE;
    int         rname_max       = 0;
    int         cigar_min       = MAX_STRING_SIZE;
    int         cigar_max       = 0;
    int         rnext_min       = MAX_STRING_SIZE;
    int         rnext_max       = 0;
    int         seq_min         = MAX_STRING_SIZE;
    int         seq_max         = 0;
    int         qual_min        = MAX_STRING_SIZE;
    int         qual_max        = 0;
    int         tags_min        = MAX_STRING_SIZE;
    int         tags_max        = 0;

    int         has_tags        = 0;

    /********************************************************************/

    /* Check that we have a filename */
    if(argc != 2)
    {
        printf("Usage: bioh5g_preparse_sam <filename>\n");
        exit(EXIT_FAILURE);
    }

    /* Open the file */
    f = fopen(argv[1], "r");
    if(NULL == f)
    {
        fprintf(stderr, "Unable to open file: %s, exiting\n", argv[1]);
        exit(EXIT_FAILURE);
    }

    printf("Working");

    /* Parse the SAM file */
    while(fgets(line, MAX_STRING_SIZE, f))
    {
        n_lines++;

        /* Show we haven't locked up */
        if(0 == n_lines % 10000)
            printf(".");

        /* Process the header line */
        if('@' == line[0])
        {
            n_header_bytes += (long int)strlen(line);
            n_header_lines++;
            continue;
        }

        /* Check that the line isn't insanely long and trim the newline
         * if it's a data line */
        len = (int)strlen(line);
        if(len > MAX_STRING_SIZE)
        {
            /* In the future I'll probably make the arrays dynamic... */
            fprintf(stderr, "ERROR: SAM line too long to process ( > %d)\n", MAX_STRING_SIZE);
            return EXIT_FAILURE;
        }
        line[len - 1] = '\0';

        /* Parse the SAM line */
        if (0 == _parse_sam_line(line, &sam))
        {
            n_sam_lines++;

            /* Having SEQ and QUAL fields not be equal is typically an error.
             * Might want to have this crash here.
             */
            if(strlen(sam.SEQ) != strlen(sam.QUAL))
            {
                fprintf(stderr, "length of SEQ does not equal length of QUAL at line %" PRId64  ": ", n_lines);
                fprintf(stderr, "SEQ: %s QUAL: %s line: %s\n", sam.SEQ, sam.QUAL, line);
            }

            /* Check lengths of fields for new min or max */
            len = (int)strlen(sam.QNAME);
            if(len < qname_min) qname_min = len;
            if(len > qname_max) qname_max = len;

            len = (int)strlen(sam.RNAME);
            if(len < rname_min) rname_min = len;
            if(len > rname_max) rname_max = len;

            len = (int)strlen(sam.CIGAR);
            if(len < cigar_min) cigar_min = len;
            if(len > cigar_max) cigar_max = len;

            len = (int)strlen(sam.RNEXT);
            if(len < rnext_min) rnext_min = len;
            if(len > rnext_max) rnext_max = len;

            len = (int)strlen(sam.SEQ);
            if(len < seq_min) seq_min = len;
            if(len > seq_max) seq_max = len;

            len = (int)strlen(sam.QUAL);
            if(len < qual_min) qual_min = len;
            if(len > qual_max) qual_max = len;
            
            if(sam.has_tags)
            {
                has_tags = 1;

                len = (int)strlen(sam.TAGS);
                if(len < tags_min) tags_min = len;
                if(len > tags_max) tags_max = len;
            }
        }
    }

    fclose(f);

    printf("\n\n");
    printf("File: %s\n", argv[1]);
    printf("Total number of lines: %" PRId64 "\n", n_lines);
    printf("\n");
    printf("Header stats:\n");
    printf("\tNumber of lines: %" PRId64 "\n", n_header_lines);
    printf("\tSize in bytes (includes @s and newlines): %" PRId64 "\n", n_header_bytes);
    printf("\n");
    printf("SAM data stats:\n");
    printf("\tNumber of valid SAM lines: %" PRId64 "\n", n_sam_lines);
    printf("\n");
    printf("\tQNAME min: %d\n", qname_min);
    printf("\tQNAME max: %d\n", qname_max);
    printf("\n");
    printf("\tRNAME min: %d\n", rname_min);
    printf("\tRNAME max: %d\n", rname_max);
    printf("\n");
    printf("\tCIGAR min: %d\n", cigar_min);
    printf("\tCIGAR max: %d\n", cigar_max);
    printf("\n");
    printf("\tRNEXT min: %d\n", rnext_min);
    printf("\tRNEXT max: %d\n", rnext_max);
    printf("\n");
    printf("\tSEQ min: %d\n", seq_min);
    printf("\tSEQ max: %d\n", seq_max);
    printf("\n");
    printf("\tQUAL min: %d\n", qual_min);
    printf("\tQUAL max: %d\n", qual_max);
    printf("\n");

    if(has_tags)
    {
        printf("\tTAGS min: %d\n", tags_min);
        printf("\tTAGS max: %d\n", tags_max);
    }
    else
    {
        printf("\tTAGS min: N/A (no tags)\n");
        printf("\tTAGS max: N/A (no tags)\n");
    }

    return EXIT_SUCCESS;
}


/* Parses a SAM line.
 *
 * Returns:
 *
 * The SAM fields in a SAM data struct (out parameter)
 *
 * The actual return value is an error code:
 *       0 = no problems
 *      -1 = parse errors
 */
static int
_parse_sam_line(char *line, /*OUT*/ sam_data *sam)
{
    int             i;
    int             j;
    int             field;
    int             len;
    char            buf[MAX_STRING_SIZE];

    /*********************************************************************/

    memset(sam, 0, sizeof(sam_data));
    field   = 0;
    len     = (int)strlen(line);
    i       = 0;

    while(i <= len)
    {
        /* initialize */
        memset(buf, 0, MAX_STRING_SIZE);
        j = 0;
        field++;

        /* Read the next token (tab-delimited) */
        /* When we get to the tags (field 12) we read to the end of the line */
        while(i <= len && (line[i] != '\t' || 12 == field))
        {
            buf[j] = line[i];
            i++;
            j++;
        }

        /* move past the tab */
        i++;

        /* store the token */
        switch(field)
        {
        case 1:
            strcpy(sam->QNAME, buf);
            break;
        case 2:
            sam->FLAGS = atoi(buf);
            break;
        case 3:
            strcpy(sam->RNAME, buf);
            break;
        case 4:
            sam->POS = atoi(buf);
            break;
        case 5:
            sam->MAPQ = atoi(buf);
            break;
        case 6:
            strcpy(sam->CIGAR, buf);
            break;
        case 7:
            strcpy(sam->RNEXT, buf);
            break;
        case 8:
            sam->PNEXT = atoi(buf);
            break;
        case 9:
            sam->TLEN = atoi(buf);
            break;
        case 10:
            strcpy(sam->SEQ, buf);
            break;
        case 11:
            strcpy(sam->QUAL, buf);
            break;
        case 12:
            sam->has_tags = 1;
            strcpy(sam->TAGS, buf);
            break;
        default:
            return -1;
        }
    }

    /* We should have read all the fields, with the possible
     * exception of the TAGS field.
     */
    if(field < 11)
        return -1;
    else
        return 0;
}

